Package: this.path
Version: 2.8.0
Date: 2026-04-09
License: MIT + file LICENSE
Title: Get Executing Script's Path
Description: Determine the path of the executing script. Compatible
        with several popular GUIs: 'Rgui', 'RStudio', 'Positron',
        'VSCode', 'Jupyter', 'Emacs', and 'Rscript' (shell). Compatible
        with several functions and packages: 'source()',
        'sys.source()', 'debugSource()' in 'RStudio',
        'compiler::loadcmp()', 'utils::Sweave()', 'box::use()',
        'knitr::knit()', 'plumber::plumb()', 'shiny::runApp()',
        'package:targets', and 'testthat::source_file()'.
Authors@R: person(given = "Iris",
                  family = "Simmons",
                  role = c("aut", "cre"),
                  email = "ikwsimmo@gmail.com")
Author: Iris Simmons [aut, cre]
Maintainer: Iris Simmons <ikwsimmo@gmail.com>
Depends: R (>= 2.15)
Suggests: utils, jsonlite, microbenchmark, rprojroot
Enhances: compiler, box, knitr, plumber, shiny, targets, testthat
URL: https://github.com/ArcadeAntics/this.path
BugReports: https://github.com/ArcadeAntics/this.path/issues
ByteCompile: TRUE
Biarch: TRUE
BuildManual: TRUE
Type: Package
NeedsCompilation: yes
Packaged: 2026-04-09 03:13:52 UTC; iris
Repository: CRAN
Date/Publication: 2026-04-10 08:50:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-04-10 10:59:35 UTC; unix
Archs: this.path.so.dSYM

Package: this.path
Version: 2.8.0
Date: 2026-04-09
License: MIT + file LICENSE
Title: Get Executing Script's Path
Description: Determine the path of the executing script. Compatible
        with several popular GUIs: 'Rgui', 'RStudio', 'Positron',
        'VSCode', 'Jupyter', 'Emacs', and 'Rscript' (shell). Compatible
        with several functions and packages: 'source()',
        'sys.source()', 'debugSource()' in 'RStudio',
        'compiler::loadcmp()', 'utils::Sweave()', 'box::use()',
        'knitr::knit()', 'plumber::plumb()', 'shiny::runApp()',
        'package:targets', and 'testthat::source_file()'.
Authors@R: person(given = "Iris",
                  family = "Simmons",
                  role = c("aut", "cre"),
                  email = "ikwsimmo@gmail.com")
Author: Iris Simmons [aut, cre]
Maintainer: Iris Simmons <ikwsimmo@gmail.com>
Depends: R (>= 2.15)
Suggests: utils, jsonlite, microbenchmark, rprojroot
Enhances: compiler, box, knitr, plumber, shiny, targets, testthat
URL: https://github.com/ArcadeAntics/this.path
BugReports: https://github.com/ArcadeAntics/this.path/issues
ByteCompile: TRUE
Biarch: TRUE
BuildManual: TRUE
Type: Package
NeedsCompilation: yes
Packaged: 2026-04-09 03:13:52 UTC; iris
Repository: CRAN
Date/Publication: 2026-04-10 08:50:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 11:03:41 UTC; unix
Archs: this.path.so.dSYM

##
## this.path : Get Executing Script's Path
## Copyright (C) 2021-2025   Iris Simmons
##


useDynLib(this.path, .registration = TRUE, .fixes = ".C_")


## backports.R     ----


if (getRversion() < "3.0.0") {
    export(.mapply)
    export(list.files)
    export(parse)
}


if (getRversion() < "3.1.0") {
    export(anyNA)
    export(anyNA.data.frame)
    export(anyNA.numeric_version)
    export(anyNA.POSIXlt)
    S3method(anyNA, data.frame)
    S3method(anyNA, numeric_version)
    S3method(anyNA, POSIXlt)
    S3method(anyNA, default, .anyNA.default)
}


if (getRversion() < "3.2.0") {
    export(isNamespaceLoaded)
    export(dir.exists)
    export(file.copy)
    export(lengths)
    S3method(lengths, default, .lengths.default)
    export(file.info)
    export(file.mtime)
    export(file.size)
}


if (getRversion() < "3.3.0") {
    export(strrep)
    export(startsWith)
    export(endsWith)
}


if (getRversion() < "3.4.0") {
    export(print.connection)
    S3method(print, connection)
    export(withAutoprint)
}


if (getRversion() < "3.5.0") {
    export(...length)
    export(isTRUE)
    export(isFALSE)
}


if (getRversion() < "3.6.0") {
    export(errorCondition)
    export(str2expression)
    export(str2lang)
    export(Sys.setFileTime)
}


if (getRversion() < "4.0.0") {
    export(deparse1)
    export(unlink)
}


if (getRversion() < "4.1.0") {
    export(...elt)
    export(bquote)
}


if (getRversion() < "4.2.0") {
    export(gettext)
    export(gettextf)
}


## checkpath.R     ----


export(check.path, check.dir , check.proj)


## defunct.R       ----


## enhances.R      ----


export(with_script_path, make_fix_file   )


## error.R         ----


export(thisPathNotExistsError, thisPathNotFoundError , tryCatch2             ,
       last.condition        , tryCatch3             )


## files.R         ----


export(basename2   , dirname2    , splitext    , removeext   ,
       ext         , "ext<-"     , path.join   , path.split  ,
       path.split.1, path.unsplit)


## ismain.R        ----


export(is.main   , from.shell)


## lineno.R        ----


export(sys.LINENO, env.LINENO, src.LINENO, LINENO    , LINE      )


## make_fix_funs.R ----


export(path.functions, make_fix_funs)


## ns-hooks.R      ----


export(OS.type)


## print.R         ----


S3method(print       , ThisPathDocumentContext)
S3method(format      , ThisPathDocumentContext)
S3method(as.character, ThisPathDocumentContext)


## progargs.R      ----


export(asArgs  , progArgs, withArgs)


## promises.R      ----


export(initwd   , getinitwd)


## relpath.R       ----


export(relpath     , rel2sys.dir , rel2sys.proj, rel2env.dir ,
       rel2env.proj, rel2src.dir , rel2src.proj, rel2here    ,
       rel2proj    )


## rprojroot.R     ----


export(sys.proj  , env.proj  , src.proj  , this.proj , reset.proj)


## setsyspath.R    ----


export(wrap.source          , set.sys.path         , unset.sys.path       ,
       set.env.path         , set.src.path         , set.sys.path.function,
       with_sys.path        )


## shfile.R        ----


export(shFILE    , try.shFILE)


## startup.R       ----


export(site.file     , init.file     , with_site.file, with_init.file)


## sys.R           ----


export(sys.srcref , sys.whiches)


## sysputenv.R     ----


export(Sys.putenv)


## thispath.R      ----


export(set.jupyter.path, set.gui.path    , sys.path        ,
       sys.dir         , env.path        , env.dir         ,
       src.path        , src.dir         , this.path       ,
       this.dir        , sys.here        , env.here        ,
       src.here        , here            , ici             ,
       try.sys.path    , try.env.path    , try.src.path    ,
       try.this.path   , FILE            )


## utils.R         ----


export(source.exprs)

this.path : Get Executing Script's Path
Copyright (C) 2022-2024   Iris Simmons


In Rgui for Windows, this.path() works by looking at the windows
handles belonging to that R process.

For an open script, the windows handle will end with " - R Editor" OR
ANY SUPPORTED TRANSLATION INTO ANOTHER LANGUAGE. For a new script, the
windows handle will be "Untitled - R Editor", again possibly translated
into another language.

What we do in "main.R" is:

  * launch a new Rgui process in each supported language and each
    supported version of R

  * open an existing script and grab the translation of " - R Editor"

  * open a new script and grab the translation of "Untitled - R Editor"

We will then save that information into files
"r-editor_msvcrt_2_15.dat", ..., "r-editor_ucrt_4_2.dat", ...,
"untitled_msvcrt_2_15.dat", ..., and "untitled_ucrt_4_2.dat", ...

Some of the files are redundant because the translations did not change
from previous R versions, and in such cases we delete those files.

```
this.path : Get Executing Script's Path
Copyright (C) 2021-2026   Iris Simmons
```


# this.path

  * [Description](#description)
  * [Installation](#installation)
  * [Details](#details)
  * [Note](#note)
  * [Alternatives](#alternatives)
     1. [Always Change Working Directory](#alternative-1-always-change-working-directory)
     2. [Source References](#alternative-2-source-references)
     3. [Absolute Paths](#alternative-3-absolute-paths)
     4. [Other Packages That Determine Current **R** Script](#alternative-4-other-packages-that-determine-current-r-script)
     5. [Packages That Determine Project Root](#alternative-5-packages-that-determine-project-root)
     6. [`package:box`](#alternative-6-packagebox)
  * [Closing](#closing)

## Description

`package:this.path` provides a mechanism for an **R** script to
retrieve its own path without the explicit need to write it elsewhere.

## Installation

Install it from CRAN:
```R
install.packages("this.path")
```

Install the development version from GitHub:
```R
install.packages(
    "this.path",
    repos = "https://raw.githubusercontent.com/ArcadeAntics/PACKAGES"
)

## or:

remotes::install_github("ArcadeAntics/this.path")
```

## Details

The most important functions from `package:this.path` are
`this.path()`, `this.dir()`, `here()`, and `this.proj()`:

*   `this.path()` returns the normalized path of the script in which it
    is written.

*   `this.dir()` returns the directory of `this.path()`.

*   `here()` constructs file paths against `this.dir()`.

*   `this.proj()` constructs file paths against the project root of
    `this.dir()`.

New additions include:

*   `LINENO()` returns the line number of the executing expression.

*   `shFILE()` looks through the command line arguments, extracting
    *FILE* from either of the following: `-f FILE` or `--file=FILE`

*   `set.sys.path()` implements `this.path()` for any `source()`-like
    functions outside of the builtins.

*   `with_init.file()` allows `this.path()` to be used in a user
    profile.

`package:this.path` also provides functions for constructing and
manipulating file paths:

*   `path.join()`, `basename2()`, and `dirname2()` are drop in
    replacements for `file.path()`, `basename()`, and `dirname()` which
    better handle drives and network shares.

*   `splitext()`, `removeext()`, `ext()`, and `ext<-()` split a path
    into root and extension, remove a file extension, get an extension,
    or set an extension for a file path.

*   `path.split()`, `path.split.1()`, and `path.unsplit()` split the
    path to a file into components.

*   `relpath()`, `rel2here()`, and `rel2proj()` turn absolute paths
    into relative paths.

## Note

This package started from a stack overflow posting:

[https://stackoverflow.com/questions/1815606/determine-path-of-the-executing-script/64129649#64129649](https://stackoverflow.com/questions/1815606/determine-path-of-the-executing-script/64129649#64129649)

If you like this package, please consider upvoting my answer so that
more people will see it! If you have an issue with this package, please
use `bug.report(package = "this.path")` to report your issue.

## Alternatives

If you are unhappy with the performance of `package:this.path`, or
would like to know other solutions, here are some alternatives:

### Alternative 1: Always Change Working Directory

The working directory could always be changed to the directory of the
executing script before running it. This would be:

```bash
cd /path/to
Rscript ./file.R
```

or:

```R
source("/path/to/file.R", chdir = TRUE)
```

This does not determine the executing script, only the executing
script's directory, so it does not provide all the same functionality
of `package:this.path`. Additionally, it fails for executable **R**
scripts and it ignores the simple fact that it is sometimes
legitimately desirable to have the working directory set elsewhere.

### Alternative 2: Source References

`utils::getSrcFilename()` provides the ability to retrieve the filename
of a source reference. Everywhere `this.path()` would be used, replace
it with `utils::getSrcFilename(function() NULL, full.names = TRUE)`.
This comes with some issues such as:

*   fails in interactive use since scripts must be run with `source()`

*   fails when running **R** scripts from a shell

*   option `keep.source` must be set to `TRUE`

*   it returns the non-normalized source file's `filename`

### Alternative 3: Absolute Paths

Instead of changing the working directory to the directory of the
executing script and referring to files with relative file paths, refer
to files with absolute file paths. This would be:

*   `"C:/path/to/file"` or `"//host/share/path/to/file"` on Windows

*   `"/path/to/file"` under Unix-alikes

This has some undesirable consequences:

*   If the files are moved to a new location, or any of the file path
    components are renamed, the file paths need to be updated for every
    use for every script.

*   If the files are hosted on a network share, users with differing
    network drive mappings or even differing operating systems may not
    have a common absolute file path to refer to the same location.

This does not determine the executing script, nor the executing
script's directory. It provides none of the functionality of
`package:this.path`.

### Alternative 4: Other Packages That Determine Current **R** Script

There are a few other packages and functions that provide the ability
to retrieve the path of the current **R** script:

*   [`package:envDocument`](https://CRAN.R-project.org/package=envDocument)
    specifically `envDocument::getScriptPath()`

*   [`package:funr`](https://CRAN.R-project.org/package=funr)
    specifically `funr::get_script_path()` and `funr::sys.script()`

*   [`package:gtools`](https://CRAN.R-project.org/package=gtools)
    specifically `gtools::script_file()`

*   [`package:scriptName`](https://CRAN.R-project.org/package=scriptName)
    specifically `scriptName::current_filename()`

*   [`package:whereami`](https://CRAN.R-project.org/package=whereami)
    specifically `whereami::thisfile()`

These are lacking in functionality compared to `package:this.path`:

*   `this.path()` is compatible with the following GUIs:

    *   'Rgui'

    *   '[RStudio](https://posit.co/products/open-source/rstudio/)'
        (including
        [background jobs](https://docs.posit.co/ide/user/ide/guide/tools/jobs.html))

    *   '[Positron](https://github.com/posit-dev/positron)'

    *   '[VSCode](https://code.visualstudio.com/)' +
        '[REditorSupport](https://code.visualstudio.com/docs/languages/r)'

    *   '[Jupyter](https://jupyter.org/)'

    *   '[Emacs](https://www.gnu.org/software/emacs/)' +
        '[ESS](https://ess.r-project.org/)'

*   `this.path()` is compatible with the following functions and
    packages:

    *   `sys.source()`

    *   `debugSource()` in 'RStudio'

    *   `compiler::loadcmp()`

    *   `utils::Sweave()`

    *   [`box::use()`](https://CRAN.R-project.org/package=box)

    *   [`knitr::knit()`](https://CRAN.R-project.org/package=knitr)

    *   [`plumber::plumb()`](https://CRAN.R-project.org/package=plumber)

    *   [`shiny::runApp()`](https://CRAN.R-project.org/package=shiny)

    *   [`package:targets`](https://CRAN.R-project.org/package=targets)

    *   [`testthat::source_file()`](https://CRAN.R-project.org/package=testthat)

*   Related to `source()`, `this.path()`:

    *   takes argument `chdir` into account.

    *   recognizes that `file = ""`, `file = "clipboard"`, and
        `file = "stdin"` are not referring to files and skips them.

    *   accounts for `file` being a URL pathname.

    *   accounts for `file` being a connection instead of a character
        string.

    *   skips calls in which `file` is missing and `exprs` is used
        instead.

*   `this.path()` does a better job of extracting *FILE* from the
    command line arguments `-f FILE` and `--file=FILE`.

*   `this.path()` saves all normalized paths within their appropriate
    environments, making it much faster subsequent times within the
    same script and independent of working directory.

*   If `this.path()` does not find an executing script, it throws an
    error. This is better than returning `NULL`; if the executing
    script cannot be found, obviously there is a problem that needs to
    be addressed, so the script should not continue.

These packages also have some objectively incorrect coding issues such
as:

*   searching only the first call on the stack

*   failing to check the command line arguments

*   imposing unnecessary restrictions on pathnames

*   searching the call stack in the wrong order

*   returning the directory instead of the path of the executing script

*   searching for a path in the wrong order

*   searching for a source call by frame variables instead of function
    value

*   loading unnecessary additional packages without explicit user
    permission

*   searching for a source call by name instead of value

*   returning a non-normalized path

*   changing global options without explicit user permission

### Alternative 5: Packages That Determine Project Root

[`package:here`](https://CRAN.R-project.org/package=here) provides
function `here::here()` with the ability to retrieve the project root,
but it lacks functionality:

*   `here::here()` returns the initial working directory when it cannot
    find the project's directory. If the project's directory cannot be
    found, there is something wrong and an error should be thrown, but
    it doesn't, and this leads to incorrect / / unexpected behaviour.

*   `here::here()` does not work when the initial working directory is
    set outside the context of the project. Occasionally, I would set
    the working directory to a temporary directory where my **R**
    script would create a log file of the details of the script. This
    leads to `package:here` not being able to find the project root and
    incorrectly returning the initial working directory.

*   `here::here()` does not work when multiple projects are in use nor
    for projects containing sub-projects. In my scenario, I had a
    project `A` and a project `B`. I would run a script in `B` which
    runs a script in `A`, but the project root is already set to the
    root of `B`, so the script in `A` fails.

[`package:rprojroot`](https://CRAN.R-project.org/package=rprojroot) is
the **R** package upon which `package:here` is built. It provides a
list of project root criteria `rprojroot::criteria`. It is not useful
on its own since, like `here::here()`, it does not work when the
working directory is set outside the context of the project. However,
you can combine it with `package:this.path` to get the best results:

```R
## replace 'rprojroot::is_r_package' with desired criterion

fix_file <- this.path::make_fix_file(rprojroot::is_r_package)
```

If the default criteria are not sufficient for your needs, you can make
your own using:

*   `rprojroot::has_basename()`

*   `rprojroot::has_dir()`

*   `rprojroot::has_file()`

*   `rprojroot::has_file_pattern()`

*   `rprojroot::root_criterion()`

`package:rprojroot` recommends using
`<criterion>$find_file(path = whereami::thisfile())` for this purpose,
but as mentioned in section **Alternative 4**, `whereami::thisfile()`
is seriously lacking compared to `this.path::this.path()`.

These can not determine the executing script, nor the executing
script's directory. They provide none of the same functionality of
`package:this.path`.

### Alternative 6: `package:box`

[`package:box`](https://CRAN.R-project.org/package=box) provides two
related functions:

*   `box::file()` constructs file paths against the directory of the
    executing script.

*   `box::use()` imports an **R** script as a module.

These both lack the same functionality as the packages listed in
**Alternative 4**. `box::file()` should not be used in favour of
`this.path::here()`. However, `box::use()` is still extremely useful,
it just needs to be combined with `package:this.path` to get the best
results:

```R
this.path::with_script_path(
box::use(
    <import 1>,
    <import 2>,
    <...>
)
)
```

This explicitly tells `package:box` the path of the current script so
that relative imports will work correctly.

## Closing

If you think I have overlooked some of the alternatives, or think there
are any improvements I could make to `package:this.path`, please let me
know, I am open to suggestions! I hope this package serves you well!

##
## this.path : Get Executing Script's Path
## Copyright (C) 2021-2026   Iris Simmons
##


.package_translations_languages <- function ()
{
    ## the languages supported by R can be listed using this:
    ##
    ## dirname(grep('/LC_MESSAGES$',list.dirs(system.file(package='translations'),full.names=FALSE),perl=TRUE,value=TRUE))
    ##
    ## or, written more plainly, using this:
    x <- system.file(package = "translations")
    x <- list.dirs(x, full.names = FALSE)
    x <- grep("/LC_MESSAGES$", x, perl = TRUE, value = TRUE)
    dirname(x)
}


.languages <- matrix(
    ncol = 4L,
    byrow = TRUE,
    dimnames = list(NULL, c(
"LANGUAGE", "Format: (* Custom Locale)"    , "locale.codepage"                                , "locale.utf8"                      )), c(
"ar"      , "Arabic (Saudi Arabia)"        , "Arabic_Saudi Arabia.1256"                       , "Arabic_Saudi Arabia.utf8"         ,
"bn"      , "Bengali (India)"              , "Bengali_India.1252"                             , "Bengali_India.utf8"               ,
"ca"      , "Catalan (Catalan)"            , "Catalan_Spain.1252"                             , "Catalan_Spain.utf8"               ,
"da"      , "Danish (Denmark)"             , "Danish_Denmark.1252"                            , "Danish_Denmark.utf8"              ,
"de"      , "German (Germany)"             , "German_Germany.1252"                            , "German_Germany.utf8"              ,
"en"      , "English (World)"              , "English_World.1252"                             , "English_World.utf8"               ,
"en@quot" , "English (World)"              , "English_World.1252"                             , "English_World.utf8"               ,
"en_GB"   , "English (United Kingdom)"     , "English_United Kingdom.1252"                    , "English_United Kingdom.utf8"      ,
"es"      , "Spanish (Spain)"              , "Spanish_Spain.1252"                             , "Spanish_Spain.utf8"               ,
"fa"      , "Persian (Iran)"               , "Persian_Iran.1256"                              , "Persian_Iran.utf8"                ,
"fr"      , "French (France)"              , "French_France.1252"                             , "French_France.utf8"               ,
"hi"      , "Hindi (India)"                , "Hindi_India.1252"                               , "Hindi_India.utf8"                 ,
"hu"      , "Hungarian (Hungary)"          , "Hungarian_Hungary.1250"                         , "Hungarian_Hungary.utf8"           ,
"id"      , "Indonesian (Indonesia)"       , "Indonesian_Indonesia.1252"                      , "Indonesian_Indonesia.utf8"        ,
"it"      , "Italian (Italy)"              , "Italian_Italy.1252"                             , "Italian_Italy.utf8"               ,
"ja"      , "Japanese (Japan)"             , "Japanese_Japan.932"                             , "Japanese_Japan.utf8"              ,
"ko"      , "Korean (Korea)"               , "Korean_Korea.949"                               , "Korean_Korea.utf8"                ,
"lt"      , "Lithuanian (Lithuania)"       , "Lithuanian_Lithuania.1257"                      , "Lithuanian_Lithuania.utf8"        ,
"ne"      , "Nepali (Nepal)"               , "Nepali_Nepal.1252"                              , "Nepali_Nepal.utf8"                ,
"nn"      , "Norwegian Nynorsk (Norway)"   , "Norwegian-Nynorsk_Norway.1252"                  , "Norwegian Nynorsk_Norway.utf8"    ,
"pl"      , "Polish (Poland)"              , "Polish_Poland.1250"                             , "Polish_Poland.utf8"               ,
"pt_BR"   , "Portuguese (Brazil)"          , "Portuguese_Brazil.1252"                         , "Portuguese_Brazil.utf8"           ,
"ru"      , "Russian (Russia)"             , "Russian_Russia.1251"                            , "Russian_Russia.utf8"              ,
"sq"      , "Albanian (Albania)"           , "Albanian_Albania.1250"                          , "Albanian_Albania.utf8"            ,
"tr"      , "Turkish (T\u00FCrkiye)"       , `Encoding<-`("Turkish_T\xFCrkiye.1254", "latin1"), "Turkish_T\u00FCrkiye.utf8"        ,
"ur"      , "Urdu (Pakistan)"              , "Urdu_Pakistan.1256"                             , "Urdu_Pakistan.utf8"               ,
"zh_CN"   , "Chinese (Simplified, China)"  , "Chinese (Simplified)_China.936"                 , "Chinese (Simplified)_China.utf8"  ,
"zh_TW"   , "Chinese (Traditional, Taiwan)", "Chinese (Traditional)_Taiwan.950"               , "Chinese (Traditional)_Taiwan.utf8"
    )
)
rownames(.languages) <- .languages[, "LANGUAGE"]


.my_unsupported_languages <- function ()
setdiff(.package_translations_languages(), rownames(.languages))


.language_envvars_list <- sapply(
    rownames(.languages),
    function(language) {
        sapply(
            c("codepage", "utf8"),
            function(x) {
                locale <- .languages[[language, paste0("locale.", x)]]
                c(
                    paste0("LANGUAGE=", language),
                    `Encoding<-`(paste0("LC_ALL=", `Encoding<-`(locale, "bytes")), Encoding(locale))
                )
            },
            simplify = FALSE
        )
    },
    simplify = FALSE
)


.language_envvars <- function (LANGUAGE = Sys.getenv("LANGUAGE"), utf8 = identical(R.version[["crt"]], "ucrt"))
{
    if (!is.character(LANGUAGE) || length(LANGUAGE) != 1L)
        stop(gettextf("'%s' must be a character string",
            "LANGUAGE", domain = "R"), domain = NA)
    if (.Platform$OS.type == "windows") {
        if (!nzchar(LANGUAGE))
            return(c("LANGUAGE=", "LC_ALL="))
        LANGUAGE <- match.arg(LANGUAGE, c(rownames(.languages), NA))
        if (is.na(LANGUAGE))
            c("LANGUAGE=", "LC_ALL=")
        else .language_envvars_list[[LANGUAGE]][[if (utf8) "utf8" else "codepage"]]
    } else {
        if (!nzchar(LANGUAGE))
            return("LANGUAGE=")
        LANGUAGE <- match.arg(LANGUAGE, c(rownames(.languages), NA))
        if (is.na(LANGUAGE))
            "LANGUAGE="
        else paste0("LANGUAGE=", LANGUAGE)
    }
}


Sys.putenv <- function (x)
{
    x <- as.character(x)
    m <- regexpr("=", x, fixed = TRUE, useBytes = TRUE)
    y <- regmatches(x, m, invert = TRUE)
    if (any(invalid <- lengths(y) != 2L)) {
        stop(
            ngettext(
                sum(invalid),
                "invalid environment variable:\n",
                "invalid environment variables:\n"
            ),
            paste(utils::capture.output(x[invalid]), collapse = "\n")
        )
    }
    y <- .mapply(`Encoding<-`, list(y, Encoding(x)), NULL)
    args <- lapply(y, `[[`, 2L)
    names(args) <- vapply(y, `[[`, "", 1L, USE.NAMES = FALSE)
    do.call("Sys.setenv", args)
}


.read_C_strings <- function (path, divisor = 1L)
{
    x <- readBin(path, "raw", file.size(path))
    if (!length(x))
        return(character())
    if (x[[length(x)]] != as.raw(0L))
        stop(sprintf("incomplete final string found on '%s'", path))
    if (length(nul <- which(x == as.raw(0L))) %% divisor)
        stop(sprintf("expected a multiple of %s strings on '%s'", divisor, path))
    to <- nul - 1L
    length.out <- diff(c(0L, nul)) - 1L
    x <- lapply(seq_along(nul), function(i) {
        rawToChar(x[seq.int(to = to[[i]], length.out = length.out[[i]])])
    })
    as.character(x)
}


.read_C_strings_with_encoding <- function (path, divisor = 1L)
{
    x <- .read_C_strings(path, divisor * 2L)
    if (length(x)) {
        encoding <- x[c(FALSE, TRUE)]
        x        <- x[c(TRUE, FALSE)]
        Encoding(x) <- encoding
    }
    x
}


if (sys.nframe() == 0L && !nzchar(Sys.getenv("R_PACKAGE_NAME"))) {
    stopifnot(.Platform$OS.type == "windows")


    main <- function() {


        testing <- FALSE
        # testing <- TRUE; warning("comment out 'testing <- TRUE' later", immediate. = TRUE)
        quiet_on_success <- testing


        tmpdir <- tempdir()
        if (!all(charToRaw(tmpdir) <= 0x7f)) {
            ## if the temporary directory does not consistent purely of ASCII
            ## characters, use a different directory
            tmpdirdir <- sprintf("%s\\tmp", Sys.getenv("SystemDrive"))
            if (!dir.exists(tmpdirdir)) {
                dir.create(tmpdirdir)
                on.exit(unlink(tmpdirdir, recursive = TRUE, force = TRUE, expand = FALSE), add = TRUE, after = FALSE)
            }
            tmpdir <- tempfile("Rtmp", tmpdirdir)
            tryCatch(dir.create(tmpdir),
            ## turn a warning into an error
            warning = stop
            )
            on.exit(unlink(tmpdir, recursive = TRUE, force = TRUE, expand = FALSE), add = TRUE, after = FALSE)
        }


        ## there was a time when I was doing something along the lines:
        ##
        ## ```
        ## for (language in rownames(.languages)) {
        ##     Sys.putenv(.language_envvars(language))
        ##     gettext("Untitled", domain = "RGui")
        ##     gettext("R Editor", domain = "RGui")
        ## }
        ## ```
        ##
        ## but it doesn't work well for Microsoft Visual C++ Runtime
        ## because the messages usually get mistranslated in the windows titles
        ## but they don't with gettext()
        ##
        ## we want the exact windows titles, mistranslated or otherwise
        ##
        ## and it also doesn't work well for Universal C Runtime in Lithuanian


        stopifnot(.Platform$OS.type == "windows")
        stopifnot(bindtextdomain("RGui") != "")
        stopifnot(identical(R.version[["crt"]], "ucrt"))
        stopifnot(getRversion() >= "4.2.0")
        stopifnot(l10n_info()[["UTF-8"]])


        exe <- "Rscript.exe"
        x <- Sys.getenv()
        path <- c(
            grep("(?i)^r_release$", names(x)),
            grep("(?i)^r_oldrel$", names(x)),
            grep("(?i)^r_[[:digit:]]+_[[:digit:]]+$", names(x))
        )
        path <- unique(path)
        path <- as.character(x[path])
        if (length(path)) {
            path <- file.path(path, exe)
            path <- path[file.access(path, 1L) == 0L]
            path <- c(file.path(R.home("bin"), exe), path)
        } else {
            path <- file.path(R.home(), "..")
            path <- normalizePath(path, "/", TRUE)
            path <- list.files(path, full.names = TRUE)
            path <- file.path(path, "bin", exe)
            path <- path[file.access(path, 1L) == 0L]
            path <- c(file.path(R.home("bin"), exe), path)
            path <- utils::shortPathName(path)
        }


        # args <- c("--version")
        # args <- c("--default-packages=NULL", "--vanilla", "-e", "utils::capture.output")
        # args <- paste(shQuote(args), collapse = " ")
        # command <- paste(shQuote(path), args)
        # names(command) <- path
        # x <- lapply(command, system, intern = TRUE)
        # print(x, quote = FALSE, width = 10)
        # stop("comment this out later")


        deparse_strings <- function(x) {
            if (!is.character(x))
                x <- as.character(x)
            vapply(
                x,
                function(x) {
                    if (nzchar(x))
                        sprintf(
                            "`Encoding<-`(rawToChar(as.raw(c(%s))), \"%s\")",
                            paste0(as.integer(charToRaw(x)), "L", collapse = ", "),
                            Encoding(x)
                        )
                    else "\"\""
                },
                ""
            )
        }


        get_R_info <- function(path) {
            tmpfiles <- tempfile(tmpdir = tmpdir, fileext = c(".R", ""))
            on.exit(unlink(tmpfiles), add = TRUE, after = FALSE)
            R_PROFILE_USER <- Sys.getenv("R_PROFILE_USER", NA)
            if (is.na(R_PROFILE_USER)) {
                on.exit(Sys.unsetenv("R_PROFILE_USER"), add = TRUE, after = FALSE)
            } else {
                on.exit(Sys.setenv(R_PROFILE_USER = R_PROFILE_USER), add = TRUE, after = FALSE)
            }
            Sys.setenv(R_PROFILE_USER = tmpfiles[[1L]])
            writeLines(
                sprintf(
                    '
                        .First <- function() {
                            conn <- file(%s, "ab", encoding = "")
                            on.exit(close(conn))
                            nul <- as.raw(0L)
                            writeBin(
                                c(
                                    charToRaw(if (identical(R.version[["crt"]], "ucrt")) "TRUE" else "FALSE"),
                                    nul,
                                    charToRaw(as.character(getRversion())),
                                    nul,
                                    charToRaw(R.home("bin")),
                                    nul,
                                    charToRaw(.Platform$r_arch),
                                    nul
                                ),
                                conn
                            )
                        }
                    ',
                    deparse_strings(tmpfiles[[2L]])
                ),
                tmpfiles[[1L]]
            )
            file.create(tmpfiles[[2L]])
            expected <- 4L
            n <- 0L
            x <- character()
            lapply(path, function(path) {
                args <- c(
                    path,
                    "--default-packages=NULL",
                    "--no-save",
                    "--no-restore",
                    "--no-site-file",
                    "--no-environ",
                    "--no-Rconsole",
                    "-e",
                    "NULL"
                )
                command <- paste(shQuote(args), collapse = " ")
                system(command, intern = TRUE)
                n <<- n + expected
                x <<- .read_C_strings(tmpfiles[[2L]])
                stopifnot(length(x) == n)
            })
            ## break into groups of 'expected'
            x <- split(x, seq.int(0L, along.with = x) %/% expected)
            x <- data.frame(
                ucrt      =       as.logical(vapply(x, `[[`, "", 1L)),
                version   = R_system_version(vapply(x, `[[`, "", 2L)),
                R_BIN_DIR =                  vapply(x, `[[`, "", 3L),
                r_arch    =                  vapply(x, `[[`, "", 4L)
            )
            x$major <- vapply(unclass(x$version), `[[`, 0L, 1L)
            x$minor <- vapply(unclass(x$version), `[[`, 0L, 2L)
            x$patch <- vapply(unclass(x$version), `[[`, 0L, 3L)
            x
        }
        x <- get_R_info(path)
        x$rgui <- file.path(x$R_BIN_DIR, "Rgui.exe")
        x <- x[x$version >= "2.15.0", , drop = FALSE]
        x <- x[order(x$version, decreasing = TRUE), , drop = FALSE]
        proto <- x[integer(), , drop = FALSE]
        x <- split(x, interaction(x$major, x$minor, drop = TRUE, lex.order = TRUE))
        x <- lapply(x, function(x) {
            if (nrow(x) == 1L)
                x
            else {
                x <- x[x[["version"]] == x[[1L, "version"]], , drop = FALSE]
                if (nrow(x) == 1L)
                    x
                else {
                    if (length(i <- which(x$r_arch == "x64")))
                        x[i[[1L]], , drop = FALSE]
                    else if (length(i <- which(x$r_arch != "i386")))
                        x[i[[1L]], , drop = FALSE]
                    else x[1L, , drop = FALSE]
                }
            }
        })
        x <- do.call("rbind", x)
        x <- rbind(proto, x)
        x <- x[order(x$version), , drop = FALSE]
        x$suffix <- paste(ifelse(x$ucrt, "ucrt", "msvcrt"), x$major, x$minor, sep = "_", recycle0 = TRUE)
        x$suffixes <- I(vector("list", nrow(x)))
        invisible(lapply(
            split(seq_len(nrow(x)), x$ucrt),
            function(i) {
                x$suffixes[i] <<- lapply(
                    seq_along(i),
                    function(j) rev(x[i[seq_len(j)], "suffix"])
                )
            }
        ))
        x$suffix <- NULL


        write_r_editor <- function(rgui, ucrt, suffixes) {


            # i <- 1L; rgui <- x[[i, "rgui"]]; ucrt <- x[[i, "ucrt"]]; suffixes <- x[[i, "suffixes"]]; warning("comment this out later", immediate. = TRUE)


            FILES <- tempfile(tmpdir = tmpdir, fileext = c(".dat", ".R", ".Rprofile", ".Renviron"))
            on.exit(unlink(FILES), add = TRUE, after = FALSE)
            FILE.dat <- FILES[[1L]]
            FILE.R   <- FILES[[2L]]
            Rprofile <- FILES[[3L]]
            Renviron <- FILES[[4L]]
            file.create(FILE.dat)
            file.create(FILE.R)


            local({
                conn <- file(Rprofile, "wb", encoding = "")
                on.exit(close(conn), add = TRUE, after = FALSE)
                writeLines(paste0("FILE.dat <- ", deparse_strings(FILE.dat)), conn, useBytes = TRUE)
                writeLines(paste0("FILE.R   <- ", deparse_strings(FILE.R  )), conn, useBytes = TRUE)
                writeLines('
                    .First <- function() {
                        options(error = function() {
                            quit(save = "no", status = 1L)
                        })
                        reg.finalizer(environment(), function(e) {
                            conn <- e$conn
                            if (!is.null(conn) && isOpen(conn))
                                close(conn)
                        }, onexit = TRUE)
                        fun <- function() {
                            text <- names(utils::getWindowsHandles())[[1L]]
                            if (Encoding(text) == "unknown") {
                                loc <- l10n_info()
                                Encoding(text) <- if (loc$`UTF-8`)
                                    "UTF-8"
                                else if (loc$`Latin-1`)
                                    "latin1"
                                else "unknown"
                            }
                            nul <- as.raw(0L)
                            writeBin(
                                c(
                                    charToRaw(text),
                                    nul,
                                    charToRaw(Encoding(text)),
                                    nul
                                ),
                                conn
                            )
                        }
                        conn <- file(FILE.dat, "ab", encoding = "")
                        utils::file.edit(FILE.R)
                        fun()
                        utils::file.edit("")
                        fun()
                        quit(save = "no")
                    }
                ', conn, useBytes = TRUE)
            })


            make_Renviron <- function(language, ucrt) {
                x <- .language_envvars(language, ucrt)
                x <- c(x, "R_DEFAULT_PACKAGES=NULL")
                conn <- file(Renviron, "wb", encoding = "")
                on.exit(close(conn), add = TRUE, after = FALSE)
                writeLines(x, conn, useBytes = TRUE)
                invisible()
            }


            R_PROFILE_USER <- Sys.getenv("R_PROFILE_USER", NA)
            if (is.na(R_PROFILE_USER)) {
                on.exit(Sys.unsetenv("R_PROFILE_USER"), add = TRUE, after = FALSE)
            } else {
                on.exit(Sys.setenv(R_PROFILE_USER = R_PROFILE_USER), add = TRUE, after = FALSE)
            }
            Sys.setenv(R_PROFILE_USER = Rprofile)


            R_ENVIRON_USER <- Sys.getenv("R_ENVIRON_USER", NA)
            if (is.na(R_ENVIRON_USER)) {
                on.exit(Sys.unsetenv("R_ENVIRON_USER"), add = TRUE, after = FALSE)
            } else {
                on.exit(Sys.setenv(R_ENVIRON_USER = R_ENVIRON_USER), add = TRUE, after = FALSE)
            }
            Sys.setenv(R_ENVIRON_USER = Renviron)


            ## we want to provide --vanilla to enable factory-default settings
            ## for Rgui.exe
            ##
            ## --vanilla is a combination of --no-save, --no-restore,
            ##           --no-site-file, --no-init-file, --no-environ,
            ##           and --no-Rconsole
            ##
            ## however, we want to run the environ and init file, so instead of
            ## --vanilla, use the other arguments except --no-environ and
            ## --no-init-file
            options <- c("--no-save", "--no-restore", "--no-site-file", "--no-Rconsole")


            n <- 0L
            args <- c(rgui, options)
            command <- paste(shQuote(args), collapse = " ")
            for (language in rownames(.languages)) {
                make_Renviron(language, ucrt)
                ans <- system(command)
                if (ans) {
                    if (ans == -1L) {
                        stop(gettextf("'%s' could not be run",
                            command,      domain = "R-base"), domain = NA)
                    } else {
                        stop(gettextf("'%s' execution failed with error code %d",
                            command, ans, domain = "R-base"), domain = NA)
                    }
                }
                n <- n + 4L
                stopifnot(length(.read_C_strings(FILE.dat)) == n)
            }


            txt <- .read_C_strings_with_encoding(FILE.dat, divisor = 2L)


            r_editor <- txt[c(TRUE, FALSE)]
            untitled <- txt[c(FALSE, TRUE)]


            dir <- "./inst/extdata"
            valid_dir <- endsWith(
                tryCatch(normalizePath(dir, "/", TRUE), error = function(e) ""),
                "/this.path/inst/extdata"
            )
            r_editor_paths <- sprintf("%s/r-editor_%s.dat", dir, suffixes)
            untitled_paths <- sprintf("%s/untitled_%s.dat", dir, suffixes)


            if (testing) {


                first <- TRUE
                print_R_version <- function() {
                    if (first) {
                        first <<- FALSE
                        cat("\nin ", rgui, "\n\n", sep = "")
                    }
                }


                if (!valid_dir || is.na(r_editor_path <- r_editor_paths[match(TRUE, file.exists(r_editor_paths))])) {
                    print_R_version()
                    warning(
                        "\n no file to compare \" - R Editor\" strings against! Window titles:\n\n",
                        paste(unique(r_editor), collapse = "\n"),
                        call. = FALSE,
                        immediate. = TRUE,
                        domain = NA
                    )
                } else if (
                    any(invalid <- !(
                        vapply(
                            r_editor,
                            function(str) {
                                paste(charToRaw(str), collapse = "")
                            },
                            ""
                        ) %in% vapply(
                            .read_C_strings_with_encoding(r_editor_path),
                            function(str) {
                                paste(c(charToRaw(FILE.R), charToRaw(str)), collapse = "")
                            },
                            ""
                        )
                    ))
                ) {
                    print_R_version()
                    warning(
                        ngettext(
                            sum(invalid),
                            "invalid \" - R Editor\" string:\n",
                            "invalid \" - R Editor\" strings:\n"
                        ),
                        paste(utils::capture.output(r_editor[invalid]), collapse = "\n"),
                        call. = FALSE,
                        immediate. = TRUE
                    )
                } else if (!quiet_on_success) {
                    print_R_version()
                    cat("\nAll \" - R Editor\" strings are valid!\n", r_editor, sep = "\n")
                }


                if (!valid_dir || is.na(untitled_path <- untitled_paths[match(TRUE, file.exists(untitled_paths))])) {
                    print_R_version()
                    warning(
                        "\n no file to compare \"Untitled - R Editor\" strings against! Window titles:\n\n",
                        paste(unique(untitled), collapse = "\n"),
                        call. = FALSE,
                        immediate. = TRUE,
                        domain = NA
                    )
                } else if (
                    any(invalid <- !(
                        vapply(
                            untitled,
                            function(str) {
                                paste(charToRaw(str), collapse = "")
                            },
                            ""
                        ) %in% vapply(
                            .read_C_strings_with_encoding(untitled_path),
                            function(str) {
                                paste(charToRaw(str), collapse = "")
                            },
                            ""
                        )
                    ))
                ) {
                    print_R_version()
                    warning(
                        ngettext(
                            sum(invalid),
                            "invalid \"Untitled - R Editor\" string:\n",
                            "invalid \"Untitled - R Editor\" strings:\n"
                        ),
                        paste(utils::capture.output(untitled[invalid]), collapse = "\n"),
                        call. = FALSE,
                        immediate. = TRUE
                    )
                } else if (!quiet_on_success) {
                    print_R_version()
                    cat("\nAll \"Untitled - R Editor\" strings are valid!\n", untitled, sep = "\n")
                }


            } else {


                ## string comparisons often involve translating between encodings
                ## so we will do raw comparisons instead to avoid translations
                r_editor <- vapply(r_editor, function(str) {
                    bytes <- charToRaw(str)
                    ## all "R Editor" strings must start with this prefix
                    prefix <- c(charToRaw(FILE.R), charToRaw(" - "))
                    if (length(bytes) < length(prefix) ||
                        any(bytes[seq_along(prefix)] != prefix))
                    {
                        stop("invalid \" - R Editor\" string: ", str, domain = NA)
                    }
                    value <- rawToChar(bytes[seq_along(bytes) > length(prefix) - 3L])
                    Encoding(value) <- Encoding(str)
                    value
                }, "", USE.NAMES = FALSE)


                ## don't use unique() since that converts strings to same encoding,
                ## instead, convert to raw then remove duplicates
                r_editor <- r_editor[!duplicated(lapply(r_editor, charToRaw))]
                untitled <- untitled[!duplicated(lapply(untitled, charToRaw))]


                if (valid_dir) {


                    .write_C_strings <- function(x, path) {
                        conn <- file(path, "wb")
                        on.exit(close(conn), add = TRUE, after = FALSE)
                        lapply(
                            rbind(x, Encoding(x)),
                            function(y) writeBin(c(charToRaw(y), as.raw(0L)), conn)
                        )
                        invisible()
                    }


                    r_editor_path <- r_editor_paths[[1L]]
                    untitled_path <- untitled_paths[[1L]]


                    .write_C_strings(r_editor, r_editor_path)
                    .write_C_strings(untitled, untitled_path)
                    ## if the files are redundant, remove them
                    for (previous in r_editor_paths[-1L]) {
                        if (file.exists(previous)) {
                            if (identical(
                                readBin(r_editor_path, "raw", file.size(r_editor_path)),
                                readBin(previous, "raw", file.size(previous))
                            )) {
                                unlink(r_editor_path)
                            }
                            break
                        }
                    }
                    for (previous in untitled_paths[-1L]) {
                        if (file.exists(previous)) {
                            if (identical(
                                readBin(untitled_path, "raw", file.size(untitled_path)),
                                readBin(previous, "raw", file.size(previous))
                            )) {
                                unlink(untitled_path)
                            }
                            break
                        }
                    }


                } else {


                    warning(
                        "\n no directory in which to write \" - R Editor\" strings:\n\n",
                        paste(r_editor, collapse = "\n"),
                        call. = FALSE,
                        immediate. = TRUE,
                        domain = NA
                    )


                    warning(
                        "\n no directory in which to write \"Untitled - R Editor\" strings:\n\n",
                        paste(untitled, collapse = "\n"),
                        call. = FALSE,
                        immediate. = TRUE,
                        domain = NA
                    )


                }


            }


            invisible(list(r_editor = r_editor, untitled = untitled))
        }


        invisible(.mapply(
            write_r_editor,
            x[c("rgui", "ucrt", "suffixes")],
            NULL
        ))
    }


    main()
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2021-2026   Iris Simmons
##


local({
    owd <- getwd()
    if (is.null(owd)) {
        message("cannot 'chdir' as current directory is unknown")
        return(invisible())
    } else on.exit(setwd(owd), add = TRUE)


    ## test for 3 specific cases of sourcing
    ## * sourcing a file by specifying its basename
    ## * sourcing a file by specifying its absolute path
    ## * sourcing a file by specifying one of its relative paths
    abs_path_R <- tempfile("test", fileext = ".R")
    on.exit(unlink(abs_path_R), add = TRUE)
    this.path:::.writeCode({
        n <- this.path:::.getframenumber()
        if (is.na(n) || n < 1L) stop("invalid traceback")
        sym <- ".this.path::document.context"
        frame <- sys.frame(n)
        stopifnot(bindingIsLocked(sym, frame))
        cat("\n> getwd()\n")
        print(getwd())
        cat("\n> ", paste(deparse(call("dynGet", sym)), collapse = "\n+ "), "\n", sep = "")
        print(frame[[sym]])
        cat("\n> this.path::sys.path(original = TRUE)\n")
        print(this.path::sys.path(original = TRUE))
        cat("\n> this.path::sys.path(for.msg = TRUE)\n")
        print(this.path::sys.path(for.msg = TRUE))
        cat("\n> sys.path(verbose = TRUE)\n")
        stopifnot(identical(
            print(this.path::sys.path(verbose = TRUE)),
            getOption("this.path::sys.path() expectation")
        ))
        cat("\n> this.path::sys.path(original = TRUE)\n")
        print(this.path::sys.path(original = TRUE))
        cat("\n> this.path::sys.path(for.msg = TRUE)\n")
        print(this.path::sys.path(for.msg = TRUE))
    }, file = abs_path_R)
    abs_path_R <- normalizePath(abs_path_R, "/", TRUE)
    abs_path_dir <- normalizePath(R.home(), "/", TRUE)
    basename_R <- this.path::basename2(abs_path_R)
    basename_dir <- this.path::dirname2(abs_path_R)


    rel_path_and_dir <- function(file) {
        x <- this.path::path.split.1(file)
        n <- length(x)
        if (n < 3L) {
            c(this.path::dirname2(file), this.path::basename2(file))
        } else {
            i <- n < seq_len(n) + max(2L, n%/%2L)
            c(this.path::path.unsplit(x[!i]), this.path::path.unsplit(x[i]))
        }
    }
    tmp <- rel_path_and_dir(abs_path_R)
    rel_path_dir <- tmp[[1L]]
    rel_path_R <- tmp[[2L]]
    rm(tmp)


    ## for 'source' and 'debugSource' specifically,
    ## try sourcing a 'file://' URL
    basename_R_URL <- this.path:::.as_file_URL(basename_R)
    rel_path_R_URL <- this.path:::.as_file_URL(rel_path_R)
    abs_path_R_URL <- this.path:::.as_file_URL(abs_path_R)


    fun <- function(expr, envir = parent.frame(),
        bquote_envir = envir, eval_envir = envir)
    {
        if (!is.environment(envir))
            stop("not an environment", domain = "R")
        expr <- call("bquote", substitute(expr), as.symbol("bquote_envir"))
        expr <- eval(expr)
        dep <- deparse(expr)
        cat("\n\n\n\n\n\n\n\n\n\n")
        cat("\n> getwd()\n")
        print(getwd())
        cat("\n> ")
        cat(dep, sep = "\n+ ")
        eval(expr, eval_envir)
    }


    oopt <- options(`this.path::sys.path() expectation` = normalizePath(abs_path_R, "/", TRUE))
    on.exit(options(oopt), add = TRUE)


    ## try using source in all possible manners
    setwd(basename_dir)
    fun(source(.(basename_R)                                 , local = TRUE, chdir = FALSE))               ## from a  basename                    without changing directory
    fun(source(.(basename_R)                                 , local = TRUE, chdir = TRUE ))               ## from a  basename                    with    changing directory (shouldn't do anything)
    fun(source(.(basename_R_URL)                             , local = TRUE               ))               ## from a  basename      'file://' URL
    fun(source(print(conn <- file(open = "r", .(basename_R))), local = TRUE               )); close(conn)  ## from a  basename      connection
    setwd(rel_path_dir)
    fun(source(.(rel_path_R)                                 , local = TRUE, chdir = FALSE))               ## from a  relative path               without changing directory
    fun(source(.(rel_path_R)                                 , local = TRUE, chdir = TRUE ))               ## from a  relative path               with    changing directory
    fun(source(.(rel_path_R_URL)                             , local = TRUE               ))               ## from a  relative path 'file://' URL
    fun(source(print(conn <- file(open = "r", .(rel_path_R))), local = TRUE               )); close(conn)  ## from a  relative path connection
    setwd(abs_path_dir)
    fun(source(.(abs_path_R)                                 , local = TRUE, chdir = FALSE))               ## from an absolute path               without changing directory
    fun(source(.(abs_path_R)                                 , local = TRUE, chdir = TRUE ))               ## from an absolute path               with    changing directory
    fun(source(.(abs_path_R_URL)                             , local = TRUE               ))               ## from an absolute path 'file://' URL
    fun(source(print(conn <- file(open = "r", .(abs_path_R))), local = TRUE               )); close(conn)  ## from an absolute path connection


    ## 'sys.source' cannot handle 'file://' URLs nor connections
    setwd(basename_dir)
    fun(sys.source(.(basename_R), envir = environment(), chdir = FALSE))
    fun(sys.source(.(basename_R), envir = environment(), chdir = TRUE ))
    setwd(rel_path_dir)
    fun(sys.source(.(rel_path_R), envir = environment(), chdir = FALSE))
    fun(sys.source(.(rel_path_R), envir = environment(), chdir = TRUE ))
    setwd(abs_path_dir)
    fun(sys.source(.(abs_path_R), envir = environment(), chdir = FALSE))
    fun(sys.source(.(abs_path_R), envir = environment(), chdir = TRUE ))


    ## 'debugSource' cannot handle connections
    if (.Platform$GUI == "RStudio") {
        debugSource <- get("debugSource", "tools:rstudio", inherits = FALSE)
        setwd(basename_dir)
        fun(debugSource(.(basename_R)    ))
        fun(debugSource(.(basename_R_URL)))
        setwd(rel_path_dir)
        fun(debugSource(.(rel_path_R)    ))
        fun(debugSource(.(rel_path_R_URL)))
        setwd(abs_path_dir)
        fun(debugSource(.(abs_path_R)    ))
        fun(debugSource(.(abs_path_R_URL)))
    }


    ## 'compiler::loadcmp' cannot handle 'file://' URLs nor connections
    if (requireNamespace("compiler", quietly = TRUE)) {
        basename_Rc <- basename_R; this.path::ext(basename_Rc) <- ".Rc"
        rel_path_Rc <- rel_path_R; this.path::ext(rel_path_Rc) <- ".Rc"
        abs_path_Rc <- abs_path_R; this.path::ext(abs_path_Rc) <- ".Rc"


        on.exit(unlink(abs_path_Rc), add = TRUE)
        compiler::cmpfile(abs_path_R, abs_path_Rc)


        opt2 <- options(`this.path::sys.path() expectation` = normalizePath(abs_path_Rc, "/", TRUE))
        setwd(basename_dir)
        fun(compiler::loadcmp(.(basename_Rc), envir = environment(), chdir = FALSE))
        fun(compiler::loadcmp(.(basename_Rc), envir = environment(), chdir = TRUE ))
        setwd(rel_path_dir)
        fun(compiler::loadcmp(.(rel_path_Rc), envir = environment(), chdir = FALSE))
        fun(compiler::loadcmp(.(rel_path_Rc), envir = environment(), chdir = TRUE ))
        setwd(abs_path_dir)
        fun(compiler::loadcmp(.(abs_path_Rc), envir = environment(), chdir = FALSE))
        fun(compiler::loadcmp(.(abs_path_Rc), envir = environment(), chdir = TRUE ))
        options(opt2)
    }


    ## 'utils::Sweave' cannot handle 'file://' URLs nor connections
    if (requireNamespace("utils", quietly = TRUE)) {
        basename_Rnw <- basename_R; this.path::ext(basename_Rnw) <- ".Rnw"
        abs_path_Rnw <- abs_path_R; this.path::ext(abs_path_Rnw) <- ".Rnw"


        on.exit(unlink(abs_path_Rnw), add = TRUE)
        writeLines(c(
            "\\documentclass{article}",
            "",
            "\\begin{document}",
            "",
            "<<>>=",
            ## remove expressions starting with 'cat'
            {
                exprs <- parse(abs_path_R)
                exprs <- exprs[!vapply(
                    exprs,
                    function(x) is.call(x) && identical(x[[1L]], as.symbol("cat")),
                    NA,
                    USE.NAMES = FALSE
                )]
                this.path:::.writeCode(exprs, NULL)
            },
            "@",
            "",
            "\\end{document}"
        ), abs_path_Rnw)


        opt2 <- options(`this.path::sys.path() expectation` = normalizePath(abs_path_Rnw, "/", TRUE))
        setwd(basename_dir)
        outputname <- fun(utils::Sweave(.(basename_Rnw)))
        writeLines(readLines(outputname))
        unlink(outputname)


        tmpdir <- tempfile("dir")
        on.exit(unlink(tmpdir, recursive = TRUE, force = TRUE), add = TRUE)
        dir.create(tmpdir)
        setwd(tmpdir)
        writeLines(readLines(fun(utils::Sweave(.(this.path::path.join("..", basename_Rnw))))))
        options(opt2)
    }


    ## 'box::use' cannot handle 'file://' URLs nor connections nor absolute paths
    if (requireNamespace("box", quietly = TRUE)) {
        tmp <- c(sub("\\.R$", ".r", abs_path_R), abs_path_R)
        tmp <- tmp[[match(TRUE, file.exists(tmp))]]
        opt2 <- options(`this.path::sys.path() expectation` = normalizePath(tmp, "/", TRUE))
        setwd(basename_dir); box::set_script_path(this.path::path.join(basename_dir, "."))
        fun(box::use(module = ./.(as.symbol(sub("\\.R$", "", basename_R))))); box::unload(module)
        if (!this.path:::.is_abs_path(rel_path_R)) {
            tmp_fun <- function(x) {
                n <- length(x)
                if (n > 1L)
                    call("/", tmp_fun(x[-n]), as.symbol(x[[n]]))
                else as.symbol(x[[1L]])
            }
            tmp <- tmp_fun(c(".", this.path::path.split.1(sub("\\.R$", "", rel_path_R))))
            setwd(rel_path_dir); box::set_script_path(this.path::path.join(rel_path_dir, "."))
            fun(box::use(module = .(tmp))); box::unload(module)
            rm(tmp, tmp_fun)
        }
        options(opt2)
    }


    ## 'knitr::knit' cannot handle 'file://' URLs
    if (requireNamespace("knitr", quietly = TRUE)) {
        basename_Rmd <- basename_R; this.path::ext(basename_Rmd) <- ".Rmd"
        rel_path_Rmd <- rel_path_R; this.path::ext(rel_path_Rmd) <- ".Rmd"
        abs_path_Rmd <- abs_path_R; this.path::ext(abs_path_Rmd) <- ".Rmd"


        on.exit(unlink(abs_path_Rmd), add = TRUE)
        writeLines(c(
            "```{r}",
            ## remove expressions starting with 'cat'
            {
                exprs <- parse(abs_path_R)
                exprs <- exprs[!vapply(
                    exprs,
                    function(x) is.call(x) && identical(x[[1L]], as.symbol("cat")),
                    NA,
                    USE.NAMES = FALSE
                )]
                this.path:::.writeCode(exprs, NULL)
            },
            "```"
        ), abs_path_Rmd)


        opt2 <- options(`this.path::sys.path() expectation` = normalizePath(abs_path_Rmd, "/", TRUE))
        setwd(basename_dir)
        fun(knitr::knit(.(basename_Rmd)                     , output = stdout(), quiet = TRUE))
        fun(knitr::knit(print(conn <- file(.(basename_Rmd))), output = stdout(), quiet = TRUE)); close(conn)
        setwd(rel_path_dir)
        fun(knitr::knit(.(rel_path_Rmd)                     , output = stdout(), quiet = TRUE))
        fun(knitr::knit(print(conn <- file(.(rel_path_Rmd))), output = stdout(), quiet = TRUE)); close(conn)
        setwd(abs_path_dir)
        fun(knitr::knit(.(abs_path_Rmd)                     , output = stdout(), quiet = TRUE))
        fun(knitr::knit(print(conn <- file(.(abs_path_Rmd))), output = stdout(), quiet = TRUE)); close(conn)
        options(opt2)
    }


    ## 'plumber::plumb' cannot handle 'file://' URLs nor connections
    if (requireNamespace("plumber", quietly = TRUE)) {
        opt2 <- options(`this.path::sys.path() expectation` = normalizePath(abs_path_R, "/", TRUE))
        setwd(basename_dir)
        fun(plumber::plumb(.(basename_R)))
        setwd(rel_path_dir)
        fun(plumber::plumb(.(rel_path_R)))
        setwd(abs_path_dir)
        fun(plumber::plumb(.(abs_path_R)))
        options(opt2)


        entrypoint_R <- this.path::path.join(basename_dir, "entrypoint.R")
        on.exit(unlink(entrypoint_R), add = TRUE)
        writeLines(c(
            readLines(abs_path_R),
            "plumber::Plumber$new()"
        ), entrypoint_R)
        opt2 <- options(`this.path::sys.path() expectation` = normalizePath(entrypoint_R, "/", TRUE))
        setwd(basename_dir)
        fun(plumber::plumb())
        setwd(rel_path_dir)
        fun(plumber::plumb(dir = .(dirname(rel_path_R))))
        setwd(abs_path_dir)
        fun(plumber::plumb(dir = .(dirname(abs_path_R))))
        options(opt2)
    }


    ## 'shiny::runApp'
    if (requireNamespace("shiny", quietly = TRUE)) {
        shinytmp <- tempfile("shinytmp", tmpdir = basename_dir)
        on.exit(unlink(shinytmp, recursive = TRUE, force = TRUE), add = TRUE)
        dir.create(shinytmp)
        file <- this.path::path.join(shinytmp, "app.R")
        writeLines(c(
            readLines(abs_path_R),
            "stop(structure(list(message = \"\", call = NULL), class = c(\"this_path_tests_R_catch_this_error\", \"error\", \"condition\")))"
        ), file)
        opt2 <- options(`this.path::sys.path() expectation` = normalizePath(file, "/", TRUE))
        rm(file)
        abs_path_app_R <- normalizePath(shinytmp, "/", TRUE)
        abs_path_app_dir <- abs_path_dir
        basename_app_R <- "."
        basename_app_dir <- abs_path_app_R
        tmp <- rel_path_and_dir(abs_path_app_R)
        rel_path_app_dir <- tmp[[1L]]
        rel_path_app_R <- tmp[[2L]]
        rm(tmp)
        setwd(basename_app_dir)
        this.path::tryCatch3({
            fun(shiny::runApp(.(basename_app_R)))
        }, this_path_tests_R_catch_this_error = )
        setwd(rel_path_app_dir)
        this.path::tryCatch3({
            fun(shiny::runApp(.(rel_path_app_R)))
        }, this_path_tests_R_catch_this_error = )
        setwd(abs_path_app_dir)
        this.path::tryCatch3({
            fun(shiny::runApp(.(abs_path_app_R)))
        }, this_path_tests_R_catch_this_error = )
        options(opt2)
    }


    ## 'testthat::source_file' cannot handle 'file://' URLs nor connections
    if (requireNamespace("testthat", quietly = TRUE)) {
        setwd(basename_dir)
        fun(testthat::source_file(.(basename_R), env = environment(), chdir = FALSE, wrap = FALSE))
        fun(testthat::source_file(.(basename_R), env = environment(), chdir = FALSE, wrap = TRUE ))
        fun(testthat::source_file(.(basename_R), env = environment(), chdir = TRUE , wrap = FALSE))
        fun(testthat::source_file(.(basename_R), env = environment(), chdir = TRUE , wrap = TRUE ))
        setwd(rel_path_dir)
        fun(testthat::source_file(.(rel_path_R), env = environment(), chdir = FALSE, wrap = FALSE))
        fun(testthat::source_file(.(rel_path_R), env = environment(), chdir = FALSE, wrap = TRUE ))
        fun(testthat::source_file(.(rel_path_R), env = environment(), chdir = TRUE , wrap = FALSE))
        fun(testthat::source_file(.(rel_path_R), env = environment(), chdir = TRUE , wrap = TRUE ))
        setwd(abs_path_dir)
        fun(testthat::source_file(.(abs_path_R), env = environment(), chdir = FALSE, wrap = FALSE))
        fun(testthat::source_file(.(abs_path_R), env = environment(), chdir = FALSE, wrap = TRUE ))
        fun(testthat::source_file(.(abs_path_R), env = environment(), chdir = TRUE , wrap = FALSE))
        fun(testthat::source_file(.(abs_path_R), env = environment(), chdir = TRUE , wrap = TRUE ))
    }


    invisible()
})


local({
    FILE.R <- tempfile(fileext = ".R")
    on.exit(unlink(FILE.R))
    this.path:::.writeCode({
        stopifnot(identical(
            this.path::this.path(),
            getOption("this.path::this.path() expectation")
        ))
    }, FILE.R)
    oopt <- options(
        `this.path::this.path() expectation` = normalizePath(FILE.R, "/", TRUE),
        keep.source = TRUE
    )
    on.exit(options(oopt), add = TRUE)
    eval(
        parse(FILE.R),
        structure(list2env(list(.packageName = FILE.R), parent = .BaseNamespaceEnv), path = FILE.R)
    )
    eval(parse(FILE.R))
})


local({
    FILE.R <- tempfile(fileext = ".R")
    on.exit(unlink(FILE.R))
    this.path:::.writeCode({
        list(
            this.path::src.path(original = TRUE),
            this.path::src.path(original = NA),
            this.path::src.path(),
            this.path::src.path(original = TRUE),
            this.path::src.path(original = NA)
        )
    }, FILE.R)
    oopt <- options(keep.source = TRUE)
    on.exit(options(oopt), add = TRUE)
    stopifnot(identical(
        eval(parse(FILE.R)),
        list(FILE.R, normalizePath(FILE.R, "/", TRUE))[c(1L, 1L, 2L, 1L, 2L)]
    ))
})


local({
    FILE1.R <- tempfile(pattern = "file1_", fileext = ".R")
    on.exit(unlink(FILE1.R), add = TRUE)
    this.path:::.writeCode({
        fun <- function(x) x
        fun1 <- function() fun(this.path::src.path())
    }, FILE1.R)
    source(FILE1.R, environment(), keep.source = TRUE)


    FILE2.R <- tempfile(pattern = "file2_", fileext = ".R")
    on.exit(unlink(FILE2.R), add = TRUE)
    this.path:::.writeCode({
        fun2 <- function() fun(this.path::src.path())
    }, FILE2.R)
    source(FILE2.R, environment(), keep.source = TRUE)


    ## it might seem weird to use eval(expression())
    ## it is just to prevent the expressions from having source references
    stopifnot(identical(eval(expression(fun1())), normalizePath(FILE1.R, "/", TRUE)))
    stopifnot(identical(eval(expression(fun2())), normalizePath(FILE2.R, "/", TRUE)))


    FILE3.R <- tempfile("file3_", fileext = ".R")
    on.exit(unlink(FILE3.R), add = TRUE)
    this.path:::.writeCode({
        x <- list(fun1(), fun2(), fun(this.path::src.path()))
    }, FILE3.R)
    source(FILE3.R, environment(), keep.source = TRUE)
    stopifnot(identical(x, as.list(normalizePath(c(FILE1.R, FILE2.R, FILE3.R), "/", TRUE))))
})

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2024   Iris Simmons
##


.tolower_ASCII <- function (x)
.External2(.C_tolower_ASCII, x)


.toupper_ASCII <- function (x)
.External2(.C_toupper_ASCII, x)


.str_equal_useBytes <- function (e1, e2)
.External2(.C_str_equal_useBytes, e1, e2)


.casefold_ASCII <- function (x, upper = FALSE)
if (upper) .toupper_ASCII(x) else .tolower_ASCII(x)


delayedAssign(".net_USE_command", {
    if (.OS_windows)
        paste(shQuote(paste0(Sys.getenv("windir"), "\\System32\\net.exe")), "USE")
})


.all_drives <- paste0(c(LETTERS, letters), ":/")


.relpath <- function (path, relative.to = getwd(), normalize = .OS_windows || !missing(relative.to), normalize.path = TRUE)
{
    if (!is.character(path))
        stop("a character vector argument expected", domain = "R")
    n <- length(path)
    if (n <= 0L) {
        character()
    } else {
        value <- character(n)
        value[] <- path
        if (any(i <- !is.na(value) & nzchar(value))) {
            path <- value[i]
            if (normalize.path)
                path <- .normalize_abspath_and_URL(path, "/", FALSE)
            if (!.IS_SCALAR_STR(relative.to))
                stop(gettextf("'%s' must be a character string", "relative.to", domain = "R"), domain = NA)
            if (normalize)
                relative.to <- .normalize_abspath_and_URL_1(relative.to, "/", TRUE)
            path <- c(relative.to, path)
            value[i] <- if (.OS_windows) {
                ## replace //LOCALHOST/C$/
                ## or      //127.0.0.1/C$/
                ## with    C:/
                enc <- Encoding(path)
                path <- sub("(?i)^[/\\\\][/\\\\](?:LOCALHOST|127\\.0\\.0\\.1)[/\\\\]([ABCDEFGHIJKLMNOPQRSTUVWXYZ])\\$([/\\\\]|$)",
                            "\\1:/", path, useBytes = TRUE)
                Encoding(path) <- enc
                p <- path.split(path)


                ## get the first element of each path, the drive,
                ## and then keep all the unique ones
                u <- unique(vapply(p, `[[`, "", 1L))
                no_convert_local <- if (length(u) == 1L) {
                    TRUE
                } else if (!any(j <- u %in% .all_drives)) {
                    TRUE
                } else if (all(j)) {
                    length(unique(.toupper_ASCII(u))) == 1L
                } else {
                    1L == length(u <- unique(.toupper_ASCII(u[j]))) &&
                        u == .toupper_ASCII(paste0(Sys.getenv("SystemDrive"), "/"))
                }
                if (no_convert_local) {
                    fix_local <- identity
                } else {
                    x <- system(.net_USE_command, intern = TRUE)
                    m <- regexec(" ([ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz]:) +(.*?) *(?:Web Client Network|Microsoft Windows Network)?$",
                        x, useBytes = TRUE)
                    ## one for the whole match, another two for the parenthesized sub-expressions
                    if (any(keep <- lengths(m) == 3L)) {
                        enc <- Encoding(x)[keep]
                        x <- regmatches(x[keep], m[keep])
                        local <- vapply(x, `[[`, "", 2L)
                        local <- .tolower_ASCII(local)
                        local <- paste0(local, "/")
                        remote <- vapply(x, `[[`, "", 3L)
                        if (any(j <- grepl("^[/\\\\]{2}", remote, useBytes = TRUE)))
                            remote[j] <- gsub("\\", "/", remote[j], fixed = TRUE, useBytes = TRUE)
                        remote <- paste0(remote, "/")
                        remote <- sub("(?i)^//(?:LOCALHOST|127\\.0\\.0\\.1)/([ABCDEFGHIJKLMNOPQRSTUVWXYZ])\\$/",
                            "\\1:/", remote, useBytes = TRUE)
                        Encoding(local) <- enc
                        Encoding(remote) <- enc
                        fix_local <- function(p) {
                            if (indx <- match(.tolower_ASCII(p[[1L]]), local, 0L)) {
                                c(remote[[indx]], p[-1L])
                            } else p
                        }
                    } else fix_local <- identity
                }


                r <- p[[1L]]
                p <- p[-1L]
                r <- fix_local(r)
                ignore_case <- !grepl("^(http|https)://", r[[1L]], useBytes = TRUE)
                fix_case <- if (ignore_case) .tolower_ASCII else identity
                r <- fix_case(r)
                len <- length(r)
                path.unsplit(lapply(p, function(p) {
                    n <- min(len, length(p))
                    q <- fix_case(fix_local(p))
                    n <- match(FALSE, .str_equal_useBytes(q[seq_len(n)], r[seq_len(n)]), n + 1L) - 1L
                    if (n == 0L) {
                        p
                    } else {
                        value <- c(rep("..", len - n), p[-seq_len(n)])
                        if (length(value) <= 0L)
                            "."
                        else if (!value[[1L]] %in% c(".", ".."))
                            c(".", value)
                        else value
                    }
                }))
            }
            else {
                p <- path.split(path)
                r <- p[[1L]]
                len <- length(r)
                path.unsplit(lapply(p[-1L], function(p) {
                    n <- min(len, length(p))
                    n <- match(FALSE, p[seq_len(n)] == r[seq_len(n)], n + 1L) - 1L
                    if (n == 0L) {
                        p
                    } else {
                        value <- c(rep("..", len - n), p[-seq_len(n)])
                        if (length(value) <= 0L)
                            "."
                        else if (!value[[1L]] %in% c(".", ".."))
                            c(".", value)
                        else value
                    }
                }))
            }
        }
        value
    }
}


relpath <- if (.Platform$OS.type == "windows") {
           function (path, relative.to = normalizePath(getwd(), "/", TRUE))
.relpath(path, relative.to, normalize = !missing(relative.to))
} else {
           function (path, relative.to = getwd())
.relpath(path, relative.to, normalize = !missing(relative.to))
}


rel2sys.dir <- function (path, local = FALSE)
{
    relative.to <- .External2(.C_sys_path, local)
    relative.to <- .dir(relative.to)
    .relpath(path, relative.to, normalize = FALSE)
}


rel2sys.proj <- function (path, local = FALSE)
{
    relative.to <- .External2(.C_sys_path, local)
    relative.to <- .dir(relative.to)
    relative.to <- .proj(relative.to)
    .relpath(path, relative.to, normalize = FALSE)
}


rel2env.dir <- function (path, n = 0L, envir = parent.frame(n + 1L), matchThisEnv = getOption("topLevelEnvironment"))
{
    n <- .External2(.C_asIntegerGE0, n)
    relative.to <- .External2(.C_env_path, envir, matchThisEnv)
    relative.to <- .dir(relative.to)
    .relpath(path, relative.to, normalize = FALSE)
}


rel2env.proj <- function (path, n = 0L, envir = parent.frame(n + 1L), matchThisEnv = getOption("topLevelEnvironment"))
{
    n <- .External2(.C_asIntegerGE0, n)
    relative.to <- .External2(.C_env_path, envir, matchThisEnv)
    relative.to <- .dir(relative.to)
    relative.to <- .proj(relative.to)
    .relpath(path, relative.to, normalize = FALSE)
}


rel2src.dir <- function (path, n = 0L, srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    relative.to <- .External2(.C_src_path, srcfile)
    relative.to <- .dir(relative.to)
    .relpath(path, relative.to, normalize = FALSE)
}


rel2src.proj <- function (path, n = 0L, srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    relative.to <- .External2(.C_src_path, srcfile)
    relative.to <- .dir(relative.to)
    relative.to <- .proj(relative.to)
    .relpath(path, relative.to, normalize = FALSE)
}


rel2here <- function (path, local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"),
    srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    relative.to <- .External2(.C_this_path, local, envir, matchThisEnv, srcfile)
    relative.to <- .dir(relative.to)
    .relpath(path, relative.to, normalize = FALSE)
}


rel2proj <- function (path, local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"),
    srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    relative.to <- .External2(.C_this_path, local, envir, matchThisEnv, srcfile)
    relative.to <- .dir(relative.to)
    relative.to <- .proj(relative.to)
    .relpath(path, relative.to, normalize = FALSE)
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2021-2026   Iris Simmons
##


## helper functions for this.path()    ----


.get_named_element <- function (x, names)
{
    for (name in names) {
        if (i <- match(name, names(x), 0L, c("", NA_character_)))
            x <- x[[i]]
        else return()
    }
    x
}


.get_contents <- function (file, encoding = getOption("encoding"))
{
    if (identical(encoding, "unknown")) {
        enc <- utils::localeToCharset()
        encoding <- enc[length(enc)]
    }
    else enc <- encoding
    if (length(enc) > 1L) {
        encoding <- NA
        owarn <- options(warn = 2)
        for (e in enc) {
            if (is.na(e))
                next
            zz <- file(file, encoding = e)
            res <- tryCatch(readLines(zz, warn = FALSE),
                error = identity)
            close(zz)
            if (!inherits(res, "error")) {
                encoding <- e
                break
            }
        }
        options(owarn)
    }
    if (is.na(encoding))
        stop("unable to find a plausible encoding")
    filename <- file
    file <- file(filename, "r", encoding = encoding)
    on.exit(close(file))
    readLines(file, warn = FALSE)
}


.get_jupyter_notebook_contents <- function (..., do.unlist = FALSE, give.f = TRUE)
{
    lines <- .get_contents(...)
    source <- jsonlite::parse_json(lines, simplifyVector = TRUE)
    source <- .get_named_element(source, c("cells", "source"))
    if (do.unlist) {
        value <- unlist(source)
        if (give.f)
            attr(value, "f") <- as.factor(rep(seq_along(source), lengths(source)))
        value
    }
    else source
}


.get_jupyter_R_notebook_contents <- function (path)
{
    ## similar to .get_jupyter_notebook_contents(), but does some error
    ## handling and checks that the metadata is valid for a Jupyter R Notebook
    path
    contents <- tryCatch(.get_contents(path), error = identity)
    if (!inherits(contents, "error")) {
        contents <- tryCatch(jsonlite::parse_json(contents, simplifyVector = TRUE),
            error = identity)
        if (!inherits(contents, "error")) {
            language <- .get_named_element(contents, c("metadata", "kernelspec", "language"))
            name     <- .get_named_element(contents, c("metadata", "language_info", "name"))
            version  <- .get_named_element(contents, c("metadata", "language_info", "version"))
            source   <- .get_named_element(contents, c("cells", "source"))
            # withAutoprint( { language; name; version; source } , spaced = TRUE, verbose = FALSE, width.cutoff = 60L); cat("\n\n\n\n\n")
            if (.scalar_streql(language, "R") &&
                .scalar_streql(name    , "R") &&
                .scalar_streql(version , as.character(getRversion())) &&
                is.list(source) && length(source) && all(vapply(source, is.character, NA, USE.NAMES = FALSE)))
            {
                return(source)
            }
        }
    }
    NULL
}


.fixNewlines <- function (srcfile)
{
    srcfile$lines <- .External2(.C_fixNewlines, srcfile$lines)
    srcfile$fixedNewlines <- TRUE
    srcfile$lines
}


.splitlines <- function (x)
.External2(.C_splitlines, x)


## path of active file in GUI          ----


.site_file_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (contents && .startup_info[["has_site_file"]])
        for.msg <- FALSE
    value <- site.file(original, for.msg, default = {
        stop(.ThisPathNotExistsError("site-wide startup profile file was not found"))
    })
    if (verbose) cat("Source: site-wide startup profile file\n")
    value
}


.init_file_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (contents && .startup_info[["has_init_file"]])
        for.msg <- FALSE
    value <- init.file(original, for.msg, default = {
        stop(.ThisPathNotExistsError("user profile was not found"))
    })
    if (verbose) cat("Source: user profile\n")
    value
}


.in_shell_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (contents && .has_shFILE)
        for.msg <- FALSE
    value <- shFILE(original, for.msg, default = {
        stop(.ThisPathNotExistsError("R is running from a shell and argument 'FILE' is missing"))
    })
    if (verbose) cat("Source: shell argument 'FILE'\n")
    if (contents) {
        if (is.na(value))
            NULL
        else .get_contents(value)
    }
    else value
}


.RStudio_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    indx <- match("tools:rstudio", search(), 0L)
    if (!indx) {
        if (for.msg)
            return(NA_character_)
        else stop(.ThisPathNotExistsError("RStudio has not finished loading"))
    }
    tools_rstudio <- as.environment(indx)


    ## ".rs.api.getActiveDocumentContext" from "tools:rstudio" returns a
    ## list of information about the document where your cursor is located
    ##
    ## ".rs.api.getSourceEditorContext" from "tools:rstudio" returns a list
    ## of information about the document open in the current tab
    ##
    ## element 'id' is a character string, an identification for the document
    ## element 'path' is a character string, the path of the document


    context <- NULL
    if (verbose) {
        .rs.api.getActiveDocumentContext <- get(".rs.api.getActiveDocumentContext", envir = tools_rstudio, inherits = FALSE)
        if (!is.function(.rs.api.getActiveDocumentContext))
            stop(
                simpleError(
                    gettextf(
                        "could not find function \"%s\"",
                        ".rs.api.getActiveDocumentContext",
                        domain = "R"
                    ),
                    quote(.rs.api.getActiveDocumentContext())
                )
            )
        context <- .rs.api.getActiveDocumentContext()
        active <- context[["id"]] != "#console"
        if (!active)
            context <- NULL
    }
    if (is.null(context)) {
        .rs.api.getSourceEditorContext <- get(".rs.api.getSourceEditorContext", envir = tools_rstudio, inherits = FALSE)
        if (!is.function(.rs.api.getSourceEditorContext))
            stop(
                simpleError(
                    gettextf(
                        "could not find function \"%s\"",
                        ".rs.api.getSourceEditorContext",
                        domain = "R"
                    ),
                    quote(.rs.api.getSourceEditorContext())
                )
            )
        context <- .rs.api.getSourceEditorContext()
    }


    if (is.null(context)) {
        if (for.msg)
            NA_character_
        else stop(.ThisPathNotExistsError("R is running from RStudio with no documents open\n (or source document has no associated path)"))
    }
    else if (contents) {
        if (verbose)
            cat(
                if (active)
                    "Source: active document in RStudio\n"
                else
                    "Source: source document in RStudio\n"
            )
        .External2(.C_remove_trailing_blank_string, context[["contents"]])
    }
    else if (nzchar(path <- context[["path"]])) {
        if (.OS_windows)
            ## on Windows, file path encoding is UTF-8
            ## (well, more specifically UCS-2, but unimportant for this)
            ## it is not explicitly set, so we have to do that ourselves
            Encoding(path) <- "UTF-8"
        if (verbose)
            cat(
                if (active)
                    "Source: active document in RStudio\n"
                else
                    "Source: source document in RStudio\n"
            )
        if (.isfalse(original))
            .normalizePath(path)
        else path
    }
    else if (for.msg)
        gettext("Untitled", domain = "RGui", trim = FALSE)
    else {
        stop(.ThisPathNotFoundError(
            if (verbose) {
                if (active)
                    "active document in RStudio has no associated path (has yet to be saved)"
                else
                    "source document in RStudio has no associated path (has yet to be saved)"
            }
            else "document in RStudio has no associated path (has yet to be saved)"
        ))
    }
}


.Positron_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    indx <- match("tools:positron", search(), 0L)
    if (!indx) {
        if (for.msg)
            return(NA_character_)
        else stop(.ThisPathNotExistsError("Positron has not finished loading"))
    }
    tools_positron <- as.environment(indx)


    ## ".ps.ui.LastActiveEditorContext" from "tools:positron" returns a
    ## list of information about the document open in the current tab


    .ps.ui.LastActiveEditorContext <- get(".ps.ui.LastActiveEditorContext", envir = tools_positron, inherits = FALSE)
    if (!is.function(.ps.ui.LastActiveEditorContext))
        stop(
            simpleError(
                gettextf(
                    "could not find function \"%s\"",
                    ".ps.ui.LastActiveEditorContext",
                    domain = "R"
                ),
                quote(.ps.ui.LastActiveEditorContext())
            )
        )
    context <- .ps.ui.LastActiveEditorContext()


    if (is.null(context)) {
        if (for.msg)
            NA_character_
        else stop(.ThisPathNotExistsError("R is running from Positron with no documents open\n (or document has no associated path)"))
    }
    else if (contents) {
        if (verbose) cat("Source: document in Positron\n")
        x <- context[["contents"]]
        storage.mode(x) <- "character"
        .External2(.C_remove_trailing_blank_string, x)
    }
    else if (context[["document"]][["is_untitled"]]) {
        if (for.msg)
            context[["document"]][["path"]]
        else stop(.ThisPathNotFoundError("document in Positron has no associated path (has yet to be saved)"))
    }
    else if (nzchar(path <- context[["document"]][["path"]])) {
        if (.OS_windows)
            Encoding(path) <- "UTF-8"
        if (verbose) cat("Source: document in Positron\n")
        if (.isfalse(original))
            .normalizePath(path)
        else path
    }
    else if (for.msg)
        gettext("Untitled", domain = "RGui", trim = FALSE)
    else stop(.ThisPathNotFoundError("document in Positron has no associated path (has yet to be saved)"))
}


.vscode_path <- evalq(envir = new.env(), {
    ## https://github.com/REditorSupport/vscode-R/blob/master/R/session/vsc.R#L1
    delayedAssign("pid", { Sys.getpid() })
    delayedAssign("wd", { getwd() })
    delayedAssign("homedir", { Sys.getenv(if (.Platform$OS.type == "windows") "USERPROFILE" else "HOME") })
    delayedAssign("dir_watcher", { Sys.getenv("VSCODE_WATCHER_DIR", file.path(homedir, ".vscode-R")) })
    delayedAssign("request_file", { file.path(dir_watcher, "request.log") })
    delayedAssign("request_lock_file", { file.path(dir_watcher, "request.lock") })
    ## https://github.com/REditorSupport/vscode-R/blob/master/R/session/vsc.R#L357
    delayedAssign("dir_session", {
        local({
            dir_session <- tempfile("vscode-R")
            dir.create(dir_session, showWarnings = FALSE, recursive = TRUE)
            dir_session
        })
    })
    ## https://github.com/REditorSupport/vscode-R/blob/master/R/session/vsc.R#L812
    response_timeout <- 5
    delayedAssign("response_lock_file", {
        local({
            ## force the other promise first
            response_file
            response_lock_file <- file.path(dir_session, "response.lock")
            writeLines("", response_lock_file)
            response_lock_file
        })
    })
    delayedAssign("response_file", {
        local({
            response_file <- file.path(dir_session, "response.log")
            file.create(response_file, showWarnings = FALSE)
            response_file
        })
    })
                function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    ## read the previous request time stamp before making a new request
    response_time_stamp <- readLines(response_lock_file)
    ## https://github.com/REditorSupport/vscode-R/blob/master/R/session/vsc.R#L204
    obj <- list(time = Sys.time(), pid = pid, wd = wd, command = "rstudioapi",
        action = "active_editor_context", args = list(), sd = dir_session)
    jsonlite::write_json(obj, request_file,
        auto_unbox = TRUE, null = "null", force = TRUE)
    cat(sprintf("%.6f", Sys.time()), file = request_lock_file)
    ## https://github.com/REditorSupport/vscode-R/blob/master/R/session/vsc.R#L838
    wait_start <- Sys.time()
    while (isTRUE(response_time_stamp == readLines(response_lock_file))) {
        if ((Sys.time() - wait_start) > response_timeout) {
            stop(.ThisPathNotFoundError(
                "Did not receive a response from VSCode-R API within ",
                response_timeout, " seconds."
            ))
        }
        Sys.sleep(0.1)
    }
    context <- jsonlite::read_json(response_file)
    if (is.null(context)) {
        if (for.msg)
            NA_character_
        else stop(.ThisPathNotExistsError("R is running from VSCode with no documents open\n (or document has no associated path)"))
    }
    else if (contents) {
        if (verbose) cat("Source: document in VSCode\n")
        .External2(.C_splitlines, context[["contents"]])
    }
    else if (context[["id"]][["scheme"]] == "untitled") {
        if (for.msg)
            context[["path"]]
        else stop(.ThisPathNotFoundError("document in VSCode has no associated path (has yet to be saved)"))
    }
    else if (nzchar(path <- context[["path"]])) {
        if (.OS_windows)
            Encoding(path) <- "UTF-8"
        if (verbose) cat("Source: document in VSCode\n")
        if (.isfalse(original))
            .normalizePath(path)
        else path
    }
    else if (for.msg)
        gettext("Untitled", domain = "RGui", trim = FALSE)
    else stop(.ThisPathNotFoundError("document in VSCode has no associated path (has yet to be saved)"))
}
})


.is_jupyter_loaded <- function ()
# .GUI_jupyter && isNamespaceLoaded("IRkernel") && .identical(sys.function(1L), IRkernel::main)
.GUI_jupyter && !is.null(ns <- .getNamespace("IRkernel")) && .identical(sys.function(1L), getExportedValue(ns, "main"))


.jupyter_path <- evalq(envir = new.env(), {
    delayedAssign("ofile", { NA_character_ })
    ofile
    delayedAssign("file" , { .normalizePath(ofile) })
                 function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (!is.na(ofile))
        return(.External2(.C_jupyter_path, verbose, original, for.msg, contents))


    if (is.null(initwd)) {
        if (for.msg)
            return(NA_character_)
        else stop(.ThisPathNotFoundError("R is running from Jupyter but the initial working directory is unknown"))
    }


    if (!.is_jupyter_loaded()) {
        if (for.msg)
            return(NA_character_)
        else stop(.ThisPathNotExistsError("Jupyter has not finished loading"))
    }


    n <- sys.frame(1L)[["kernel"]][["executor"]][["nframe"]] + 2L
    ocall <- sys.call(n)
    call <- .removeSource(ocall)


    files <- list.files(initwd, all.files = TRUE, full.names = TRUE, no.. = TRUE)
    files <- files[!dir.exists(files)]
    i <- grepl("\\.ipynb$", files, useBytes = TRUE)
    ipynb <- files[i]
    files <- files[!i]
    i <- grepl("\\.ipynb$", files, ignore.case = TRUE, useBytes = TRUE)
    IPYNB <- files[i]
    files <- files[!i]


    for (file in c(ipynb, IPYNB, files)) {
        for (lines in .get_jupyter_R_notebook_contents(file)) {
            exprs <- tryCatch(parse(text = lines, srcfile = NULL, keep.source = FALSE),
                error = identity)
            if (!inherits(exprs, "error")) {
                for (expr in exprs) {
                    if (.identical(expr, call)) {
                        .External2(.C_set_jupyter_path, file, skipCheck = TRUE)
                        return(.External2(.C_jupyter_path, verbose, original, for.msg, contents))
                    }
                }
            }
        }
    }


    if (for.msg)
        NA_character_
    else {
        stop(.ThisPathNotFoundError(
            sprintf("R is running from Jupyter with initial working directory '%s'\n", encodeString(initwd)),
            " but could not find a file with contents matching:\n",
            {
                t <- attr(ocall, "srcref", exact = TRUE)
                paste(if (is.integer(t)) as.character(t) else deparse(ocall), collapse = "\n")
            }
        ))
    }
}
})


set.jupyter.path <- function (...)
{
    if (!.GUI_jupyter)
        stop(gettextf("'%s' can only be called in Jupyter", "set.jupyter.path"))
    if (!.is_jupyter_loaded())
        stop(gettextf("'%s' can only be called after Jupyter has finished loading", "set.jupyter.path"))
    path <- if (missing(...) || ...length() == 1L && (is.null(..1) || is.atomic(..1) && length(..1) == 1L && is.na(..1)))
        NA_character_
    else if (is.null(initwd))
        path.join(...)
    else path.join(initwd, ...)
    .External2(.C_set_jupyter_path, path)
}


.emacs_path <- evalq(envir = new.env(), {
    .emacsclient <- function ()
{
    ## https://www.gnu.org/software/emacs/manual/html_node/emacs/Misc-Variables.html#index-emacs_005fdir
    if (.OS_unix) {
        x <- Sys.which("emacsclient")
        if (!nzchar(x)) stop(simpleError("command 'emacsclient' not found; add to PATH and retry", sys.call(-1L)))
        return(x)
    }
    x <- Sys.getenv("emacs_dir")
    if (nzchar(x)) {
        if (requireNamespace("utils", quietly = TRUE))
            x <- utils::shortPathName(x)
        else x <- .fixbackslash(x)
        if (endsWith(x, "\\"))
            x <- paste0(x, "bin\\emacsclient.exe")
        else x <- paste0(x, "\\bin\\emacsclient.exe")
        return(x)
    }
    x <- Sys.which("emacsclient")
    if (nzchar(x)) return(x)
    msg_start <- "command \"emacsclient.exe\" not found on PATH\n nor in default install location because "
    msg_end <- "\n add \"emacsclient.exe\" to PATH and retry"
    path <- Sys.getenv("SystemDrive")
    if (!nzchar(path))
        stop(msg_start, "environment variable 'SystemDrive' is not set", msg_end)
    path <- paste0(path, "/PROGRA~1")
    if (!dir.exists(path))
        stop(msg_start, path, " is not an existing directory", msg_end)
    x <- character(0)
    v <- character(0)
    ## check $SystemDrive/PROGRA~1/Emacs for directories matching
    ## emacs-{version}
    xx <- list.dirs(paste0(path, "/Emacs"), recursive = FALSE)
    if (length(xx)) {
        pattern <- "^emacs-((?:[[:digit:]]+[.-]){1,}[[:digit:]]+)$"
        y <- basename2(xx)
        m <- regexec(pattern, y)
        keep <- which(lengths(m) == 2L)
        xx <- xx[keep]
        vv <- regmatches(y[keep], m[keep])
        vv <- vapply(vv, `[[`, 2L, FUN.VALUE = "", USE.NAMES = FALSE)
        x <- c(x, xx)
        v <- c(v, vv)
    }
    ## check $SystemDrive/PROGRA~1 for directories matching
    ## GNU Emacs {version}
    xx <- list.dirs(path, recursive = FALSE)
    if (length(xx)) {
        pattern <- "^GNU Emacs ((?:[[:digit:]]+[.-]){1,}[[:digit:]]+)$"
        y <- basename2(xx)
        m <- regexec(pattern, y)
        keep <- which(lengths(m) == 2L)
        xx <- xx[keep]
        vv <- regmatches(y[keep], m[keep])
        vv <- vapply(vv, `[[`, 2L, FUN.VALUE = "", USE.NAMES = FALSE)
        x <- c(x, xx)
        v <- c(v, vv)
    }
    ## find the highest version and select it
    v <- numeric_version(v)
    x <- x[which.max(xtfrm(v))]
    if (length(x) != 1L)
        stop(msg_start, path, " does not contain a directory\n  'Emacs/emacs-{version}' nor 'GNU Emacs {version}'", msg_end)
    x <- paste0(x, "/bin/emacsclient.exe")
    if (file.access(x, 1L) != 0L)
        stop(msg_start, x, " does not exist or is not executable", msg_end)
    .fixbackslash(x)
}
    delayedAssign("emacsclient", { .emacsclient() })
    ## https://www.gnu.org/software/emacs/manual/html_node/elisp/Finding-All-Frames.html
    ## https://www.gnu.org/software/emacs/manual/html_node/elisp/Buffer-File-Name.html
    expr <- "
        (let ((R-pid    %d)
              (filename %s)
              (contents %s)
              (frames (frame-list-z-order))
              ess-r-mode-buffer
              found-matching-pid-anywhere
              active)
          ;; for testing purposes
          (setq found-matching-pid-anywhere (= R-pid 0))
          (while frames
            (setq ess-r-mode-buffer nil)
            (setq active t)
            (let ((buffers (frame-parameter (car frames) 'buffer-list))
                  (found-matching-pid (= R-pid 0)))
              (while buffers
                (with-current-buffer (car buffers)
                  (if buffer-file-name
                      (if (and (not ess-r-mode-buffer) (or (string= major-mode \"ess-r-mode\") (string= major-mode \"ess-mode\"))) (progn
                          (setq ess-r-mode-buffer (current-buffer))
                          (if found-matching-pid (setq active nil))
                      ))
                  ;; else
                  (if (and (not found-matching-pid) (or (string= major-mode \"inferior-ess-r-mode\") (string= major-mode \"inferior-ess-mode\")))
                      (let ((process (get-buffer-process (current-buffer))))
                        (if (and process (= (process-id process) R-pid)) (progn
                            (setq found-matching-pid t)
                            (setq found-matching-pid-anywhere t)
                        ))
                      )
                  ))
                )
                (if (and ess-r-mode-buffer found-matching-pid) (progn
                    (setq buffers nil)
                    (setq frames nil)
                ))
                (setq buffers (cdr buffers))
              )
            )
            (setq frames (cdr frames))
          )
          (if found-matching-pid-anywhere
              (if ess-r-mode-buffer
                  (with-current-buffer ess-r-mode-buffer
                    (if contents (progn
                        (write-region nil nil filename nil 0)
                        (if active 'success-active 'success-source)
                    ) ;; else
                    (if buffer-file-number (progn
                        (write-region buffer-file-name nil filename nil 0)
                        (if active 'success-active 'success-source)
                    ) ;; else
                        (if active 'untitled-active 'untitled-source)
                    ))
                  )
              )
          ;; else
              'no-matching-pid
          )
        )
    "
    expr <- gsub("^[ \t\n]+|[ \t\n]+$", "", expr)
    expr <- gsub(";.*?\n", "\n", expr)
    expr <- gsub("[ \t\n]+", " ", expr)
    expr <- gsub("( ", "(", expr, fixed = TRUE)
    expr <- gsub(" )", ")", expr, fixed = TRUE)
               function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    # exe <- Sys.which("emacsclient"); contents <- TRUE; Sys.getpid <- function() 0L; stop("comment this out later")


    file <- tempfile(fileext = ".txt")
    on.exit(unlink(file))
    ## create right now so no other program
    ## may claim this filename in the meantime
    file.create(file)
    exe <- .External2(.C_forcePromise_no_warn, "emacsclient")
    EXPR <- sprintf(expr,
        Sys.getpid(),
        encodeString(file, quote = "\""),
        if (contents) "t" else "nil"
    )
    args <- c(exe, "-e", EXPR)
    command <- paste(shQuote(args), collapse = " ")
    # cat(command, sep = "\n")
    rval <- suppressWarnings(system(command, intern = TRUE))
    if (!is.null(status <- attr(rval, "status")) && status) {
        if (status == -1L)
            stop(gettextf("'%s' could not be run", "emacsclient", domain = "R-base"), domain = NA)
        else stop(sprintf("'%s' execution failed with error code %d and message:\n\n", "emacsclient", status),
            paste(rval, collapse = "\n"),
            "\n\n perhaps add (server-start) to your ~/.emacs file and restart the session\n",
            " or type M-x server-start in your current session?", domain = NA)
    }


    if (!.IS_SCALAR_STR(rval))
        stop("command 'emacsclient' failed to return a character string")


    if (.scalar_streql(rval, "success-active")) {
        x <- readChar(file, nchars = file.size(file), useBytes = TRUE)
        if (verbose) cat("Source: active document in Emacs\n")
        if (contents)
            .External2(.C_splitlines, x)
        else if (.isfalse(original))
            .normalizePath(x)
        else x
    }
    else if (.scalar_streql(rval, "success-source")) {
        x <- readChar(file, nchars = file.size(file), useBytes = TRUE)
        if (verbose) cat("Source: source document in Emacs\n")
        if (contents)
            .External2(.C_splitlines, x)
        else if (.isfalse(original))
            .normalizePath(x)
        else x
    }
    else if (.scalar_streql(rval, "untitled-active")) {
        if (for.msg)
            gettext("Untitled", domain = "RGui", trim = FALSE)
        else stop(.ThisPathNotFoundError("active document in Emacs has no associated path (has yet to be saved)"))
    }
    else if (.scalar_streql(rval, "untitled-source")) {
        if (for.msg)
            gettext("Untitled", domain = "RGui", trim = FALSE)
        else stop(.ThisPathNotFoundError("source document in Emacs has no associated path (has yet to be saved)"))
    }
    else if (.scalar_streql(rval, "nil")) {
        if (for.msg)
            NA_character_
        else stop(.ThisPathNotFoundError("R is running from Emacs with no documents open"))
    }
    else if (.scalar_streql(rval, "no-matching-pid")) {
        if (for.msg)
            NA_character_
        else {
            stop(.ThisPathNotFoundError(
                sprintf(
                    "R process %d not found in primary Emacs session\n this.path() only works in primary Emacs session\n\n if you want to run multiple R sessions, do not run\n multiple Emacs sessions, one R session each.\n\n instead run one Emacs session with multiple frames,\n one R session each. this gives the same appearance\n while allowing this.path() to work across all R sessions.",
                    Sys.getpid()
                ),
                domain = NA
            ))
        }
    }
    else stop("invalid 'emacsclient' return value")
}
})


.rkward_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (for.msg)
        NA_character_
    else stop(.ThisPathNotFoundError("R is running from RKWard for which 'this.path' is currently unimplemented\n consider using RStudio, Positron, VSCode, or Emacs until such a time when this is implemented"))
}


.powerbi_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (for.msg)
        NA_character_
    else stop(.ThisPathNotFoundError("R is running from Power BI for which 'this.path' is unimplemented\n you should not need to use this.path() in Power BI since you can import data directly"))
}


.in_callr_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (for.msg)
        NA_character_
    else stop(.ThisPathNotFoundError("R is running from 'package:callr' for which 'this.path' is unimplemented\n 'package:callr' calls a function in a separate session,\n you should not need the path of the script where it was written"))
}


.r_editor_info <- sapply(
    c("r-editor", "untitled"),
    function(name) {
        dir <- "./inst/extdata"
        pattern <- sprintf("^%s_(msvcrt|ucrt)_([[:digit:]]+)_([[:digit:]]+)\\.dat$", name)
        files <- .BaseNamespaceEnv$list.files(dir, pattern)
        matches <- regmatches(files, regexec(pattern, files))
        info <- data.frame(
            crt = vapply(matches, `[[`, "", 2L),
            vrsn = R_system_version(sprintf(
                "%s.%s.0",
                vapply(matches, `[[`, "", 3L),
                vapply(matches, `[[`, "", 4L)
            )),
            as.data.frame.vector(lapply(file.path(dir, files), .read_C_strings_with_encoding), nm = "matches")
        )
        info <- info[order(info$vrsn, decreasing = TRUE), , drop = FALSE]
        info
    },
    simplify = FALSE
)


delayedAssign(".r_editor", {
    if (.OS_windows) local({
        x <- .r_editor_info$`r-editor`
        i <- match(TRUE, x$crt == (if (.ucrt) "ucrt" else "msvcrt") &
                         x$vrsn <= getRversion())
        if (is.na(i))
            " - R Editor"
        else x[[i, "matches"]]
    })
    else NULL
})
delayedAssign(".untitled", {
    if (.OS_windows) local({
        x <- .r_editor_info$untitled
        i <- match(TRUE, x$crt == (if (.ucrt) "ucrt" else "msvcrt") &
                         x$vrsn <= getRversion())
        if (is.na(i))
            "Untitled - R Editor"
        else x[[i, "matches"]]
    })
    else NULL
})


.Rgui_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
.External2(.C_Rgui_path, verbose, original, for.msg, contents, .untitled, .r_editor)


.AQUA_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (for.msg)
        NA_character_
    else stop(.ThisPathInAQUAError())
}


.Tk_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (for.msg)
        NA_character_
    else stop(.ThisPathNotExistsError("R is running from Tk which does not make use of its -f FILE, --file=FILE arguments"))
}


.gui_path <- function (verbose = FALSE, original = FALSE, for.msg = FALSE, contents = FALSE)
{
    if (.in_shell) {
        .in_shell_path(verbose, original, for.msg, contents)
    }
    else if (.GUI_RStudio) {
        .RStudio_path(verbose, original, for.msg, contents)
    }
    else if (.GUI_Positron) {
        .Positron_path(verbose, original, for.msg, contents)
    }
    else if (.GUI_vscode) {
        .vscode_path(verbose, original, for.msg, contents)
    }
    else if (.GUI_jupyter) {
        .jupyter_path(verbose, original, for.msg, contents)
    }
    else if (.GUI_emacs) {
        .emacs_path(verbose, original, for.msg, contents)
    }
    else if (.GUI_rkward) {
        .rkward_path(verbose, original, for.msg, contents)
    }
    else if (.GUI_powerbi) {
        .powerbi_path(verbose, original, for.msg, contents)
    }
    else if (.in_callr) {
        .in_callr_path(verbose, original, for.msg, contents)
    }
    else if (.GUI_Rgui) {
        ## we could use .Rgui_path(verbose, original, for.msg, contents)
        ## but it's slightly slower
        .External2(.C_Rgui_path, verbose, original, for.msg, contents, .untitled, .r_editor)
    }
    else if (.GUI_AQUA) {
        .AQUA_path(verbose, original, for.msg, contents)
    }
    ## running from a shell under Unix-alikes with GUI 'Tk'
    else if (.GUI_Tk) {
        .Tk_path(verbose, original, for.msg, contents)
    }
    ## running R in another manner
    else {
        if (for.msg)
            NA_character_
        else stop(.ThisPathUnrecognizedMannerError())
    }
}


set.gui.path <- function (...)
.External2(.C_set_gui_path)


## this.path(), this.dir(), and here() ----


## check that class name is C files matches class name in R files
local({
    tmp <- readLines("./src/ns-hooks.c")
    tmp <- tmp[[grep("^[ \t]*#[ \t]*define[ \t]+ThisPathNotExistsErrorClass_string[ \t]*\\\\[ \t]*$", tmp) + 1L]]
    tmp <- str2lang(tmp)
    if (!is.character(tmp) || length(tmp) != 1L || is.na(tmp))
        stop("could not determine class name")
    expected <- "ThisPathNotExistsError"
    if (tmp != expected)
        stop(sprintf("ThisPathNotExistsErrorClass found in ./src/ns-hooks.c (%s)\n does not match expected value in ./R/thispath.R (%s)",
            tmp, expected))
})


sys.path <- function (verbose = getOption("verbose"), original = FALSE, for.msg = FALSE,
    contents = FALSE, local = FALSE, default, else.)
{
    if (missing(default)) {
        if (missing(else.))
            .External2(.C_sys_path, verbose, original, for.msg, contents, local)
        else stop("'sys.path' with 'else.' but not 'default' makes no sense")
    }
    else {
        ## force the promises
        verbose
        original
        for.msg
        contents
        local
        if (missing(else.))
            tryCatch({
                .External2(.C_sys_path, verbose, original, for.msg, contents, local)
            }, ThisPathNotExistsError = function(e) default)
        else {
            tryCatch2({
                value <- .External2(.C_sys_path, verbose, original, for.msg, contents, local)
            }, ThisPathNotExistsError = function(e) default,
                else. = (else.)(value))
        }
    }
}


## part of _sys_path in ./src/thispath.c
.getframenumber <- function ()
.External2(.C_getframenumber)


sys.dir <- function (verbose = getOption("verbose"), local = FALSE, default, else.)
{
    if (missing(default)) {
        if (missing(else.)) {
            value <- .External2(.C_sys_path, verbose, local)
            .dir(value)
        }
        else stop("'sys.dir' with 'else.' but not 'default' makes no sense")
    }
    else {
        ## force the promises
        verbose
        local
        if (missing(else.))
            tryCatch({
                value <- .External2(.C_sys_path, verbose, local)
                .dir(value)
            }, ThisPathNotExistsError = function(e) default)
        else {
            tryCatch2({
                value <- .External2(.C_sys_path, verbose, local)
                value <- .dir(value)
            }, ThisPathNotExistsError = function(e) default,
                else. = (else.)(value))
        }
    }
}


env.path <- function (verbose = getOption("verbose"), original = FALSE, for.msg = FALSE,
    contents = FALSE, n = 0L, envir = parent.frame(n + 1L), matchThisEnv = getOption("topLevelEnvironment"),
    default, else.)
{
    n <- .External2(.C_asIntegerGE0, n)
    if (missing(default)) {
        if (missing(else.))
            .External2(.C_env_path, verbose, original, for.msg,contents, envir, matchThisEnv)
        else stop("'env.path' with 'else.' but not 'default' makes no sense")
    }
    else {
        ## force the promises
        verbose
        original
        for.msg
        contents
        envir
        matchThisEnv
        if (missing(else.))
            tryCatch({
                .External2(.C_env_path, verbose, original, for.msg, contents, envir, matchThisEnv)
            }, ThisPathNotExistsError = function(e) default)
        else {
            tryCatch2({
                value <- .External2(.C_env_path, verbose, original, for.msg, contents, envir, matchThisEnv)
            }, ThisPathNotExistsError = function(e) default,
                else. = (else.)(value))
        }
    }
}


env.dir <- function (verbose = getOption("verbose"), n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"), default, else.)
{
    n <- .External2(.C_asIntegerGE0, n)
    if (missing(default)) {
        if (missing(else.)) {
            value <- .External2(.C_env_path, verbose, envir, matchThisEnv)
            .dir(value)
        }
        else stop("'env.dir' with 'else.' but not 'default' makes no sense")
    }
    else {
        ## force the promises
        verbose
        envir
        matchThisEnv
        if (missing(else.))
            tryCatch({
                value <- .External2(.C_env_path, verbose, envir, matchThisEnv)
                .dir(value)
            }, ThisPathNotExistsError = function(e) default)
        else {
            tryCatch2({
                value <- .External2(.C_env_path, verbose, envir, matchThisEnv)
                value <- .dir(value)
            }, ThisPathNotExistsError = function(e) default,
                else. = (else.)(value))
        }
    }
}


src.path <- function (verbose = getOption("verbose"), original = FALSE, for.msg = FALSE,
    contents = FALSE, n = 0L, srcfile = if (n) sys.parent(n) else 0L, default, else.)
{
    n <- .External2(.C_asIntegerGE0, n)
    if (missing(default)) {
        if (missing(else.))
            .External2(.C_src_path, verbose, original, for.msg, contents, srcfile)
        else stop("'src.path' with 'else.' but not 'default' makes no sense")
    }
    else {
        ## force the promises
        verbose
        original
        for.msg
        contents
        srcfile
        if (missing(else.))
            tryCatch({
                .External2(.C_src_path, verbose, original, for.msg, contents, srcfile)
            }, ThisPathNotExistsError = function(e) default)
        else {
            tryCatch2({
                value <- .External2(.C_src_path, verbose, original, for.msg, contents, srcfile)
            }, ThisPathNotExistsError = function(e) default,
                else. = (else.)(value))
        }
    }
}


src.dir <- function (verbose = getOption("verbose"), n = 0L, srcfile = if (n) sys.parent(n) else 0L,
    default, else.)
{
    n <- .External2(.C_asIntegerGE0, n)
    if (missing(default)) {
        if (missing(else.)) {
            value <- .External2(.C_src_path, verbose, srcfile)
            .dir(value)
        }
        else stop("'src.dir' with 'else.' but not 'default' makes no sense")
    }
    else {
        ## force the promises
        verbose
        srcfile
        if (missing(else.))
            tryCatch({
                value <- .External2(.C_src_path, verbose, srcfile)
                .dir(value)
            }, ThisPathNotExistsError = function(e) default)
        else {
            tryCatch2({
                value <- .External2(.C_src_path, verbose, srcfile)
                value <- .dir(value)
            }, ThisPathNotExistsError = function(e) default,
                else. = (else.)(value))
        }
    }
}


this.path <- function (verbose = getOption("verbose"), original = FALSE, for.msg = FALSE,
    contents = FALSE, local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"), srcfile = if (n) sys.parent(n) else 0L,
    default, else.)
{
    n <- .External2(.C_asIntegerGE0, n)
    if (missing(default)) {
        if (missing(else.))
            .External2(.C_this_path, verbose, original, for.msg, contents, local, envir, matchThisEnv, srcfile)
        else stop("'this.path' with 'else.' but not 'default' makes no sense")
    }
    else {
        ## force the promises
        verbose
        original
        for.msg
        contents
        local
        envir
        matchThisEnv
        srcfile
        if (missing(else.))
            tryCatch({
                .External2(.C_this_path, verbose, original, for.msg, contents, local, envir, matchThisEnv, srcfile)
            }, ThisPathNotExistsError = function(e) default)
        else {
            tryCatch2({
                value <- .External2(.C_this_path, verbose, original, for.msg, contents, local, envir, matchThisEnv, srcfile)
            }, ThisPathNotExistsError = function(e) default,
                else. = (else.)(value))
        }
    }
}


this.dir <- function (verbose = getOption("verbose"), local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"), srcfile = if (n) sys.parent(n) else 0L,
    default, else.)
{
    n <- .External2(.C_asIntegerGE0, n)
    if (missing(default)) {
        if (missing(else.)) {
            value <- .External2(.C_this_path, verbose, local, envir, matchThisEnv, srcfile)
            .dir(value)
        }
        else stop("'this.dir' with 'else.' but not 'default' makes no sense")
    }
    else {
        ## force the promises
        verbose
        local
        envir
        matchThisEnv
        srcfile
        if (missing(else.))
            tryCatch({
                value <- .External2(.C_this_path, verbose, local, envir, matchThisEnv, srcfile)
                .dir(value)
            }, ThisPathNotExistsError = function(e) default)
        else {
            tryCatch2(value <- {
                value <- .External2(.C_this_path, verbose, local, envir, matchThisEnv, srcfile)
                .dir(value)
            }, ThisPathNotExistsError = function(e) default,
                else. = (else.)(value))
        }
    }
}


sys.here <- function (..., local = FALSE, .. = 0L)
{
    base <- .External2(.C_sys_path, local)
    base <- .here(base, ..)
    path.join(base, ...)
}


env.here <- function (..., n = 0L, envir = parent.frame(n + 1L), matchThisEnv = getOption("topLevelEnvironment"),
    .. = 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    base <- .External2(.C_env_path, envir, matchThisEnv)
    base <- .here(base, ..)
    path.join(base, ...)
}


src.here <- function (..., n = 0L, srcfile = if (n) sys.parent(n) else 0L,
    .. = 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    base <- .External2(.C_src_path, srcfile)
    base <- .here(base, ..)
    path.join(base, ...)
}


here <- function (..., local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"), srcfile = if (n) sys.parent(n) else 0L,
    .. = 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    base <- .External2(.C_this_path, local, envir, matchThisEnv, srcfile)
    base <- .here(base, ..)
    path.join(base, ...)
}
delayedAssign("ici", { here })





try.sys.path <- function (contents = FALSE, local = FALSE)
{
    ## force the promises before proceeding
    contents
    local
    success <- tryCatch({
        value <- .External2(.C_sys_path, FALSE, FALSE, FALSE, contents, local)
        TRUE
    }, error = function(e) FALSE)
    if (success)
        value
    else {
        .External2(.C_sys_path, FALSE, FALSE, TRUE, contents, local)
    }
}


try.env.path <- function (contents = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"))
{
    n <- .External2(.C_asIntegerGE0, n)
    contents
    envir
    matchThisEnv
    success <- tryCatch({
        value <- .External2(.C_env_path, FALSE, FALSE, FALSE, contents, envir, matchThisEnv)
        TRUE
    }, error = function(e) FALSE)
    if (success)
        value
    else {
        .External2(.C_env_path, FALSE, FALSE, TRUE, contents, envir, matchThisEnv)
    }
}


try.src.path <- function (contents = FALSE, n = 0L, srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    contents
    srcfile
    success <- tryCatch({
        value <- .External2(.C_src_path, FALSE, FALSE, FALSE, contents, srcfile)
        TRUE
    }, error = function(e) FALSE)
    if (success)
        value
    else {
        .External2(.C_src_path, FALSE, FALSE, TRUE, contents, srcfile)
    }
}


try.this.path <- function (contents = FALSE, local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"),
    srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    contents
    local
    envir
    matchThisEnv
    srcfile
    success <- tryCatch({
        value <- .External2(.C_this_path, FALSE, FALSE, FALSE,
            contents, local, envir, matchThisEnv, srcfile)
        TRUE
    }, error = function(e) FALSE)
    if (success)
        value
    else {
        .External2(.C_this_path, FALSE, FALSE, TRUE, contents,
            local, envir, matchThisEnv, srcfile)
    }
}


FILE <- function ()
{
    contents <- FALSE
    local <- FALSE
    envir <- parent.frame()
    matchThisEnv <- getOption("topLevelEnvironment")
    srcfile <- 0L
    success <- tryCatch({
        value <- .External2(.C_this_path, FALSE, FALSE, FALSE, contents, local, envir, matchThisEnv, srcfile)
        TRUE
    }, error = function(e) FALSE)
    if (success)
        value
    else {
        .External2(.C_this_path, FALSE, FALSE, TRUE, contents, local, envir, matchThisEnv, srcfile)
    }
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2024   Iris Simmons
##


tmp <- list(
    testfun = list(),
    desc = character(0),
    subdir = NULL,
    find_file = function() NULL,
    make_fix_file = function() NULL
)
class(tmp) <- "root_criterion"


i <- 0L


## has_file(".here")
i <- i + 1L
tmp$testfun[[i]] <- function (path)
.isfalse(.isdir(path.join(path, ".here")))
tmp$desc[[i]] <- "contains a file \".here\""


## is_rstudio_project
i <- i + 1L
tmp$testfun[[i]] <- function (path)
{
    files <- list.files(path, "\\.Rproj$", all.files = TRUE, full.names = TRUE)
    for (file in files[!dir.exists(files)])
        if (any(startsWith(readLines(file, 1L, warn = FALSE), "Version: ")))
            return(TRUE)
    FALSE
}
tmp$desc[[i]] <- "contains a file matching \"[.]Rproj$\" with contents matching \"^Version: \" in the first line"


## is_r_package
i <- i + 1L
tmp$testfun[[i]] <- function (path)
{
    file <- path.join(path, "DESCRIPTION")
    .isfalse(.isdir(file)) &&
        any(startsWith(readLines(file, warn = FALSE), "Package: "))
}
tmp$desc[[i]] <- "contains a file \"DESCRIPTION\" with contents matching \"^Package: \""


## is_remake_project
i <- i + 1L
tmp$testfun[[i]] <- function (path)
.isfalse(.isdir(path.join(path, "remake.yml")))
tmp$desc[[i]] <- "contains a file \"remake.yml\""


## is_projectile_project
i <- i + 1L
tmp$testfun[[i]] <- function (path)
.isfalse(.isdir(path.join(path, ".projectile")))
tmp$desc[[i]] <- "contains a file \".projectile\""


## is_vcs_root
i <- i + 1L
tmp$testfun[[i]] <- function (path)
dir.exists(path.join(path, ".git"))
tmp$desc[[i]] <- "contains a directory \".git\""
i <- i + 1L
tmp$testfun[[i]] <- function (path)
{
    file <- path.join(path, ".git")
    .isfalse(.isdir(file)) &&
        any(startsWith(readLines(file, warn = FALSE), "gitdir: "))
}
tmp$desc[[i]] <- "contains a file \".git\" with contents matching \"^gitdir: \""
i <- i + 1L
tmp$testfun[[i]] <- function (path)
dir.exists(path.join(path, ".svn"))
tmp$desc[[i]] <- "contains a directory \".svn\""


rm(i)


tmp$find_file <- function (..., path = ".")
{
    path <- normalizePath(path, winslash = "/", mustWork = TRUE)
    root <- .find_root(path = path, criterion = package_here_criterion_if_rprojroot_is_not_available)
    path.join(root, ...)
}


tmp$make_fix_file <- .removeSource_from_inner_functions(
                     function (path = getwd(), subdir = NULL)
{
    path <- normalizePath(path, winslash = "/", mustWork = TRUE)
    root <- .find_root(path = path, criterion = package_here_criterion_if_rprojroot_is_not_available)
    value <- function(...) NULL
    body(value) <- call("path.join", root, as.symbol("..."))
    environment(value) <- .mynamespace
    value
}
)


.find_root <- evalq(envir = new.env(), {
    environment(tmp$find_file) <<- environment()
    environment(tmp$make_fix_file) <<- environment()
    package_here_criterion_if_rprojroot_is_not_available <- tmp
    delayedAssign("package_here_criterion", {
        if (requireNamespace("rprojroot", quietly = TRUE))
            rprojroot::has_file(".here")     |
            rprojroot::is_rstudio_project    |
            rprojroot::is_r_package          |
            rprojroot::is_remake_project     |
            rprojroot::is_projectile_project |
            rprojroot::is_vcs_root
        else
            package_here_criterion_if_rprojroot_is_not_available
    })
    format.root_criterion <- function (x, ...)
{
    if (length(x$desc) > 1L)
        c("Root criterion: one of", paste0("- ", x$desc))
    else paste0("Root criterion: ", x$desc)
}
    print.root_criterion <- function (x, ...)
{
    cat(format(x), sep = "\n")
    invisible(x)
}
              function (path = getwd(), verbose = FALSE, criterion = package_here_criterion)
{
    # path <- "\\\\host\\share\\path\\to\\file\\"
    if (is.null(path))
        stop("cannot '.find_root' as current directory is unknown")
    if (.OS_windows && missing(path))
        path <- .normalizePath(path)
    else if (grepl("^(https|http|ftp|ftps)://", path, useBytes = TRUE))
        stop("this.proj() does not work for URL pathnames")
    if (!inherits(criterion, "root_criterion"))
        criterion <- rprojroot::as.root_criterion(criterion)
    opath <- path
    p <- path.split.1(path)
    while (n <- length(p)) {
        path <- path.unsplit(p)
        if (verbose) {
            for (i in seq_along(criterion$testfun)) {
                if (criterion$testfun[[i]](path)) {
                    cat("this.proj source: ", criterion$desc[[i]], "\n", sep = "")
                    return(path)
                }
            }
        }
        else {
            for (f in criterion$testfun) {
                if (f(path))
                    return(path)
            }
        }
        p <- p[-n]
    }
    stop(sprintf("no root directory found in %s or its parent directories\n%s",
        encodeString(opath, quote = "\""),
        paste(format(criterion), collapse = "\n")))
}
})


rm(tmp)


.proj <- evalq(envir = new.env(), {
    x <- structure(character(0), names = character(0))
         function (path, verbose = FALSE)
{
    ## 'path' should be normalized
    if (indx <- match(path, names(x), 0L))
        x[[indx]]
    else (x[[path]] <<- .find_root(path, verbose))
}
})


sys.proj <- function (..., local = FALSE)
{
    base <- .External2(.C_sys_path, local)
    base <- .dir(base)
    base <- .proj(base)
    path.join(base, ...)
}


env.proj <- function (..., n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"))
{
    n <- .External2(.C_asIntegerGE0, n)
    base <- .External2(.C_env_path, envir, matchThisEnv)
    base <- .dir(base)
    base <- .proj(base)
    path.join(base, ...)
}


src.proj <- function (..., n = 0L, srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    base <- .External2(.C_src_path, srcfile)
    base <- .dir(base)
    base <- .proj(base)
    path.join(base, ...)
}


this.proj <- function (..., local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"),
    srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    base <- .External2(.C_this_path, local, envir, matchThisEnv, srcfile)
    base <- .dir(base)
    base <- .proj(base)
    path.join(base, ...)
}


reset.proj <- function ()
.External2(.C_reset_proj)

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2024   Iris Simmons
##


`%check%` <- function (x, y)
stopifnot(identical(x, y))


path.join <- this.path:::.windows_path_join


## zero input gives zero output
path.join() %check% character()


## vector of length zero gives zero output
path.join(letters, character()) %check% character()


## binding to an empty string does nothing
path.join(letters, "") %check% letters


path.join("p1", "/p2", "p3/", "p4", "p5/") %check% "/p2/p3/p4/p5/"


path.join("C:", "p1") %check% "C:p1"
path.join("C:/", "p1") %check% "C:/p1"
path.join("C:/path/to/file1", "/path/to/file2") %check% "C:/path/to/file2"
path.join("//host/share/path/to/file1", "/path/to/file2") %check% "//host/share/path/to/file2"
path.join("C:p1", "C:/p2", "~", "~/p3", "//host",
    "//host/share", "//host/share/path/to/file", "p4") %check% "//host/share/path/to/file/p4"
path.join("p1", "c:/p2", "p3", "//host/share/p4", "p5", "c:p6", "p7", "C:p8", "p9") %check% "C:p6/p7/p8/p9"


path.join("C:/path/to/file1", "//host/share/path/to/file2") %check% "//host/share/path/to/file2"


path.join("//h/s", "p1", "C:p2", "p3/", "p4") %check% "C:p2/p3/p4"


path.join("//h/s/", "p1", "c:p2", "p3", "c:/p4", "p5", "C:p6", "p7/") %check% "C:/p4/p5/p6/p7/"


path.join("~/p1", "p2", "~/p3", "p4/") %check% "~/p3/p4/"


path.join("c:p1", "p2", "p3", "C:") %check% "C:p1/p2/p3"


## in 1.0.2-60 fixed mistake where forgot to
## nul terminate a string with no pathspec
path.join("//host/share") %check% "//host/share"


## in 1.1.0 fixed mistake where the non empty path spec was ignored
## because preceding path specs were empty
path.join("C:", "c:test") %check% "c:test"




path.join <- this.path:::.unix_path_join


## zero input gives zero output
path.join() %check% character()


## vector of length zero gives zero output
path.join(letters, character()) %check% character()


## binding to an empty string does nothing
path.join(letters, "") %check% letters


path.join("p1", "/p2", "p3/", "p4", "p5/") %check% "/p2/p3/p4/p5/"


path.join("/path/to/file1", "//host/share/path/to/file2") %check% "//host/share/path/to/file2"


path.join("//h/s", "p1", "/p2", "p3/", "p4") %check% "/p2/p3/p4"


path.join("/testing", "~/", "p1", "p2", "p3/") %check% "~/p1/p2/p3/"

##
## this.path : Get Executing Script's Path
## Copyright (C) 2024   Iris Simmons
##


this.path::env.path(verbose = FALSE)


#' @export
doc_cntxt <- get(
    ".this.path::document.context",
    envir = sys.frame(this.path:::.getframenumber()),
    inherits = FALSE
)


#' @export
fun_that_calls_src_path <- function (...)
{
    this.path::src.path(...)
}


#' @export
fun_that_calls_env_path <- function (...)
{
    this.path::env.path(...)
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2025   Iris Simmons
##


.windows_basename2 <- function (path)
.External2(.C_windows_basename2, path)


.unix_basename2 <- function (path)
.External2(.C_unix_basename2, path)


basename2 <- function (path, expand = TRUE)
.External2(.C_basename2, if (expand) path.expand(path) else path)


.windows_dirname2 <- function (path)
.External2(.C_windows_dirname2, path)


.unix_dirname2 <- function (path)
.External2(.C_unix_dirname2, path)


dirname2 <- function (path, expand = TRUE)
.External2(.C_dirname2, if (expand) path.expand(path) else path)


.windows_splitext <- function (path, compression = FALSE)
.External2(.C_windows_splitext, path, compression)


.unix_splitext <- function (path, compression = FALSE)
.External2(.C_unix_splitext, path, compression)


splitext <- function (path, compression = FALSE, expand = TRUE)
.External2(.C_splitext, if (expand) path.expand(path) else path, compression)


.windows_removeext <- function (path, compression = FALSE)
.External2(.C_windows_removeext, path, compression)


.unix_removeext <- function (path, compression = FALSE)
.External2(.C_unix_removeext, path, compression)


removeext <- function (path, compression = FALSE, expand = TRUE)
.External2(.C_removeext, if (expand) path.expand(path) else path, compression)


.windows_ext <- function (path, compression = FALSE)
.External2(.C_windows_ext, path, compression)


.unix_ext <- function (path, compression = FALSE)
.External2(.C_unix_ext, path, compression)


ext <- function (path, compression = FALSE, expand = TRUE)
.External2(.C_ext, if (expand) path.expand(path) else path, compression)


`.windows_ext<-` <- function (path, compression = FALSE, value)
.External2(`.C_windows_ext<-`, path, compression, value)


`.unix_ext<-` <- function (path, compression = FALSE, value)
.External2(`.C_unix_ext<-`, path, compression, value)


`ext<-` <- function (path, compression = FALSE, expand = TRUE, value)
{
    if (expand)
        path[] <- path.expand(path)
    .External2(`.C_ext<-`, path, compression, value)
}


.windows_path_join <- function (...)
.External2(.C_windows_path_join)


.unix_path_join <- function (...)
.External2(.C_unix_path_join)


path.join <- function (...)
.External2(.C_path_join)


.windows_path_split <- function (path)
.External2(.C_windows_path_split, path)


.unix_path_split <- function (path)
.External2(.C_unix_path_split, path)


path.split <- function (path)
.External2(.C_path_split, path)


.windows_path_split_1 <- function (path)
.External2(.C_windows_path_split_1, path)


.unix_path_split_1 <- function (path)
.External2(.C_unix_path_split_1, path)


path.split.1 <- function (path)
.External2(.C_path_split_1, path)


.windows_path_unsplit <- function (...)
.External2(.C_windows_path_unsplit)


.unix_path_unsplit <- function (...)
.External2(.C_unix_path_unsplit)


path.unsplit <- function (...)
.External2(.C_path_unsplit)


.is_abs_path <- function (path)
.External2(.C_is_abs_path, path)


.is_clipboard <- function (file)
.External2(.C_is_clipboard, file)


.fixslash <- function (s)
.External2(.C_fixslash, s)


.fixbackslash <- function (s)
.External2(.C_fixbackslash, s)


.as_file_URL <- function (path)
{
    if (!length(path))
        character(0)
    else if (.OS_windows) {
        ## on Windows we have file:///C:/path/to/file or similar
        enc <- Encoding(path)
        path <- gsub("\\", "/", path, fixed = TRUE, useBytes = TRUE)
        Encoding(path) <- enc
        i <- grepl("^.:", path, useBytes = TRUE)
        if (all(i))
            paste0("file:///", path)
        else if (any(i)) {
            path[ i] <- paste0("file:///", path[ i])
            path[!i] <- paste0("file://" , path[!i])
            path
        }
        else paste0("file://", path)
    }
    else paste0("file://", path)
}


.file_URL_path <- function (path)
.External2(.C_file_URL_path, path)


.normalizeURL <- function (path)
{
    # x <- "https://raw.githubusercontent.com////////////ArcadeAntics///testing/.././this.path/./main/tests/this.path_w_URLs.R"
    # print(c(x, this.path:::.normalizeURL(x)))
    # source(x)


    path.unsplit(lapply(path.split(path), function(p) {
        # x <- path.split("https://raw.githubusercontent.com////////////ArcadeAntics///testing/.././this.path/./main/tests/this.path_w_URLs.R")[[1L]]


        p <- p[p != "."]
        while (i <- match("..", p, 0L)) {
            p <- if (i == 2L) p[-2L] else p[-i + 0L:1L]
        }
        p
    }))
}


.normalizeURL_1 <- function (path)
{
    # path <- "https://raw.githubusercontent.com////////////ArcadeAntics///testing/.././this.path/./main/tests/this.path_w_URLs.R"


    p <- path.split.1(path)
    p <- p[p != "."]
    while (i <- match("..", p, 0L)) {
        p <- if (i == 2L) p[-2L] else p[-i + 0L:1L]
    }
    path.unsplit(p)
}


## turn a path, whether that be a basename, relative path, or absolute path,
## into an absolute path against the working directory
##
## does path expansion
.abspath <- function (path)
{
    wd <- getwd()
    if (is.null(wd))
        path.expand(path)
    else path.join(wd, path.expand(path))
}


## make a path absolute, then normalize it
.normalize_abspath <- function (path, ...)
normalizePath(path = if (.OS_windows) path else .abspath(path), ...)


.normalize_abspath_and_URL <- function (path, ...)
{
    ## a version of normalizePath that will also normalize URLs
    if (any(i <- startsWith(path, "file://")))
        path[i] <- .file_URL_path(path[i])
    if (any(i <- !i & grepl("^(https|http|ftp|ftps)://", path, useBytes = TRUE))) {
        path[i] <- .normalizeURL(path[i])
        path[!i] <- .normalize_abspath(path = path[!i], ...)
        path
    }
    else .normalize_abspath(path = path, ...)
}


.normalize_abspath_and_URL_1 <- function (path, ...)
{
    if (startsWith(path, "file://"))
        .normalize_abspath(path = .file_URL_path(path), ...)
    else if (grepl("^(ftp|ftps|http|https)://", path, useBytes = TRUE))
        .normalizeURL_1(path)
    else .normalize_abspath(path = path, ...)
}


.normalizePath <- function (path, winslash = "/", mustWork = TRUE)
normalizePath(path, winslash, mustWork)


.normalizePath_not_dir <- function (path, winslash = "/", mustWork = TRUE)
{
    x <- normalizePath(path, winslash, mustWork)
    if (.isfalse(.isdir(x)))
        x
    else stop(sprintf("'%s' is not a regular file", path), domain = NA)
}


.normalizePath_fix_dir <- function (path, winslash = "/", mustWork = TRUE)
{
    x <- normalizePath(path, winslash, mustWork)
    if (.istrue(.isdir(x)))
        path.join(x, ".")
    else x
}


.normalizePath_against <- function (wd, path, winslash = "/", mustWork = TRUE)
{
    owd <- getwd()
    if (is.null(owd))
        stop("cannot '.normalizePath_against' as current directory is unknown")
    on.exit(setwd(owd))
    setwd(wd)
    normalizePath(path, winslash, mustWork)
}


.normalizePath_not_dir_against <- function (wd, path, winslash = "/", mustWork = TRUE)
{
    owd <- getwd()
    if (is.null(owd))
        stop("cannot '.normalizePath_against' as current directory is unknown")
    on.exit(setwd(owd))
    setwd(wd)
    x <- normalizePath(path, winslash, mustWork)
    if (.isfalse(.isdir(x)))
        x
    else stop(sprintf("'%s' is not a regular file", path), domain = NA)
}


.normalizePath_fix_dir_against <- function (wd, path, winslash = "/", mustWork = TRUE)
{
    owd <- getwd()
    if (is.null(owd))
        stop("cannot '.normalizePath_against' as current directory is unknown")
    on.exit(setwd(owd))
    setwd(wd)
    x <- normalizePath(path, winslash, mustWork)
    if (.istrue(.isdir(x)))
        path.join(x, ".")
    else x
}


.normalizePath_srcfilealias <- function (original, path, winslash = "/", mustWork = TRUE)
{
    value <- .External2(.C_src_path, original)
    value <- .dir(value)
    if (grepl("^(https|http|ftp|ftps)://", value, useBytes = TRUE)) {
        ## do not use file.path(), on old versions of R it will convert text to native encoding
        .normalizeURL(paste(value, path, sep = "/"))
    }
    else {
        owd <- getwd()
        if (is.null(owd))
            stop("cannot '.normalizePath_against' as current directory is unknown")
        on.exit(setwd(owd))
        setwd(value)
        normalizePath(path, winslash, mustWork)
    }
}


.dir <- function (path)
{
    if (grepl("^(https|http|ftp|ftps)://", path, useBytes = TRUE)) {
        # path <- "https://raw.githubusercontent.com/ArcadeAntics/this.path/main/tests/this.path_w_URLs.R"
        p <- path.split.1(path)
        path.unsplit(if (length(p) >= 2L) p[-length(p)] else p)
    }
    else dirname2(path)
}


.here <- function (path, .. = 0L)
{
    if (grepl("^(https|http|ftp|ftps)://", path, useBytes = TRUE)) {
        # path <- "https://raw.githubusercontent.com/ArcadeAntics/this.path/main/tests/this.path_w_URLs.R"
        # .. <- "2"


        p <- path.split.1(path)
        n <- length(p) - length(seq_len(..)) - 1L
        path.unsplit(if (n < 1L) p[1L] else p[seq_len(n)])
    }


    # path <- "//host/share/path/to/file"
    # path <- "C:/Users/iris/Documents/this.path/man/this.path.Rd"
    # .. <- "10"
    else .External2(.C_dirname2, path.expand(path), ..)
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2021-2024   Iris Simmons
##


.defunctError <- function (new, package = NULL, msg, old = as.character(sys.call(sys.parent()))[1L])
{
    msg <- if (missing(msg)) {
        msg <- gettextf("'%s' is defunct.\n", old, domain = "R-base")
        if (!missing(new))
            msg <- c(msg, gettextf("Use '%s' instead.\n", new, domain = "R-base"))
        c(msg, if (!is.null(package))
            gettextf("See help(\"Defunct\") and help(\"%s-defunct\").", package, domain = "R-base")
        else gettext("See help(\"Defunct\")", domain = "R-base"))
    }
    else as.character(msg)
    msg <- paste(msg, collapse = "")
    if (missing(new))
        new <- NULL
    errorCondition(msg, old = old, new = new, package = package,
        class = "defunctError")
}


.getCurrentCall <- function (n = 2L, which = sys.parent(n))
{
    n <- .External2(.C_asIntegerGE0, n)
    if (which)
        sys.call(which)
    else NULL
}


.ThisPathInAQUAError <- function (call = .getCurrentCall(), call. = TRUE)
.External2(.C_ThisPathInAQUAError, if (call.) call)


.ThisPathInZipFileError <- function (description, call = .getCurrentCall(), call. = TRUE)
.External2(.C_ThisPathInZipFileError, if (call.) call, description)


.ThisPathNotExistsError <- function (..., call. = TRUE, domain = NULL, call = .getCurrentCall())
.External2(.C_ThisPathNotExistsError, .makeMessage(..., domain = domain), call = if (call.) call)


delayedAssign("thisPathNotExistsError", { .ThisPathNotExistsError })


.ThisPathNotFoundError <- function (..., call. = TRUE, domain = NULL, call = .getCurrentCall())
.External2(.C_ThisPathNotFoundError, .makeMessage(..., domain = domain), call = if (call.) call)


delayedAssign("thisPathNotFoundError", { .ThisPathNotFoundError })


.ThisPathNotImplementedError <- function (..., call. = TRUE, domain = NULL, call = .getCurrentCall())
.External2(.C_ThisPathNotImplementedError, .makeMessage(..., domain = domain), call = if (call.) call)


.ThisPathUnrecognizedConnectionClassError <- function (con, call = .getCurrentCall(), call. = TRUE)
.External2(.C_ThisPathUnrecognizedConnectionClassError, if (call.) call, con)


.ThisPathUnrecognizedMannerError <- function (call = .getCurrentCall(), call. = TRUE)
.External2(.C_ThisPathUnrecognizedMannerError, if (call.) call)





tryCatch2 <- function (expr, ..., else., finally)
.External2(.C_tryCatch2)


.last.condition <- function (c)
.External2(.C_last_condition, c)


last.condition <- function ()
.External2(.C_last_condition)


tryCatch3 <- function (expr, ..., else., finally)
.External2(.C_tryCatch3)


# tryCatch3(message("testing"), message = , error = writeLines("caught error"), warning = 6, test = , finally = writeLines("finally"), else. = writeLines("else."))

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2026   Iris Simmons
##


if (FALSE) {


.eval_with_message <- evalq(envir = new.env(), {
    indent <- 0L
function (sym, val)
{
    writeLines(sprintf("%sevaluating '%s'", strrep(" ", indent), sym))
    oindent <- indent
    on.exit(indent <<- oindent)
    indent <<- indent + if (indent < 16L) 4L else 2L
    val
}
})


tmp <- function (..., envir = parent.frame())
{
    if (!is.environment(envir))
        envir <- as.environment(envir)
    syms <- as.list(substitute(list(...)))[-1L]
    if (any(i <- vapply(syms, typeof, "") != "symbol"))
        stop(sprintf(ngettext(sum(i), "argument %s is not a symbol",
                                      "arguments %s are not symbols"),
            paste(which(i), collapse = ", ")))
    lapply(as.character(syms), function(sym) {
        if (!exists(sym, envir = envir, inherits = FALSE))
            stop(gettextf("object '%s' not found", sym, domain = "R"), domain = NA)
        expr <- call("substitute", as.symbol(sym), quote(envir))
        expr <- eval(expr)
        expr <- call(".eval_with_message", sym, expr)
        expr <- call("delayedAssign", quote(sym), expr, quote(envir), quote(envir))
        eval(expr)
    })
    invisible()
}


tmp(
    ## promises.R ----


    .OS_unix, .OS_windows,
    .GUI_RStudio,
    .GUI_Positron,
    .OS_unix_maybe_unembedded_shell, .OS_windows_maybe_unembedded_shell, .maybe_unembedded_shell,
    .shINFO,
    .console_radian,
    .GUI_vscode,
    .GUI_jupyter,
    .GUI_emacs,
    .GUI_rkward,
    .GUI_powerbi,
    .in_callr,
    .GUI_AQUA, .GUI_Rgui, .GUI_Tk,
    .OS_unix_in_shell, .OS_windows_in_shell, .in_shell,
    .unrecognized_manner,
    .ucrt, .GUI,
    initwd,


    ## relpath.R  ----


    .net_USE_command,


    ## shfile.R   ----


    .has_shFILE,


    ## thispath.R ----


    .r_editor, .untitled
)


## rprojroot.R ----
envir <- environment(.find_root)
tmp(package_here_criterion, envir = envir)
rm(envir)


## shfile.R  ----
envir <- environment(.shFILE)
tmp(ofile, file, envir = envir)
rm(envir)


rm(tmp)


}


if (getRversion() < "3.0.0") {


    R_Visible <- local({
        pattern <- "^\\{\"([^\"]+)\" *, \\(DL_FUNC\\) &do_[^ ]+ *, +-?[[:digit:]]+\\},(?:| // R_Visible (updatable|off))$"
        x <- readLines("./src/0.c")
        m <- regexec(pattern, x)
        keep <- which(vapply(m, length, 0L) > 1L)
        x <- x[keep]
        m <- m[keep]
        y <- regmatches(x, m)
        y <- structure(
            vapply(y, `[`, 3L, FUN.VALUE = "", USE.NAMES = FALSE),
            names = vapply(y, `[`, 2L, FUN.VALUE = "", USE.NAMES = FALSE)
        )
        y <- y[nzchar(y)]
        if (length(y))
            names(y) <- paste0(".C_", names(y))
        y
    })


    fix_External2 <- function(x, envir = parent.frame()) {
        if (!is.environment(envir))
            envir <- as.environment(envir)
        sym <- substitute(x)
        if (typeof(sym) != "symbol")
            stop("not a symbol", domain = "R")
        sym <- as.character(sym)
        if (!exists(sym, envir = envir, inherits = FALSE))
            stop(gettextf("object '%s' not found", sym, domain = "R"), domain = NA)
        recurse <- function(e) {
            if (is.function(e)) {
                if (!is.primitive(e)) {
                    at <- attributes(e)
                    body_is_bytecode <- (typeof(.Internal(bodyCode(e))) == "bytecode")
                    formals(e) <- recurse(formals(e))
                    body(e) <- recurse(body(e))
                    if (body_is_bytecode && requireNamespace("compiler", quietly = TRUE))
                        e <- compiler::cmpfun(e)
                    if (!is.null(at))
                        attributes(e) <- at
                }
                e
            }
            else if (is.pairlist(e))
                as.pairlist(lapply(e, recurse))
            else if (is.call(e)) {
                if (identical(e[[1L]], quote(.External2))) {
                    # e <- body(this.path:::print.ThisPathDocumentContext)
                    e[[1L]] <- quote(.External)
                    e <- append(
                        as.list(e),
                        after = 2L,
                        list(
                            call("quote", e),
                            quote(.External),
                            call("environment")
                        )
                    )
                    e <- as.call(e)
                    m <- match(as.character(e[[2L]]), names(R_Visible), 0L)
                    if (m) {
                        if (R_Visible[[m]] == "off")
                            e <- call("invisible", e)
                        else if (R_Visible[[m]] == "updatable") {
                            e <- call("{",
                                quote(.this_path_value <- NULL),
                                quote(.this_path_visible <- TRUE),
                                e,
                                quote(if (.this_path_visible)
                                    .this_path_value
                                else invisible(.this_path_value))
                            )
                        }
                    }
                    e
                }
                else as.call(lapply(e, recurse))
            }
            else e
        }
        x <- eval(call("substitute", as.symbol(sym), quote(envir)))
        value <- recurse(x)
        if (typeof(x) == "closure") {
            assign(sym, value, envir = envir, inherits = FALSE)
        } else {
            x <- call("delayedAssign", sym, value, quote(envir), quote(envir))
            eval(x)
        }
        invisible()
    }


    ## backports.R     ----


    fix_External2(anyNA)
    fix_External2(anyNA.data.frame)
    fix_External2(anyNA.numeric_version)
    fix_External2(.anyNA.default)

    fix_External2(isNamespaceLoaded)
    fix_External2(dir.exists)
    fix_External2(lengths)
    fix_External2(.lengths.default)

    fix_External2(strrep)
    fix_External2(startsWith)
    fix_External2(endsWith)

    fix_External2(...length)

    fix_External2(...elt)


    ## checkpath.R     ----


    fix_External2(check.path)
    fix_External2(check.dir)
    fix_External2(check.proj)


    ## enhances.R      ----


    fix_External2(with_script_path)
    fix_External2(make_fix_file)


    ## error.R         ----


    fix_External2(.getCurrentCall)
    fix_External2(.ThisPathInAQUAError)
    fix_External2(.ThisPathInZipFileError)
    fix_External2(.ThisPathNotExistsError)
    fix_External2(.ThisPathNotFoundError)
    fix_External2(.ThisPathNotImplementedError)
    fix_External2(.ThisPathUnrecognizedConnectionClassError)
    fix_External2(.ThisPathUnrecognizedMannerError)


    fix_External2(tryCatch2)
    fix_External2(.last.condition)
    fix_External2(last.condition)
    fix_External2(tryCatch3)


    ## files.R         ----


    fix_External2(.windows_basename2)
    fix_External2(.unix_basename2)
    fix_External2(basename2)
    fix_External2(.windows_dirname2)
    fix_External2(.unix_dirname2)
    fix_External2(dirname2)
    fix_External2(.windows_splitext)
    fix_External2(.unix_splitext)
    fix_External2(splitext)
    fix_External2(.windows_removeext)
    fix_External2(.unix_removeext)
    fix_External2(removeext)
    fix_External2(.windows_ext)
    fix_External2(.unix_ext)
    fix_External2(ext)
    fix_External2(`.windows_ext<-`)
    fix_External2(`.unix_ext<-`)
    fix_External2(`ext<-`)
    fix_External2(.windows_path_join)
    fix_External2(.unix_path_join)
    fix_External2(path.join)
    fix_External2(.windows_path_split)
    fix_External2(.unix_path_split)
    fix_External2(path.split)
    fix_External2(.windows_path_split_1)
    fix_External2(.unix_path_split_1)
    fix_External2(path.split.1)
    fix_External2(.windows_path_unsplit)
    fix_External2(.unix_path_unsplit)
    fix_External2(path.unsplit)
    fix_External2(.is_abs_path)
    fix_External2(.is_clipboard)
    fix_External2(.fixslash)
    fix_External2(.fixbackslash)
    fix_External2(.file_URL_path)
    fix_External2(.normalizePath_srcfilealias)
    fix_External2(.here)


    ## ismain.R        ----


    fix_External2(is.main)
    fix_External2(from.shell)


    ## lineno.R        ----


    fix_External2(.LINENO)
    fix_External2(sys.LINENO)
    fix_External2(env.LINENO)
    fix_External2(src.LINENO)
    fix_External2(LINENO)
    fix_External2(LINE)


    ## make_fix_funs.R ----


    fix_External2(path.functions)


    ## ns-hooks.R      ----


    fix_External2(.mbcslocale)
    fix_External2(.R_MB_CUR_MAX)

    if (!isTRUE(getOption("R_THIS_PATH_DEVEL")))
        fix_External2(.get_ptrs)

    fix_External2(.onLoad)
    fix_External2(.onUnload)


    ## print.R         ----


    fix_External2(.PrintValueEnv)
    fix_External2(.maybePrintValueEnv)
    fix_External2(print.ThisPathDocumentContext)


    ## progargs.R      ----


    fix_External2(progArgs)
    fix_External2(withArgs)


    ## promises.R      ----


    fix_External2(.OS_windows_maybe_unembedded_shell)
    fix_External2(.shINFO)
    fix_External2(.GUI_Rgui)


    ## relpath.R       ----


    fix_External2(.tolower_ASCII)
    fix_External2(.toupper_ASCII)
    fix_External2(.str_equal_useBytes)
    fix_External2(rel2sys.dir)
    fix_External2(rel2sys.proj)
    fix_External2(rel2env.dir)
    fix_External2(rel2env.proj)
    fix_External2(rel2src.dir)
    fix_External2(rel2src.proj)
    fix_External2(rel2here)
    fix_External2(rel2proj)


    ## rprojroot.R     ----


    fix_External2(sys.proj)
    fix_External2(env.proj)
    fix_External2(src.proj)
    fix_External2(this.proj)
    fix_External2(reset.proj)


    ## setsyspath.R    ----


    fix_External2(wrap.source)
    fix_External2(set.sys.path)
    fix_External2(unset.sys.path)
    fix_External2(set.env.path)
    fix_External2(set.src.path)
    fix_External2(set.sys.path.function)


    ## shfile.R        ----


    fix_External2(.shFILE)


    ## startup.R       ----


    fix_External2(.site_file)
    fix_External2(.init_file)
    fix_External2(with_site.file)
    fix_External2(with_init.file)


    ## sys.R           ----


    fix_External2(sys.srcref)
    fix_External2(sys.whiches)


    ## tests.R         ----


    fix_External2(.faster_subsequent_times_test)


    ## thispath.R      ----


    fix_External2(.fixNewlines)
    fix_External2(.splitlines)
    fix_External2(.RStudio_path)
    fix_External2(.Positron_path)
    fix_External2(.vscode_path)
    fix_External2(.jupyter_path)
    fix_External2(set.jupyter.path)
    fix_External2(.emacs_path)
    fix_External2(.Rgui_path)
    fix_External2(.gui_path)
    fix_External2(set.gui.path)
    fix_External2(sys.path)
    fix_External2(.getframenumber)
    fix_External2(sys.dir)
    fix_External2(env.path)
    fix_External2(env.dir)
    fix_External2(src.path)
    fix_External2(src.dir)
    fix_External2(this.path)
    fix_External2(this.dir)
    fix_External2(sys.here)
    fix_External2(env.here)
    fix_External2(src.here)
    fix_External2(here)
    fix_External2(try.sys.path)
    fix_External2(try.env.path)
    fix_External2(try.src.path)
    fix_External2(try.this.path)
    fix_External2(FILE)


    ## utils.R         ----


    fix_External2(.istrue)
    fix_External2(.isfalse)
    fix_External2(.asLogical)
    fix_External2(.asInteger)
    fix_External2(.asIntegerGE0)
    fix_External2(.IS_SCALAR_STR)
    fix_External2(.AS_SCALAR_STR)
    fix_External2(.scalar_streql)


    rm(fix_External2)
}


if (requireNamespace("compiler", quietly = TRUE)) {
    ocompilePKGS <- compiler::compilePKGS(FALSE)
    compiler::compilePKGS(ocompilePKGS)
    if (ocompilePKGS) {
        tmp <- environment(.find_root)
        tmp$package_here_criterion_if_rprojroot_is_not_available$testfun[] <- lapply(
            tmp$package_here_criterion_if_rprojroot_is_not_available$testfun,
            compiler::cmpfun
        )
        tmp$package_here_criterion_if_rprojroot_is_not_available$find_file <-
            compiler::cmpfun(tmp$package_here_criterion_if_rprojroot_is_not_available$find_file)
        tmp$package_here_criterion_if_rprojroot_is_not_available$make_fix_file <-
            compiler::cmpfun(tmp$package_here_criterion_if_rprojroot_is_not_available$make_fix_file)
        tmp$format.root_criterion <- compiler::cmpfun(tmp$format.root_criterion)
        tmp$print.root_criterion <- compiler::cmpfun(tmp$print.root_criterion)
        rm(tmp)


        tmp <- environment(.emacs_path)
        tmp$.emacsclient <- compiler::cmpfun(tmp$.emacsclient)
        rm(tmp)
    }
    rm(ocompilePKGS)
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2026   Iris Simmons
##


wrap.source <- function (expr, path.only = FALSE, character.only = path.only,
    file.only = path.only, conv2utf8 = FALSE, allow.blank.string = FALSE,
    allow.clipboard = !file.only, allow.stdin = !file.only, allow.url = !file.only,
    allow.file.uri = !path.only, allow.unz = !path.only, allow.pipe = !file.only,
    allow.terminal = !file.only, allow.textConnection = !file.only,
    allow.rawConnection = !file.only, allow.sockconn = !file.only,
    allow.servsockconn = !file.only, allow.customConnection = !file.only,
    ignore.all = FALSE, ignore.blank.string = ignore.all, ignore.clipboard = ignore.all,
    ignore.stdin = ignore.all, ignore.url = ignore.all, ignore.file.uri = ignore.all)
.External2(.C_wrap_source, character.only, conv2utf8, allow.blank.string,
    allow.clipboard, allow.stdin, allow.url, allow.file.uri,
    allow.unz, allow.pipe, allow.terminal, allow.textConnection,
    allow.rawConnection, allow.sockconn, allow.servsockconn,
    allow.customConnection, ignore.blank.string, ignore.clipboard,
    ignore.stdin, ignore.url, ignore.file.uri)


set.sys.path <- function (file, path.only = FALSE, character.only = path.only,
    file.only = path.only, conv2utf8 = FALSE, allow.blank.string = FALSE,
    allow.clipboard = !file.only, allow.stdin = !file.only, allow.url = !file.only,
    allow.file.uri = !path.only, allow.unz = !path.only, allow.pipe = !file.only,
    allow.terminal = !file.only, allow.textConnection = !file.only,
    allow.rawConnection = !file.only, allow.sockconn = !file.only,
    allow.servsockconn = !file.only, allow.customConnection = !file.only,
    ignore.all = FALSE, ignore.blank.string = ignore.all, ignore.clipboard = ignore.all,
    ignore.stdin = ignore.all, ignore.url = ignore.all, ignore.file.uri = ignore.all,
    Function = NULL, ofile, delayed = FALSE, assign_to = NULL)
.External2(.C_set_sys_path, character.only, conv2utf8, allow.blank.string,
    allow.clipboard, allow.stdin, allow.url, allow.file.uri,
    allow.unz, allow.pipe, allow.terminal, allow.textConnection,
    allow.rawConnection, allow.sockconn, allow.servsockconn,
    allow.customConnection, ignore.blank.string, ignore.clipboard,
    ignore.stdin, ignore.url, ignore.file.uri, Function, delayed,
    assign_to)


local({
    f1 <- formals(wrap.source)
    f2 <- formals(set.sys.path)
    i <- 4L
    stopifnot(
        length(f1) + i == length(f2),
        identical(f1[-1L], f2[2L:(length(f2) - i)]),
        TRUE
    )
    b1 <- body(wrap.source)
    b2 <- body(set.sys.path)
    stopifnot(
        length(b1) + (i - 1L) == length(b2),
        identical(b1[-2L], b2[-c(2L, length(b2) - ((i - 2L):0L))])
    )
})


unset.sys.path <- function ()
.External2(.C_unset_sys_path)


set.env.path <- function (envir, matchThisEnv = getOption("topLevelEnvironment"))
.External2(.C_set_env_path, envir, matchThisEnv)


set.src.path <- function (srcfile)
.External2(.C_set_src_path, srcfile)


set.sys.path.function <- function (fun)
.External2(.C_set_sys_path_function, fun)


with_sys.path <- function (file, expr, ...)
{
    set.sys.path(file = file, ..., Function = c("with_sys.path", "this.path"), assign_to = NULL)
    expr
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2024   Iris Simmons
##


is.main <- function ()
# {
#     n <- .External2(.C_getframenumber)
#     if (is.na(n))
#         NA
#     else if (n)
#         FALSE
#     else TRUE
# }
!.External2(.C_getframenumber)


delayedAssign(".has_shINPUT", { .in_shell && .shINFO[["has_input"]] })


from.shell <- function ()
{
    n <- .External2(.C_getframenumber)
    if (is.na(n))
        NA
    else if (n)
        FALSE
    else .has_shINPUT
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2024   Iris Simmons
##


check <- function (path, root, ext, compression = FALSE)
{
    stopifnot(
        identical(splitext(path, compression), matrix(c(character(), root, ext), 2, byrow = TRUE, dimnames = list(c("root", "ext"), NULL))),
        identical(removeext(path, compression), root),
        identical(ext(path, compression), ext)
    )
}


splitext  <- this.path:::.windows_splitext
removeext <- this.path:::.windows_removeext
ext       <- this.path:::.windows_ext


check(character(), character(), character())
check("", "", "")


check("d:", "d:", "")
check("//host/shar.e", "//host/shar.e", "")


check("d:/", "d:/", "")
check("//host/shar.e/", "//host/shar.e/", "")
check("/", "/", "")


check("d:/file.ext", "d:/file", ".ext")
check("//host/share/file.ext", "//host/share/file", ".ext")
check("d:file.ext", "d:file", ".ext")
check("file.ext", "file", ".ext")
check("/file.ext", "/file", ".ext")


check("d:/path/t.o/file", "d:/path/t.o/file", "")
check("//host/shar.e/file", "//host/shar.e/file", "")
check("//host/share/pa.th/to/file", "//host/share/pa.th/to/file", "")
check("/path/t.o/file", "/path/t.o/file", "")
check("path/t.o/file", "path/t.o/file", "")


check("file"    , "file", "")
check("file."   , "file.", "")
check(".file"   , ".file", "")
check("...ext"  , "...ext", "")
check("file.ext", "file", ".ext")


check("file.tar.gz" , "file"  , ".tar.gz" , compression = TRUE)
check("file.tar.bz2", "file"  , ".tar.bz2", compression = TRUE)
check("file.tar.xz" , "file"  , ".tar.xz" , compression = TRUE)
check("file.gz"     , "file"  , ".gz"     , compression = TRUE)
check("file..gz"    , "file." , ".gz"     , compression = TRUE)
check(".file.gz"    , ".file" , ".gz"     , compression = TRUE)
check("...ext.gz"   , "...ext", ".gz"     , compression = TRUE)
check("file.ext.gz" , "file"  , ".ext.gz" , compression = TRUE)


check(letters, letters, character(26))



splitext  <- this.path:::.unix_splitext
removeext <- this.path:::.unix_removeext
ext       <- this.path:::.unix_ext


check(character(), character(), character())
check("", "", "")


check("//host/shar.e", "//host/shar.e", "")


check("//host/shar.e/", "//host/shar.e/", "")
check("/", "/", "")


check("//host/shar.e/file", "//host/shar.e/file", "")
check("/file", "/file", "")


check("file",     "file", "")
check("file.",    "file.", "")
check(".file",    ".file", "")
check("...ext",   "...ext", "")
check("file.ext", "file", ".ext")


check("file.tar.gz" , "file"  , ".tar.gz" , compression = TRUE)
check("file.tar.bz2", "file"  , ".tar.bz2", compression = TRUE)
check("file.tar.xz" , "file"  , ".tar.xz" , compression = TRUE)
check("file.gz"     , "file"  , ".gz"     , compression = TRUE)
check("file..gz"    , "file." , ".gz"     , compression = TRUE)
check(".file.gz"    , ".file" , ".gz"     , compression = TRUE)
check("...ext.gz"   , "...ext", ".gz"     , compression = TRUE)
check("file.ext.gz" , "file"  , ".ext.gz" , compression = TRUE)


check(letters, letters, character(26))


## windows replacement tests ----


ext     <- this.path:::.windows_ext
`ext<-` <- this.path:::`.windows_ext<-`


x   <- c(NA, "", ""    , "C:"  , "//host/share", "C:/" , "/path/to/fi", "/path/to/.", "/path/to/..", "/"   )
val <- c("", NA, ".png", ".png", ".png"        , ".png", ".png"       , ".png"      , ".png"       , ".png")
names(x) <- rev(LETTERS)[seq_along(x)]
ext(x) <- val
stopifnot(identical(x, c(Z = NA_character_, Y = NA_character_, X = "", W = "C:", V = "//host/share", U = "C:/", T = "/path/to/fi.png", S = "/path/to/.", R = "/path/to/..", Q = "/")))


x <- paste0("this.path_2.8.0.tar", c(".gz", ".bz2", ".xz"))
names(x) <- x
ext(x, compression = TRUE) <- ".png"
stopifnot(identical(x, c(this.path_2.8.0.tar.gz = "this.path_2.8.0.png", this.path_2.8.0.tar.bz2 = "this.path_2.8.0.png", this.path_2.8.0.tar.xz = "this.path_2.8.0.png")))


x   <- c("C:/path/to/file", "C:/path/to/.file", "C:/path/to/file.", "C:/path/to/.....", "C:/path/to/file.jpg", "C:/path/to/file.jpg")
val <- c(".png"           , ".png"            , ".png"            , ".png"            , ".png"               , ""                   )
names(x) <- rev(LETTERS)[seq_along(x)]
ext(x) <- val
stopifnot(identical(x, c(Z = "C:/path/to/file.png", Y = "C:/path/to/.file.png", X = "C:/path/to/file..png", W = "C:/path/to/.....", V = "C:/path/to/file.png", U = "C:/path/to/file")))


x <- "a"
ext(x) <- 5
stopifnot(identical(x, "a.5"))


## unix replacement tests ----


ext     <- this.path:::.unix_ext
`ext<-` <- this.path:::`.unix_ext<-`


x <- "C:.jpg"
ext(x) <- ".png"
stopifnot(identical(x, "C:.png"))


x   <- c(NA, "", ""    , "C:"  , "//host/share", "C:/" , "/path/to/fi", "/path/to/.", "/path/to/..", "/"   )
val <- c("", NA, ".png", ".png", ".png"        , ".png", ".png"       , ".png"      , ".png"       , ".png")
names(x) <- rev(LETTERS)[seq_along(x)]
ext(x) <- val
stopifnot(identical(x, c(Z = NA_character_, Y = NA_character_, X = "", W = "C:.png", V = "//host/share", U = "C:.png", T = "/path/to/fi.png", S = "/path/to/.", R = "/path/to/..", Q = "/")))


x <- paste0("this.path_2.8.0.tar", c(".gz", ".bz2", ".xz"))
names(x) <- x
ext(x, compression = TRUE) <- ".png"
stopifnot(identical(x, c(this.path_2.8.0.tar.gz = "this.path_2.8.0.png", this.path_2.8.0.tar.bz2 = "this.path_2.8.0.png", this.path_2.8.0.tar.xz = "this.path_2.8.0.png")))


x   <- c("/path/to/file", "/path/to/.file", "/path/to/file.", "/path/to/.....", "/path/to/file.jpg", "/path/to/file.jpg")
val <- c(".png"         , ".png"          , ".png"          , ".png"          , ".png"             , ""                 )
names(x) <- rev(LETTERS)[seq_along(x)]
ext(x) <- val
stopifnot(identical(x, c(Z = "/path/to/file.png", Y = "/path/to/.file.png", X = "/path/to/file..png", W = "/path/to/.....", V = "/path/to/file.png", U = "/path/to/file")))


x <- "a"
ext(x) <- 5
stopifnot(identical(x, "a.5"))

##
## this.path : Get Executing Script's Path
## Copyright (C) 2024   Iris Simmons
##


.faster_subsequent_times_test <- function ()
{
    first_time <- microbenchmark::microbenchmark(
        `first time` = .External2(.C_sys_path),
        times = 1
    )
    subsequent <- microbenchmark::microbenchmark(
        subsequent = .External2(.C_sys_path),
        times = 100
    )
    rbind(first_time, subsequent)
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2024   Iris Simmons
##


.PrintValueEnv <- function (x, envir = parent.frame())
.External2(.C_PrintValueEnv, x, envir)


.maybePrintValueEnv <- function (x, envir = parent.frame())
{
    if (withVisible(x)$visible)
        .External2(.C_PrintValueEnv, x, envir)
    else invisible(x)
}


print.ThisPathDocumentContext <- function (x, ..., quote = TRUE)
.External2(.C_print.ThisPathDocumentContext, x, quote)


if (getRversion() >= "3.5.0") {


format.ThisPathDocumentContext <- function (x, ...)
{
    value <- NULL
    conn <- textConnection("value", "w", local = TRUE)
    on.exit(close(conn))
    sink(conn)
    on.exit(sink(), add = TRUE, after = FALSE)
    print.ThisPathDocumentContext(x = x, ...)
    on.exit()
    sink()
    close(conn)
    value
}


} else {


## on.exit(after = TRUE) was added in R 3.5.0
format.ThisPathDocumentContext <- function (x, ...)
{
    value <- NULL
    conn <- textConnection("value", "w", local = TRUE)
    sink(conn)
    on.exit(sink())
    on.exit(close(conn), add = TRUE)
    print.ThisPathDocumentContext(x = x, ...)
    on.exit()
    sink()
    close(conn)
    value
}


}


as.character.ThisPathDocumentContext <- function (x, ...)
paste(format.ThisPathDocumentContext(x = x, ...), collapse = "\n")

##
## this.path : Get Executing Script's Path
## Copyright (C) 2024   Iris Simmons
##


.toplevel.nframe <- function ()
{
    if (.GUI_jupyter) {
        if (.is_jupyter_loaded())
            sys.frame(1L)$kernel$executor$nframe + 1L
        else 0L
    }
    else 0L
}


sys.srcref <- function (n = 1L, which = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    .External2(.C_sys.srcref, which)
}


sys.whiches <- function (n = 1L, which = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    .External2(.C_sys.whiches, which)
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2021-2024   Iris Simmons
##


if (getRversion() >= "3.4.0") {
    sys.source(
        "./inst/extdata/main.R", environment(),
        keep.source = !is.null(attr(function() NULL, "srcref")),
        toplevel.env = getOption("topLevelEnvironment")
    )
} else {
    sys.source(
        "./inst/extdata/main.R", environment(),
        keep.source = !is.null(attr(function() NULL, "srcref"))
    )
}


tmp <- evalq(envir = new.env(), {
    `.Platform$OS.type == "windows"` <- quote(.Platform$OS.type == "windows")
    .OS_windows <- quote(.OS_windows)
    `identical(R.version[["crt"]], "ucrt")` <- quote(identical(R.version[["crt"]], "ucrt"))
    .ucrt <- quote(.ucrt)
    `!is.character(LANGUAGE) || length(LANGUAGE) != 1L` <- quote(!is.character(LANGUAGE) || length(LANGUAGE) != 1L)
    `!.IS_SCALAR_STR(LANGUAGE)` = quote(!.IS_SCALAR_STR(LANGUAGE))
function (expr)
{
    if (typeof(expr) == "closure") {
        at <- attributes(expr)
        formals(expr) <- tmp(formals(expr))
        body(expr) <- tmp(body(expr))
        if (!is.null(at))
            attributes(expr) <- at
        expr
    }
    else if (is.pairlist(expr)) {
        as.pairlist(lapply(expr, tmp))
    }
    else if (is.call(expr)) {
        if (identical(expr, `.Platform$OS.type == "windows"`))
            .OS_windows
        else if (identical(expr, `identical(R.version[["crt"]], "ucrt")`))
            .ucrt
        else if (identical(expr, `!is.character(LANGUAGE) || length(LANGUAGE) != 1L`))
            `!.IS_SCALAR_STR(LANGUAGE)`
        else as.call(lapply(expr, tmp))
    }
    else expr
}
})


.language_envvars <- tmp(.language_envvars)
Sys.putenv <- tmp(Sys.putenv)


rm(tmp)

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2024   Iris Simmons
##


.Rscript <- function (options = NULL, trailing = character(), dry.run = FALSE,
    show.command = TRUE, intern = TRUE, show.output.on.console = show.command,
    ...)
{
    command <- path.join(
        R.home("bin"),
        if (.OS_windows) "Rscript.exe" else "Rscript"
    )
    args <- c(command, options)
    args <- c(shQuote(args), trailing)
    command <- paste(args, collapse = " ")
    if (dry.run)
        return(command)
    if (show.command)
        cat(command, "\n", sep = "")
    value <- suppressWarnings(system(command = command, intern = intern, ...))
    if (show.output.on.console)
        writeLines(value)
    invisible(value)
}


.build.this <- function (chdir = FALSE, file = here(n = 1L), which = "tar")
{
    ## .build.this               package:this.path               R Documentation
    ##
    ## Building Packages
    ##
    ##
    ##
    ## Description:
    ##
    ## This provides a more general method of making packages, not specifically
    ## R packages, for distribution and version control.
    ##
    ##
    ##
    ## Usage:
    ##
    ## .build.this(chdir = FALSE, file = here(n = 1), which = "tar")
    ##
    ##
    ##
    ## Arguments:
    ##
    ## chdir
    ##
    ##     TRUE or FALSE; change directory to directory of package source
    ##     before building tar/zip archives? This will not affect the build
    ##     process, only the location in which the archive is made.
    ##
    ## file
    ##
    ##     name of a directory which is to be packaged
    ##
    ## which
    ##
    ##     which type of archive do you want to make, "tar" or "zip"?
    ##     Can be both i.e. 'which = c("tar", "zip")'
    ##
    ##
    ##
    ## Details:
    ##
    ## .build.this builds package archives similar to 'R CMD build', and takes
    ## the name of the package and its version from the 'DESCRIPTION' file.
    ## If you are unfamiliar with the 'DESCRIPTION' file, here is a outline:
    ##
    ## Package: example
    ## Version: 0.1.0
    ## Date: %Y-%m-%d
    ## License: What license is it under?
    ## Title: What the Package Does (Title Case)
    ## Description: More about what it does (maybe more than one line)
    ##     Use four spaces when indenting paragraphs within the Description.
    ## Author: Who wrote it
    ## Maintainer: The package maintainer <yourself@somewhere.net>
    ##
    ## The package name should start with a letter, end with a letter or
    ## number, and contain any combination of letters, numbers, underscores,
    ## periods, and hyphens in the middle.
    ##
    ## The package version should be two or more numbers separated by hyphens
    ## and/or periods like '12.34.56-78'.



    ## determine which types of archive we want to make
    choices <- c("tar", "zip")
    nms <- c("tarball", "zip archive")
    which <- unique(match.arg(which, choices, several.ok = TRUE))
    nms <- nms[match(which, choices)]


    ## check that 'file' is valid, and 'chdir' if required
    if (!.IS_SCALAR_STR(file)) {
        stop(gettextf("'%s' must be a character string", "file", domain = "R"), domain = NA)
    } else if (grepl("^(ftp|ftps|http|https)://", file, useBytes = TRUE)) {
        stop("cannot '.build.this' on a URL")
    } else if (chdir && (path <- file) != ".") {
        file <- "."
        owd <- getwd()
        if (is.null(owd))
            stop("cannot 'chdir' as current directory is unknown")
        on.exit(setwd(owd))
        setwd(path)
    }


    # op <- options(error = quote({
    #     quit(save = "no", status = 1, runLast = FALSE)
    # }))
    # on.exit(options(op), add = TRUE)


    ## look for the appropriate applications to create archives
    ## if all are unavailable, throw an error
    ## if some are unavailable, throw a warning and continue with the rest
    val <- Sys.which(which)
    i <- is.na(val) | val == ""
    if (any(i)) {
        msgs <- paste0(which, " unavailable; cannot create ", nms)
        if (all(i)) {
            stop(paste(msgs, collapse = "\n "))
        } else warning(paste(msgs[i], collapse = "\n"))
        which <- which[!i]
        nms <- nms[!i]
    }


    ## check for file "DESCRIPTION"
    DESCRIPTION <- path.join(file, "DESCRIPTION")
    cat("* checking for file '", DESCRIPTION, "' ... ", sep = "")
    if (!file.exists(DESCRIPTION)) {
        cat("NO\n\n")
        stop("cannot create ", paste(nms, collapse = ", "),
            " without file '", DESCRIPTION, "'")
    }
    cat("OK\n")


    packageInfo <- read.dcf(DESCRIPTION,
        fields = c("Package", "Version"))
    pkgname <- packageInfo[[1L, "Package"]]
    version <- packageInfo[[1L, "Version"]]


    ## check that the package name and version exist in "DESCRIPTION"
    problems <- c(
        if (is.na(pkgname))
            "invalid 'Package' field\n\n",
        if (is.na(version))
            "invalid 'Version' field\n\n"
    )
    if (!is.null(problems)) {
        cat(" ERROR\n", problems, sep = "")
        stop("invalid package DESCRIPTION file")
    }


    ## check that the package name and version are valid names and versions
    cat("* checking DESCRIPTION meta-information ... ", sep = "")
    valid_package_name <- "([[:alpha:]][[:alnum:]_.-]*[[:alnum:]])"
    valid_package_version <- "(([[:digit:]]+[.-]){1,}[[:digit:]]+)"
    problems <- c(
        if (!grepl(paste0("^", valid_package_name   , "$"), pkgname))
            "Malformed package name\n\n",
        if (!grepl(paste0("^", valid_package_version, "$"), version))
            "Malformed package version.\n\n"
    )
    if (!is.null(problems)) {
        cat("ERROR\n", problems, sep = "")
        stop("invalid package DESCRIPTION file")
    }
    cat("OK\n")


    ## files to exclude, and all files
    exclude <- NULL
    files <- list.files(path = file, all.files = TRUE, recursive = TRUE, include.dirs = TRUE)


    ## directories to always exclude
    i <- basename2(files) %in% c(

        ## directories from source control systems
        "CSV", ".svn", ".arch-ids", ".bzr", ".git", ".hg",

        ## directories from eclipse
        ".metadata",

        "check", "chm"
    )


    ## more directories to exclude, ending with Old or old
    i <- i | grepl("(Old|old)$", files)


    ## and, of course, are directories
    i <- i & dir.exists(path.join(file, files))


    if (any(i)) {
        exclude <- c(exclude, files[i])
        files <- files[!i]
    }


    ## files to always exclude
    i <- grepl(
        pattern = paste0(
            "(",
            c(
                "^GNUmakefile$",
                "^Read-and-delete-me$",

                ## starts with .#
                "^\\.#",

                ## starts and ends with #
                "^#.*#$",

                ## ends with ~ or .bak or .swp
                "~$", "\\.bak$", "\\.swp$"
            ),
            ")", collapse = "|"
        ),
        basename2(files)
    )


    if (any(i)) {
        exclude <- c(exclude, files[i])
        files <- files[!i]
    }


    ## do not put previous archives in the new archives
    prev_build_patterns <- paste0(
        "^",
        gsub(".", "\\.", pkgname, fixed = TRUE),
        "_",
        valid_package_version,
        c("\\.tar\\.gz", "\\.tgz", "\\.zip"),
        "$"
    )


    for (exclude_pattern in prev_build_patterns) {
        if (any(i <- grepl(exclude_pattern, files))) {
            exclude <- c(exclude, files[i])
            files <- files[!i]
        }
    }


    ## look for a ".buildignore" file containing
    ## a list of Perl patterns (case insensitive)
    ## specifying files to ignore when archiving
    ignore_file <- path.join(file, ".buildignore")
    if (file.exists(ignore_file)) {
        for (exclude_pattern in readLines(ignore_file, warn = FALSE, encoding = "UTF-8")) {
            if (any(i <- grepl(exclude_pattern, files, ignore.case = TRUE, perl = TRUE))) {
                exclude <- c(exclude, files[i])
                files <- files[!i]
            }
        }
    }


    ## for directories in 'exclude', also
    ## exclude the files within said directories
    exclude_dirs <- exclude[dir.exists(path.join(file, exclude))]
    for (exclude_prefix in paste0(exclude_dirs, "/", recycle0 = TRUE)) {
        if (any(i <- startsWith(files, exclude_prefix))) {
            exclude <- c(exclude, files[i])
            files <- files[!i]
        }
    }


    ## create a new directory to hold the temporary files and archives
    dir.create(my_tmpdir <- tempfile("dir"))
    on.exit(unlink(my_tmpdir, recursive = TRUE, force = TRUE, expand = FALSE), add = TRUE)


    ## create another directory to hold the temporary files
    dir.create(pkgdir <- path.join(my_tmpdir, pkgname))


    ## within said directory, make the appropriate sub-directories
    isdir <- dir.exists(path.join(file, files))
    dirs <- files[isdir]
    for (path in path.join(pkgdir, dirs))
        dir.create(path, showWarnings = TRUE, recursive = TRUE)


    ## fill the directory and sub-directories with their files,
    ## while maintaining file modify time
    if (any(i <- !file.copy(
        path.join(file  , files[!isdir]),
        path.join(pkgdir, files[!isdir]),
        copy.date = TRUE
    )))
        stop(ngettext(sum(i), "unable to copy file: ",
                              "unable to copy files:\n  "),
             paste(encodeString(files[!isdir][i], quote = "\""), collapse = "\n  "))


    ## set the modify time of the sub-directories to their original values
    Sys.setFileTime(
                   path.join(pkgdir, dirs),
        file.mtime(path.join(file  , dirs))
    )


    for (apk in which) {
        switch(apk, tar = {
            build_name <- paste0(pkgname, "_", version, ".tar.gz")
            build_path <- path.join(my_tmpdir, build_name)
            args <- c("tar", "-czf", shQuote(build_path), "-C", shQuote(my_tmpdir), shQuote(pkgname))
            command <- paste(args, collapse = " ")
            cat("* building '", build_name, "'\n", sep = "")
            res <- system(command, ignore.stdout = TRUE)
            if (res == -1L) {
                stop("'", command, "' could not be run")
            } else if (res) {
                stop("'", command, "' execution failed with error code ", res)
            } else if (i <- !file.copy(
                build_path,
                path.join(file, build_name),
                overwrite = TRUE
            )) {
                stop("unable to move:\n  ",
                     encodeString(build_path, quote = "\""),
                     "\nto:\n  ",
                     encodeString(path.join(file, build_name), quote = "\""))
            }
        }, zip = {
            build_name <- paste0(pkgname, "_", version, ".zip")
            build_path <- path.join(my_tmpdir, build_name)
            args <- c("zip", "-r", shQuote(build_path), shQuote(pkgname))
            command <- paste(args, collapse = " ")
            cat("* building '", build_name, "'\n", sep = "")
            res <- .with_chdir(my_tmpdir, {
                system(command, ignore.stdout = TRUE)
            })
            if (res == -1L) {
                stop("'", command, "' could not be run")
            } else if (res) {
                stop("'", command, "' execution failed with error code ", res)
            } else if (i <- !file.copy(
                build_path,
                path.join(file, build_name),
                overwrite = TRUE
            )) {
                stop("unable to move:\n  ",
                     encodeString(build_path, quote = "\""),
                     "\nto:\n  ",
                     encodeString(path.join(file, build_name), quote = "\""))
            }
        }, stop("invalid 'which'; should not happen, please report!"))
    }
}


.with_chdir <- function (wd, expr)
{
    owd <- getwd()
    if (is.null(owd))
        stop("cannot 'chdir' as current directory is unknown", domain = "R-base")
    on.exit(setwd(owd))
    setwd(wd)
    expr
}


.R_BraceSymbol <- as.symbol("{")


.maybeQuote <- function (expr, evaluated, simplify.brace = TRUE)
{
    ## possibly quote an expression
    ##
    ## when 'evaluated' is missing, 'expr' is quoted only if it is an
    ## (unevaluated) call to `{`
    ##
    ## when 'evaluated' is FALSE, 'expr' is always quoted
    ##
    ## when 'evaluated' is TRUE, 'expr' is evaluated as normal
    ##
    ## this is intended to be called inside another function, like match.fun()


    if (missing(expr))
        expr


    if (missing(evaluated)) {
        if (is.call(e <- eval(as.call(list(substitute, substitute(expr))), parent.frame())) &&
            identical(e[[1L]], .R_BraceSymbol))
        {
            expr <- e
        }
    }
    else if (!evaluated)
        expr <- eval(as.call(list(substitute, substitute(expr))), parent.frame())


    ## force the evaluated of 'expr'
    else expr


    if (is.call(expr)) {
        ## always turn a call into an expression so that it
        ## cannot be misinterpreted later
        if (simplify.brace) {
            if (identical(expr[[1L]], .R_BraceSymbol))
                as.expression(as.list(expr[-1L]))
            else as.expression(list(expr))
        }
        else as.expression(list(expr))
    }
    else expr
}


.code2character <- function (x, width.cutoff = 60L,
    deparseCtrl = c("keepInteger", "showAttributes", "useSource", "keepNA", "digits17"))
{
    ## if 'x' is an (unevaluated) call to `{`
    ## deparse each sub-expression
    ##
    ## if 'x' is an expression() vector
    ## make a list of deparsed expressions
    ##
    ## if 'x' is a language object
    ## deparse the expression
    ##
    ## if 'x' is a character vector
    ## leave as is


    fun <- function(xx) {
        deparse1(xx, collapse = "\n", width.cutoff = width.cutoff,
            backtick = TRUE, control = deparseCtrl)
    }


    if (is.expression(x)) {
        if ("useSource" %in% deparseCtrl) {
            getLines <- function(srcref) {
                srcfile <- attr(srcref, "srcfile", exact = TRUE)
                getSrcLines(srcfile, srcref[7L], srcref[8L])
            }
            trimLines <- function(lines, srcref) {
                if (length(lines)) {
                    enc <- Encoding(lines)
                    Encoding(lines) <- "bytes"
                    if (length(lines) > srcref[8L] - srcref[7L])
                        lines[length(lines)] <- substring(lines[length(lines)], 1L, srcref[4L])
                    lines[1L] <- substring(lines[1L], srcref[2L])
                    Encoding(lines) <- enc
                }
                lines
            }
            # x <- parse(text = "  5 +\n 6\t; 7 +8  ;  ", keep.source = TRUE)
            srcref <- attr(x, "wholeSrcref", exact = TRUE)
            if (!is.null(srcref))
                tryCatch2({
                    lines <- getLines(srcref)
                }, error = function(e) {
                }, else. = {
                    lines <- trimLines(lines, srcref)
                    return(paste(lines, collapse = "\n"))
                })
            srcrefs <- attr(x, "srcref", exact = TRUE)
            if (!is.null(srcrefs)) {
                y <- character(length(srcrefs))
                for (i in seq_along(y)) {
                    srcref <- srcrefs[[i]]
                    y[[i]] <- if (!is.null(srcref))
                        tryCatch2({
                            lines <- getLines(srcref)
                        }, error = function(e) {
                            fun(x[[i]])
                        }, else. = {
                            lines <- trimLines(lines, srcref)
                            paste(lines, collapse = "\n")
                        })
                    else fun(x[[i]])
                }
                y
            }
            else vapply(x, fun, "")
        }
        else vapply(x, fun, "")
    }
    else if (is.language(x))
        fun(x)
    else if (is.character(x))
        x
    else stop("invalid 'x', expected a language object or character vector")
}


.writeCode <- function (x, file = stdout(), evaluated, simplify.brace = TRUE,
    width.cutoff = 60L, deparseCtrl = c("keepInteger", "showAttributes", "useSource", "keepNA", "digits17"))
{
    x <- .maybeQuote(x, evaluated, simplify.brace)
    x <- .code2character(x, width.cutoff, deparseCtrl)
    if (is.null(file))
        x
    else {
        writeLines(x, file, useBytes = !.utf8locale)
        invisible(x)
    }
}


if (getRversion() < "3.2.0") {


    tmp <- formals(.code2character)[["deparseCtrl"]]
    formals(.code2character)[["deparseCtrl"]] <- tmp[!vapply(tmp, identical, "digits17", FUN.VALUE = NA)]


    tmp <- formals(.writeCode)[["deparseCtrl"]]
    formals(.writeCode)[["deparseCtrl"]] <- tmp[!vapply(tmp, identical, "digits17", FUN.VALUE = NA)]


    rm(tmp)
}


.readFiles <- function (files)
vapply(files, function(file) paste0(readLines(file), "\n", collapse = ""), "")


.envvars <- function (...)
{
    args <- list(...)
    if (length(args) == 0L)
        as.list(Sys.getenv())
    else {
        visible <- FALSE
        if (length(args) == 1L &&
            typeof(args[[1L]]) %in% c("NULL", "pairlist", "language",
                "...", "list", "expression") &&
            is.null(names(args)))
        {
            args <- args[[1L]]
            if (typeof(args) == "...") {
                # ... <- args
                assign("...", args)
                args <- list(...)
            }
            else if (!is.vector(args) || is.object(args))
                args <- as.list(args)
        }
        if (!length(args))
            value <- structure(list(), names = character())
        else if (is.null(tags <- names(args)) ||
                 all(i <- !nzchar(tags)))
        {
            visible <- TRUE
            tags <- vapply(args, .AS_SCALAR_STR, "", USE.NAMES = FALSE)
            value <- as.list(Sys.getenv(tags, unset = NA, names = TRUE))
        }
        else {
            args <- lapply(args, .AS_SCALAR_STR)
            if (any(i)) {
                visible <- TRUE
                tags[i] <- names(args)[i] <- as.character(args[i])
                # args[i] <- list(NULL)
            }
            tags[is.na(tags)] <- "NA"
            if (.Platform$OS.type == "windows")
                tags <- tolower(tags)
            i <- !i
            base_case <- function(args, i) {
                # print(args)
                # cat("\n\n\n\n\n")
                value <- as.list(Sys.getenv(names(args), unset = NA, names = TRUE))
                args <- args[i]
                na <- is.na(args)
                if (any(na)) {
                    Sys.unsetenv(names(args)[na])
                    args <- args[!na]
                }
                if (length(args))
                    do.call("Sys.setenv", args)
                value
            }
            recurse <- function(args, tags, i) {
                if (first_setter <- match(TRUE, i, 0L)) {
                    n <- length(args)
                    if (dup <- anyDuplicated(tags[first_setter:n])) {
                        dup <- dup - 1L + first_setter
                        indxs <- seq_len(dup - 1L)
                        value <- base_case(args[indxs], i[indxs])
                        indxs <- dup:n
                        c(value, recurse(args[indxs], tags[indxs], i[indxs]))
                    }
                    else base_case(args, i)
                }
                else as.list(Sys.getenv(names(args), unset = NA, names = TRUE))
            }
            value <- recurse(args, tags, i)
        }
        if (visible)
            value
        else invisible(value)
    }
}


.istrue <- function (x)
.External2(.C_istrue, x)


.isfalse <- function (x)
.External2(.C_isfalse, x)


.asLogical <- function (x)
.External2(.C_asLogical, x)


.asInteger <- function (x)
.External2(.C_asInteger, x)


.asIntegerGE0 <- function (x)
.External2(.C_asIntegerGE0, x)


.IS_SCALAR_STR <- function (x)
.External2(.C_IS_SCALAR_STR, x)


.AS_SCALAR_STR <- function (x)
{
    switch (typeof(x),
    logical = ,
    integer = ,
    double = ,
    complex = ,
    character = ,
    raw = {
        if (is.object(x))
            x <- unclass(x)
        if (length(x))
            as.character(x[[1L]])
        else NA_character_
    },
    char = as.character(list(x)),
    symbol = as.character(x),
    NA_character_)
}


.AS_SCALAR_STR <- function (x)
.External2(.C_AS_SCALAR_STR, x)


.scalar_streql <- function (e1, e2)
.External2(.C_scalar_streql, e1, e2)


.identical <- if (getRversion() >= "4.2.0") {


              function (x, y)
identical(x, y, num.eq = FALSE, single.NA = FALSE, attrib.as.set = FALSE,
    ignore.bytecode = FALSE, ignore.environment = FALSE, ignore.srcref = FALSE,
    extptr.as.ref = TRUE)


} else if (getRversion() >= "3.4.0") {


              function (x, y)
identical(x, y, num.eq = FALSE, single.NA = FALSE, attrib.as.set = FALSE,
    ignore.bytecode = FALSE, ignore.environment = FALSE, ignore.srcref = FALSE)


} else if (getRversion() >= "3.0.0") {


              function (x, y)
identical(x, y, num.eq = FALSE, single.NA = FALSE, attrib.as.set = FALSE,
    ignore.bytecode = FALSE, ignore.environment = FALSE)


} else if (getRversion() >= "2.14.0") {


              function (x, y)
identical(x, y, num.eq = FALSE, single.NA = FALSE, attrib.as.set = FALSE,
    ignore.bytecode = FALSE)


} else if (getRversion() >= "2.10.0") {


              function (x, y)
identical(x, y, num.eq = FALSE, single.NA = FALSE, attrib.as.set = FALSE)


} else {


              function (x, y)
identical(x, y)


}


.source <- function (file, local = FALSE, echo = verbose, print.eval = echo,
    exprs, spaced = use_file, verbose = getOption("verbose"),
    prompt.echo = getOption("prompt"), max.deparse.length = 150,
    width.cutoff = 60L, deparseCtrl = "showAttributes", chdir = FALSE,
    encoding = getOption("encoding"), continue.echo = getOption("continue"),
    skip.echo = 0, keep.source = getOption("keep.source"))
{
    file <- set.sys.path(file)
    envir <- if (isTRUE(local))
        parent.frame()
    else if (isFALSE(local))
        .GlobalEnv
    else if (is.environment(local))
        local
    else stop("'local' must be TRUE, FALSE or an environment")
    if (!missing(echo)) {
        if (!is.logical(echo))
            stop("'echo' must be logical")
        if (!echo && verbose) {
            warning("'verbose' is TRUE, 'echo' not; ... coercing 'echo <- TRUE'")
            echo <- TRUE
        }
    }
    if (verbose) {
        cat("'envir' chosen:")
        print(envir)
    }
    if (use_file <- missing(exprs)) {
        ofile <- file
        from_file <- FALSE
        srcfile <- NULL
        if (is.character(file)) {
            if (!length(file) || file == "")
                stop("empty file/url name")
            have_encoding <- !missing(encoding) && !identical(encoding, "unknown")
            if (identical(encoding, "unknown")) {
                enc <- utils::localeToCharset()
                encoding <- enc[length(enc)]
            }
            else enc <- encoding
            if (length(enc) > 1L) {
                encoding <- NA
                owarn <- options(warn = 2)
                for (e in enc) {
                  if (is.na(e))
                    next
                  zz <- file(file, encoding = e)
                  res <- tryCatch(readLines(zz, warn = FALSE),
                    error = identity)
                  close(zz)
                  if (!inherits(res, "error")) {
                    encoding <- e
                    break
                  }
                }
                options(owarn)
            }
            if (is.na(encoding))
                stop("unable to find a plausible encoding")
            if (verbose)
                cat(gettextf("encoding = \"%s\" chosen", encoding),
                  "\n", sep = "")
            {
                filename <- file
                file <- file(filename, "r", encoding = encoding)
                on.exit(close(file))
                if (isTRUE(keep.source)) {
                  lines <- readLines(file, warn = FALSE)
                  on.exit()
                  close(file)
                  srcfile <- srcfilecopy(filename, lines, file.mtime(filename)[1],
                    isFile = .is_abs_path(filename))
                }
                else {
                  from_file <- TRUE
                  srcfile <- filename
                }
                loc <- utils::localeToCharset()[1L]
                encoding <- if (have_encoding)
                  switch(loc, `UTF-8` = "UTF-8", `ISO8859-1` = "latin1",
                    "unknown")
                else "unknown"
            }
        }
        else {
            lines <- readLines(file, warn = FALSE)
            srcfile <- if (isTRUE(keep.source))
                srcfilecopy(deparse1(substitute(file), ""), lines)
            else deparse1(substitute(file), "")
        }
        if (verbose) {
            cat(sprintf(" --> from_file='%s'\n lines:", from_file))
            utils::str(lines)
        }
        exprs <- if (!from_file) {
            if (length(lines) && is.character(lines))
                parse(file = stdin(), n = -1, text = lines, keep.source = FALSE, srcfile = srcfile, encoding = encoding)
            else expression()
        }
        else parse(file = file, n = -1, keep.source = FALSE, srcfile = srcfile, encoding = encoding)
        on.exit()
        if (from_file)
            close(file)
        if (verbose)
            cat("--> parsed", length(exprs), "expressions; now eval(.)ing them:\n")
        if (chdir) {
            if (is.character(ofile)) {
                if (grepl("^(ftp|ftps|http|https|file)://", ofile, useBytes = TRUE))
                  warning("'chdir = TRUE' makes no sense for a URL")
                else if ((path <- dirname(ofile)) != ".") {
                  owd <- getwd()
                  if (is.null(owd))
                    stop("cannot 'chdir' as current directory is unknown")
                  on.exit(setwd(owd), add = TRUE)
                  setwd(path)
                }
            }
            else {
                warning("'chdir = TRUE' makes no sense for a connection")
            }
        }
    }
    else {
        if (!missing(file))
            stop("specify either 'file' or 'exprs' but not both")
        if (!is.expression(exprs))
            exprs <- as.expression(exprs)
    }
    Ne <- length(exprs)
    if (echo) {
        sd <- "\""
        nos <- "[^\"]*"
        oddsd <- paste0("^", nos, sd, "(", nos, sd, nos, sd,
            ")*", nos, "$")
        trySrcLines <- function(srcfile, showfrom, showto) {
            tryCatch(suppressWarnings(getSrcLines(srcfile, showfrom,
                showto)), error = function(e) character())
        }
    }
    yy <- NULL
    lastshown <- 0
    srcrefs <- attr(exprs, "srcref", exact = TRUE)
    if (verbose && !is.null(srcrefs)) {
        cat("has srcrefs:\n")
        utils::str(srcrefs)
    }
    for (i in seq_len(Ne + echo)) {
        tail <- i > Ne
        if (!tail) {
            if (verbose)
                cat("\n>>>> eval(expression_nr.", i, ")\n\t\t =================\n")
            ei <- exprs[i]
        }
        if (echo) {
            nd <- 0
            srcref <- if (tail)
                attr(exprs, "wholeSrcref", exact = TRUE)
            else if (i <= length(srcrefs))
                srcrefs[[i]]
            if (!is.null(srcref)) {
                if (i == 1)
                  lastshown <- min(skip.echo, srcref[8L] - 1)
                if (lastshown < srcref[8L]) {
                  srcfile <- attr(srcref, "srcfile", exact = TRUE)
                  dep <- trySrcLines(srcfile, lastshown + 1,
                    srcref[8L])
                  if (length(dep)) {
                    leading <- if (tail)
                      length(dep)
                    else srcref[7L] - lastshown
                    lastshown <- srcref[8L]
                    while (length(dep) && grepl("^[[:blank:]]*$",
                      dep[1L])) {
                      dep <- dep[-1L]
                      leading <- leading - 1L
                    }
                    dep <- paste0(rep.int(c(prompt.echo, continue.echo),
                      pmax(0L, c(leading, length(dep) - leading))),
                      dep, collapse = "\n")
                    nd <- nchar(dep, "c")
                  }
                  else srcref <- NULL
                }
            }
            if (is.null(srcref)) {
                if (!tail) {
                  dep <- substr(paste(deparse(ei, width.cutoff = width.cutoff,
                    control = deparseCtrl), collapse = "\n"),
                    12L, 1000000L)
                  dep <- paste0(prompt.echo, gsub("\n", paste0("\n",
                    continue.echo), dep, fixed = TRUE))
                  nd <- nchar(dep, "c") - 1L
                }
            }
            if (nd) {
                do.trunc <- nd > max.deparse.length
                dep <- substr(dep, 1L, if (do.trunc)
                  max.deparse.length
                else nd)
                cat(if (spaced)
                  "\n", dep, if (do.trunc)
                  paste(if (grepl(sd, dep) && grepl(oddsd, dep))
                    " ...\" ..."
                  else " ....", "[TRUNCATED] "), "\n", sep = "")
            }
        }
        if (!tail) {
            yy <- withVisible(eval(ei, envir))
            i.symbol <- mode(ei[[1L]]) == "name"
            if (!i.symbol) {
                curr.fun <- ei[[1L]][[1L]]
                if (verbose) {
                  cat("curr.fun:")
                  utils::str(curr.fun)
                }
            }
            if (verbose >= 2) {
                cat(".... mode(ei[[1L]])=", mode(ei[[1L]]), "; paste(curr.fun)=")
                utils::str(paste(curr.fun))
            }
            if (print.eval && yy$visible)
                .PrintValueEnv(yy$value, envir)
            if (verbose)
                cat(" .. after ", sQuote(deparse(ei, control = unique(c(deparseCtrl,
                  "useSource")))), "\n", sep = "")
        }
    }
    invisible(yy)
}


source.exprs <- function (exprs, evaluated = FALSE, envir = parent.frame(), echo = TRUE, print.eval = TRUE)
{
    if (!evaluated) {
        exprs <- substitute(exprs)
        if (is.call(exprs)) {
            if (identical(exprs[[1]], .R_BraceSymbol))
                exprs <- as.list(exprs)[-1]
        }
    }
    if (!is.expression(exprs))
        exprs <- as.expression(exprs)
    Ne <- length(exprs)
    if (echo)
        trySrcLines <- function(srcfile, from, to) {
            tryCatch(suppressWarnings(getSrcLines(srcfile, from, to)),
                error = function(e) character())
        }
    yy <- NULL
    lastshown <- 0L
    srcrefs <- attr(exprs, "srcref")
    for (i in seq_len(Ne + echo)) {
        tail <- i > Ne
        if (!tail)
            ei <- exprs[i]
        if (echo) {
            srcref <- if (tail)
                attr(exprs, "wholeSrcref")
            else if (i <= length(srcrefs))
                srcrefs[[i]]
            if (!is.null(srcref)) {
                firstl <- srcref[7L]
                lastl <- srcref[8L]
                if (i == 1L)
                    lastshown <- firstl - 1L
                if (lastshown < lastl) {
                    srcfile <- attr(srcref, "srcfile")
                    dep <- trySrcLines(srcfile, lastshown + 1L, lastl)
                    if (length(dep)) {
                        leading <- if (tail)
                            length(dep)
                        else firstl - lastshown
                        lastshown <- lastl
                        dep <- paste0(rep.int(c(getOption("prompt"), getOption("continue")),
                            pmax(0L, c(         leading            , length(dep) - leading))),
                            dep, collapse = "\n")
                        cat(dep, "\n", sep = "")
                    }
                }
            }
        }
        if (!tail) {
            yy <- withVisible(eval(ei, envir))
            if (print.eval && yy$visible)
                .PrintValueEnv(yy$value, envir)
        }
    }
    invisible(yy)
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2024-2025   Iris Simmons
##


## Defunct in <pkgname> <version> (<date>)


# <old> <- function (<formals>)
# stop(.defunctError("<new>", "this.path", old = "<old>"))

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2026   Iris Simmons
##


.site_file <- evalq(envir = new.env(), {
    delayedAssign("file_info", {
        local({
            filename <- local({
                ## https://github.com/wch/r-source/blob/trunk/src/main/startup.c#L98
                p <- Sys.getenv("R_PROFILE", NA_character_)
                if (!is.na(p)) {
                    if (nzchar(p)) return(p)
                    else return(NA_character_)
                }
                if (nzchar(p <- .Platform$r_arch)) {
                    p <- sprintf("%s/etc/%s/Rprofile.site", R.home(), p)
                    if (file.exists(p)) return(p)
                }
                p <- sprintf("%s/etc/Rprofile.site", R.home())
                return(p)
            })
            list(filename = filename, wd = getwd(), exists = file.exists(filename))
        })
    })
    delayedAssign("ofile", {
        if (isTRUE(.shINFO[["no_site_file"]]))
            NA_character_
        else if (file_info[["exists"]])
            file_info[["filename"]]
        else NA_character_
    })
    delayedAssign("file", {
        if (is.na(ofile))
            .normalizePath(ofile)
        else if (is.null(file_info[["wd"]]))
            .normalizePath(ofile)
        else .normalizePath_against(file_info[["wd"]], ofile)
    })
              function (original = TRUE, for.msg = FALSE)
.External2(.C_site_file, original, for.msg)
})


.init_file <- evalq(envir = new.env(), {
    delayedAssign("file_info", {
        local({
            filename <- local({
                if (.OS_windows) {
                    ## https://github.com/wch/r-source/blob/trunk/src/gnuwin32/sys-win32.c#L47
                    p <- Sys.getenv("R_PROFILE_USER", NA_character_)
                    if (!is.na(p)) {
                        if (!nzchar(p)) return(NA_character_)
                        return(p)
                    }
                    if (file.exists(p <- ".Rprofile"))
                        return(p)
                    p <- Sys.getenv("R_USER", NA_character_)
                    if (!is.na(p)) {
                        p <- sprintf("%s/.Rprofile", p)
                    }
                    return(p)
                } else {
                    ## https://github.com/wch/r-source/blob/trunk/src/unix/sys-unix.c#L68
                    p <- Sys.getenv("R_PROFILE_USER", NA_character_)
                    if (!is.na(p)) {
                        if (!nzchar(p)) return(NA_character_)
                        return(p)
                    }
                    if (file.exists(p <- ".Rprofile"))
                        return(p)
                    if (is.na(home <- Sys.getenv("HOME", NA_character_)))
                        return(NA_character_)
                    p <- sprintf("%s/.Rprofile", home)
                    return(p)
                }
            })
            list(filename = filename, wd = getwd(), exists = file.exists(filename))
        })
    })
    delayedAssign("ofile", {
        if (isTRUE(.shINFO[["no_init_file"]]))
            NA_character_
        else if (file_info[["exists"]])
            file_info[["filename"]]
        else NA_character_
    })
    delayedAssign("file", {
        if (is.na(ofile))
            .normalizePath(ofile)
        else if (is.null(file_info[["wd"]]))
            .normalizePath(ofile)
        else .normalizePath_against(file_info[["wd"]], ofile)
    })
              function (original = TRUE, for.msg = FALSE)
.External2(.C_init_file, original, for.msg)
})





.startup_info <- evalq(envir = new.env(), {
    delayedAssign("has_init_file", { !is.na(.init_file()) })
    delayedAssign("has_site_file", { !is.na(.site_file()) })
    environment()
})


.in_site_file <- evalq(envir = new.env(), {
    x <- TRUE
    if (getRversion() >= "4.2.0") {
                 function ()
{
    if (x) {
        sym <- ".Library.site"
        if (exists(sym, envir = .GlobalEnv, inherits = FALSE) &&
            bindingIsActive(sym, .GlobalEnv) &&
            exists(sym, envir = .BaseNamespaceEnv, inherits = FALSE) &&
            bindingIsActive(sym, .BaseNamespaceEnv) &&
            identical(
                activeBindingFunction(sym, .GlobalEnv),
                activeBindingFunction(sym, .BaseNamespaceEnv)
            ))
        {
            ## we are still in the site file, so do nothing
        }
        else {
            ## we are no longer in the site file, permanently set to FALSE
            x <<- FALSE
            lockBinding("x", parent.env(environment()))
        }
    }
    x
}
    } else {
                 function ()
{
    if (x) {
        if (sys.parents()[1L] == 1L) {
            ## we are still in the site file, so do nothing
        }
        else {
            ## we are no longer in the site file, permanently set to FALSE
            x <<- FALSE
            lockBinding("x", parent.env(environment()))
        }
    }
    x
}
    }
})





site.file <- function (original = FALSE, for.msg = FALSE, default, else.)
{
    if (missing(default)) {
        if (missing(else.))
            .site_file(original, for.msg)
        else stop("'site.file' with 'else.' but not 'default' makes no sense")
    }
    else {
        if (missing(else.)) {
            if (.startup_info[["has_site_file"]])
                .site_file(original, for.msg)
            else if (for.msg)
                NA_character_
            else default
        }
        else {
            if (.startup_info[["has_site_file"]]) {
                value <- .site_file(original, for.msg)
                (else.)(value)
            }
            else if (for.msg) {
                value <- NA_character_
                (else.)(value)
            }
            else default
        }
    }
}


init.file <- function (original = FALSE, for.msg = FALSE, default, else.)
{
    if (missing(default)) {
        if (missing(else.))
            .init_file(original, for.msg)
        else stop("'init.file' with 'else.' but not 'default' makes no sense")
    }
    else {
        if (missing(else.)) {
            if (.startup_info[["has_init_file"]])
                .init_file(original, for.msg)
            else if (for.msg)
                NA_character_
            else default
        }
        else {
            if (.startup_info[["has_init_file"]]) {
                value <- .init_file(original, for.msg)
                (else.)(value)
            }
            else if (for.msg) {
                value <- NA_character_
                (else.)(value)
            }
            else default
        }
    }
}





with_site.file <- function (expr)
.External2(.C_with_startup_file)


with_init.file <- function (expr)
{
    if (
        ## this condition must be first because it checks the validity of 'expr'
        .External2(.C_is_valid_init_file_expr) &&

        ## check that this expression is part of the top of the call stack
        ## i.e. evaluated in global environment WITHOUT using .Internal(eval())
        sys.parent() == 0L &&
        identical(parent.frame(), .GlobalEnv) &&
        .External2(.C_sys.whiches, 0L)[1L] == 1L &&

        ## check that the first call has no source reference
        ## it might if GUI runs a file
        is.null(attributes(sys.call(1L))) &&

        ## check that there are no other files being run right now
        .External2(.C_getframenumber) == 0L &&

        ## this condition must be last because it forces the promises in .init_file
        .startup_info[["has_init_file"]])
    {
        .External2(.C_set_init_file)
        on.exit(.External2(.C_unset_init_file))
    }
    .External2(.C_with_startup_file)
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2024   Iris Simmons
##


stopifnot(identical(
    this.path:::.is_clipboard(c("clipboard", "clipboard-128", "X11_primary", "X11_secondary", "X11_clipboard")),
    if (.Platform$OS.type == "windows")
        c(TRUE , TRUE , FALSE, FALSE, FALSE)
    else
        c(TRUE , FALSE, TRUE , TRUE , TRUE )
))

##
## this.path : Get Executing Script's Path
## Copyright (C) 2024   Iris Simmons
##


with_script_path <- function (expr, file, local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"), srcfile = if (n) sys.parent(n) else 0L)
{
    if (missing(file)) {
        n <- .External2(.C_asIntegerGE0, n)
        local
        envir
        matchThisEnv
        srcfile
        new_script_path <- tryCatch3({
            .External2(.C_this_path, local, envir, matchThisEnv, srcfile)
        }, thisPathNotFoundError = {
            if (is.null(wd <- getwd()))
                "."
            else path.join(wd, ".")
        })
    }
    else {
        new_script_path <- set.sys.path(file, path.only = TRUE,
            allow.file.uri = TRUE)
        unset.sys.path()
    }


    ## a function that will make and evaluate a promise when called
    eval.env <- environment()
    first_time <- TRUE
    set_script_path_fun <- function(pkgname, pkgpath) {
        set_script_path_expr <- NULL  ## this makes R CMD check happy
        delayedAssign(
            "set_script_path_expr",
            {
                oscript_path <- box::script_path()
                ## first time only, set another on.exit to restore script_path
                if (first_time) {
                    on.exit(
                        if (isNamespaceLoaded("box")) box::set_script_path(oscript_path),
                        add = TRUE
                    )
                    first_time <- FALSE
                }
                box::set_script_path(new_script_path)
                ## value of the promise is NULL, not that it matters
                NULL
            },
            eval.env
        )
        set_script_path_expr
    }
    ## append the function to the package:box onLoad hook list,
    ## and on exit remove the hook from the list
    ##
    ## this leaves the hook list the same as it was before entering
    ##
    ## why use a hook? we do not want to force package:box to load, and we want
    ## to re-run the function if the user reloads package:box
    hookName <- packageEvent("box")
    on.exit({
        hooks <- getHook(hookName)
        for (i in seq_along(hooks)) {
            if (.identical(hooks[[i]], set_script_path_fun)) {
                setHook(hookName, hooks[-i], "replace")
                break
            }
        }
    })
    setHook(hookName, set_script_path_fun)
    if (isNamespaceLoaded("box")) set_script_path_fun()


    expr
}


make_fix_file <- function (criterion, local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"), srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    path <- .External2(.C_this_path, local, envir, matchThisEnv, srcfile)
    path <- .dir(path)
    if (grepl("^(https|http|ftp|ftps)://", path, useBytes = TRUE))
        stop("make_fix_file() does not work for URL pathnames")
    if (!inherits(criterion, "root_criterion"))
        criterion <- rprojroot::as.root_criterion(criterion)
    make_fix_funs(criterion$find_file(path = path))$here
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2024   Iris Simmons
##


if (getRversion() < "3.0.0") {


delayedAssign(".C_mapply", { getNativeSymbolInfo("do_mapply", PACKAGE = "base") })


.mapply <- function (FUN, dots, MoreArgs)
.Call(.C_mapply, match.fun(FUN), dots, MoreArgs, environment())


list.files <- function (path = ".", pattern = NULL, all.files = FALSE, full.names = FALSE,
    recursive = FALSE, ignore.case = FALSE, include.dirs = FALSE,
    no.. = FALSE)
{
    no.. <- .asLogical(no..)
    if (is.na(no..))
        stop(gettextf("invalid '%s' argument", "no..", domain = "R"), domain = NA)
    value <- .BaseNamespaceEnv$list.files(path = path, pattern = pattern,
        all.files = all.files, full.names = full.names, recursive = recursive,
        ignore.case = ignore.case, include.dirs = include.dirs)
    if (all.files && no..)
        value[!(basename2(value) %in% c(".", ".."))]
    else value
}


## parse(keep.source = getOption("keep.source")) was added in R 3.0.0
parse <- .removeSource(evalq(envir = .BaseNamespaceEnv,
         function (file = "", n = NULL, text = NULL, prompt = "?", keep.source = getOption("keep.source"),
    srcfile = NULL, encoding = "unknown")
{
    if (missing(srcfile)) {
        if (!missing(keep.source)) {
            opt.keep.source <- getOption("keep.source")
            if (isTRUE(keep.source) != isTRUE(opt.keep.source)) {
                on.exit(options(keep.source = opt.keep.source))
                options(keep.source = keep.source)
            }
        }
        parse(file = file, n = n, text = text, prompt = prompt, srcfile = , encoding = encoding)
    }
    else parse(file = file, n = n, text = text, prompt = prompt, srcfile = srcfile, encoding = encoding)
}
))


}


if (getRversion() < "3.1.0") {


anyNA <- function (x, recursive = FALSE)
.External2(.C_anyNA, x, recursive)


.anyNA_dispatch <- function (x, recursive = FALSE)
UseMethod("anyNA")


anyNA.data.frame <- function (x, recursive = FALSE)
.External2(.C_anyNA.data.frame, x, recursive)


anyNA.numeric_version <- function (x, recursive = FALSE)
.External2(.C_anyNA.numeric_version, x)


anyNA.POSIXlt <- function (x, recursive = FALSE)
anyNA(as.POSIXct.POSIXlt(x))


.anyNA.default <- function (x, recursive = FALSE)
.External2(.C_anyNA.default, x, recursive)


}


if (getRversion() < "3.2.0") {


isNamespaceLoaded <- function (name)
.External2(.C_isRegisteredNamespace, name)


dir.exists <- function (paths)
.External2(.C_dir.exists, paths)


file.copy <- function (from, to, overwrite = recursive, recursive = FALSE,
    copy.mode = TRUE, copy.date = FALSE)
{
    okay <- .BaseNamespaceEnv$file.copy(from, to, overwrite = overwrite,
        recursive = recursive, copy.mode = copy.mode)
    if (copy.date) {
        if (any(okay)) {
            if (length(to) == 1L && dir.exists(to)) {
                from <- from[okay]
                to <- path.join(to, basename2(from))
                if (recursive) {
                    n <- lapply(from, list.files, all.files = TRUE,
                        recursive = TRUE, include.dirs = TRUE)
                    times <- lengths(n)
                    n <- unlist(n)
                    from <- c(path.join(rep(from, times), n), from)
                    to <- c(path.join(rep(to, times), n), to)
                }
                Sys.setFileTime(to, file.mtime(from))
            }
            else Sys.setFileTime(to[okay], file.mtime(from[okay]))
        }
    }
    okay
}


lengths <- function (x, use.names = TRUE)
.External2(.C_lengths, x, use.names)


.lengths.default <- function (x, use.names = TRUE)
.External2(.C_lengths.default, x, use.names)


## file.info(extra_cols = TRUE) was added in R 3.2.0
file.info <- .removeSource(evalq(envir = .BaseNamespaceEnv,
             function (..., extra_cols = TRUE)
{
    if (extra_cols)
        file.info(...)
    else file.info(...)[1:6]
}
))


file.mtime <- .removeSource(evalq(envir = .BaseNamespaceEnv,
              function (...)
file.info(...)$mtime
))


file.size <- .removeSource(evalq(envir = .BaseNamespaceEnv,
             function (...)
file.info(...)$size
))


.isdir <- .removeSource(evalq(envir = .BaseNamespaceEnv,
          function (...)
file.info(...)$isdir
))


} else {  ## (getRversion() >= "3.2.0")


.isdir <- .removeSource(evalq(envir = .BaseNamespaceEnv,
          function (...)
file.info(..., extra_cols = FALSE)$isdir
))


}


if (getRversion() < "3.3.0") {


strrep <- function (x, times)
{
    if (!is.character(x))
        x <- as.character(x)
    .External2(.C_strrep, x, as.integer(times))
}


startsWith <- function (x, prefix)
.External2(.C_startsWith, x, prefix)


endsWith <- function (x, suffix)
.External2(.C_endsWith, x, suffix)


}


if (getRversion() < "3.4.0") {


print.connection <- .removeSource(evalq(envir = .BaseNamespaceEnv,
                    function (x, ...)
{
    usumm <- tryCatch(unlist(summary(x)), error = function(e) {
    })
    if (is.null(usumm)) {
        cl <- oldClass(x)
        cl <- cl[cl != "connection"]
        cat("A connection, ", if (length(cl))
            paste0("specifically, ", paste(sQuote(cl), collapse = ", "),
                ", "), "but invalid.\n", sep = "")
    }
    else {
        cat("A connection with")
        print(cbind(` ` = usumm), ...)
    }
    invisible(x)
}
))


.withAutoprint <- function (exprs, evaluated = FALSE, local = parent.frame(), print. = TRUE,
    echo = TRUE, max.deparse.length = Inf, width.cutoff = max(20, getOption("width")),
    deparseCtrl = c("keepInteger", "showAttributes", "keepNA"),
    skip.echo = 0, spaced = FALSE, ...)
{
    if (!evaluated) {
        exprs <- substitute(exprs)
        if (is.call(exprs)) {
            if (typeof(exprs[[1L]]) == "symbol" && exprs[[1L]] == "{") {
                exprs <- as.list(exprs)[-1L]
                if (missing(skip.echo) &&
                    length(exprs) &&
                    typeof(srcref <- attr(exprs, "srcref", exact = TRUE)) == "list")
                {
                    skip.echo <- srcref[[1L]][7L] - 1L
                }
            }
        }
    }
    if (!is.expression(exprs))
        exprs <- as.expression(exprs)
    conn <- textConnection(.code2character(exprs, width.cutoff = width.cutoff, deparseCtrl = deparseCtrl))
    on.exit(close(conn))
    source(file = conn, local = local, print.eval = print., echo = echo,
        max.deparse.length = max.deparse.length, skip.echo = skip.echo, ...)
}


withAutoprint <- .withAutoprint


} else {  ## (getRversion() >= "3.4.0")


.withAutoprint <- .removeSource(evalq(envir = .BaseNamespaceEnv,
                  function (exprs, evaluated = FALSE, local = parent.frame(), print. = TRUE,
    echo = TRUE, max.deparse.length = Inf, width.cutoff = max(20, getOption("width")),
    deparseCtrl = c("keepInteger", "showAttributes", "keepNA"),
    skip.echo = 0, ...)
{
    if (!evaluated) {
        exprs <- substitute(exprs)
        if (is.call(exprs)) {
            if (typeof(exprs[[1L]]) == "symbol" && exprs[[1L]] == "{") {
                exprs <- as.list(exprs)[-1L]
                if (missing(skip.echo) &&
                    length(exprs) &&
                    typeof(srcref <- attr(exprs, "srcref", exact = TRUE)) == "list")
                {
                    skip.echo <- srcref[[1L]][7L] - 1L
                }
            }
        }
    }
    source(exprs = exprs, local = local, print.eval = print.,
        echo = echo, max.deparse.length = max.deparse.length,
        width.cutoff = width.cutoff, deparseCtrl = deparseCtrl,
        skip.echo = skip.echo, ...)
}
))


}


if (getRversion() < "3.5.0") {


...length <- function ()
.External2(.C_...length)


## isTRUE(x) was previously defined as:
## function (x)
## identical(TRUE, x)
isTRUE <- .removeSource(evalq(envir = .BaseNamespaceEnv,
          function (x)
is.logical(x) && length(x) == 1L && !is.na(x) && x
))


isFALSE <- .removeSource(evalq(envir = .BaseNamespaceEnv,
           function (x)
is.logical(x) && length(x) == 1L && !is.na(x) && !x
))


}


if (getRversion() < "3.6.0") {


errorCondition <- .removeSource(evalq(envir = .BaseNamespaceEnv,
                  function (message, ..., class = NULL, call = NULL)
structure(list(message = as.character(message), call = call, ...),
    class = c(class, "error", "condition"))
))


str2expression <- function (text)
{
    if (!is.character(text))
        stop("argument must be character", domain = "R")
    parse(text = text, n = -1, keep.source = FALSE, srcfile = NULL)
}


str2lang <- function (s)
{
    if (!.IS_SCALAR_STR(s))
        stop("argument must be a character string", domain = "R")
    ans <- parse(text = s, n = -1, keep.source = FALSE, srcfile = NULL)
    if (length(ans) != 1L)
        stop(gettextf("parsing result not of length one, but %d", length(ans), domain = "R"), domain = NA)
    ans[[1L]]
}


Sys.setFileTime <- function (path, time)
{
    if (!is.character(path))
        stop("invalid 'path' argument", domain = "R-base")
    time <- as.POSIXct(time)
    if (any(is.na(time)))
        stop("invalid 'time' argument", domain = "R-base")
    n <- length(path)
    if (!n)
        return(logical())
    m <- length(time)
    if (!m)
        stop(gettextf("'%s' must be of length at least one",
            "time", domain = "R"), domain = NA)
    if (m > n)
        time <- time[seq_len(n)]
    as.logical(.mapply(.BaseNamespaceEnv$Sys.setFileTime, list(path, time), NULL))
}


}


if (getRversion() < "4.0.0") {


deparse1 <- .removeSource(evalq(envir = .BaseNamespaceEnv,
            function (expr, collapse = " ", width.cutoff = 500L, ...)
paste(deparse(expr, width.cutoff, ...), collapse = collapse)
))


unlink <- .removeSource(evalq(envir = .BaseNamespaceEnv,
          function (x, recursive = FALSE, force = FALSE, expand = TRUE)
unlink(x, recursive = recursive, force = force)
))


}


if (getRversion() < "4.1.0") {


...elt <- function (n)
.External2(.C_...elt, n)


## bquote(splice = TRUE) was added in R 4.1.0
bquote <- .removeSource(evalq(envir = .BaseNamespaceEnv,
          function (expr, where = parent.frame(), splice = FALSE)
{
    if (!is.environment(where))
        where <- as.environment(where)
    unquote <- function(e) {
        if (is.pairlist(e))
            as.pairlist(lapply(e, unquote))
        else if (is.call(e)) {
            if (typeof(e[[1L]]) == "symbol" && e[[1L]] == ".")
                eval(e[[2L]], where)
            else if (splice) {
                if (typeof(e[[1L]]) == "symbol" && e[[1L]] == "..")
                    stop("can only splice inside a call", call. = FALSE)
                else as.call(unquote.list(e))
            }
            else as.call(lapply(e, unquote))
        }
        else e
    }
    is.splice.macro <- function(e) is.call(e) && typeof(e[[1L]]) == "symbol" && e[[1L]] == ".."
    unquote.list <- function(e) {
        p <- Position(is.splice.macro, e, nomatch = NULL)
        if (is.null(p))
            lapply(e, unquote)
        else {
            n <- length(e)
            head <- if (p == 1)
                NULL
            else e[1:(p - 1)]
            tail <- if (p == n)
                NULL
            else e[(p + 1):n]
            macro <- e[[p]]
            mexp <- eval(macro[[2L]], where)
            if (!is.vector(mexp) && !is.expression(mexp))
                stop("can only splice vectors")
            c(lapply(head, unquote), mexp, as.list(unquote.list(tail)))
        }
    }
    unquote(substitute(expr))
}
))


}


if (getRversion() < "4.2.0") {


## gettext(trim = TRUE) was added in R 4.2.0
gettext <- .removeSource(evalq(envir = .BaseNamespaceEnv,
           function (..., domain = NULL, trim = TRUE)
{
    gettext(..., domain = domain)
}
))


## gettextf(trim = TRUE) was added in R 4.2.0
gettextf <- .removeSource(evalq(envir = .BaseNamespaceEnv,
            function (fmt, ..., domain = NULL, trim = TRUE)
sprintf(gettext(fmt, domain = domain), ...)
))


}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2024   Iris Simmons
##


print(loadedNamespaces())


this.path:::.HAVE_AQUA
this.path:::.PATH_MAX
this.path:::.NAMEDMAX

this.path:::.mbcslocale
this.path:::.utf8locale
this.path:::.latin1locale
this.path:::.R_MB_CUR_MAX

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2024   Iris Simmons
##


.cat_condition <- function (c, outfile = stderr())
{
    # c <- simpleError("testing th\U{0001D11E}s\n out", quote(5 + 6)); warning("comment out this later")


    deparse1s <- function(call) deparse(call, backtick = TRUE, nlines = 1L)
    wd <- function(x) {
        value <- nchar(x, "width", allowNA = TRUE)
        if (any(i <- is.na(value)))
            value[i] <- nchar(x[i], "bytes")
        value
    }
    strlen <- function(x) nchar(x, "bytes")
    strchr <- function(text, pattern) regexpr(pattern, text, fixed = TRUE, useBytes = TRUE)
    snprintf <- function(n, fmt, ..., printTrunc = FALSE) {
        buf <- sprintf(fmt = fmt, ...)
        if (strlen(buf) >= n) {
            buf <- substr(buf, 1L, sum(cumsum(strlen(strsplit(buf, NULL)[[1L]])) < n))
            if (printTrunc) {
                msg <- gettext("[... truncated]", domain = "R")
                if (strlen(buf) + 1L + strlen(msg) < BUFSIZE) {
                    buf <- rawToChar(c(charToRaw(buf), charToRaw(" "), charToRaw(msg)))
                }
            }
        }
        buf
    }
    strncat <- function(n, x, y) {
        buf <- rawToChar(c(charToRaw(x), charToRaw(y)))
        if (strlen(buf) > n)
            buf <- substr(buf, 1L, sum(cumsum(strlen(strsplit(buf, NULL)[[1L]])) <= n))
        buf
    }
    LONGWARN <- 75L
    BUFSIZE <- 8192L
    R_WarnLength <- 1000L
    if (inherits(c, "warning")) {
        call <- conditionCall(c)
        format <- "%s"
        ap <- conditionMessage(c)
        dcall <- if (!is.null(call)) {
            deparse1s(call)
        } else ""
        psize <- min(BUFSIZE, R_WarnLength + 1)
        buf <- snprintf(psize, format, ap, printTrunc = TRUE)
        buf <- c(
            if (nzchar(dcall)) {
                c(
                    gettextf("Warning in %s :", dcall, domain = "R"),
                    if (!(( .mbcslocale && 18L + wd(dcall) + wd(buf) <= LONGWARN) ||
                          (!.mbcslocale && 18L + strlen(dcall) + strlen(buf) <= LONGWARN)))
                        "\n "
                )
            } else gettext("Warning:", domain = "R"),
            sprintf(" %s\n", buf)
        )
    }
    else if (inherits(c, "error")) {
        call <- conditionCall(c)
        format <- "%s"
        ap <- conditionMessage(c)
        msg_len <- min(BUFSIZE, R_WarnLength) + 1
        if (!is.null(call)) {
            head <- gettext("Error in ", domain = "R", trim = FALSE); tail <- "\n  "
            dcall <- deparse1s(call)
            tmp2 <- snprintf(BUFSIZE, "%s", head)
            tmp <- snprintf(max(msg_len - strlen(head), 0L), format, ap)
            if (strlen(tmp2) + strlen(tail) + strlen(tmp) < BUFSIZE) {
                errbuf <- snprintf(BUFSIZE, gettext("Error in %s : ", domain = "R", trim = FALSE), dcall)
                if (.mbcslocale) {
                    msgline1 <- strchr(tmp, "\n")
                    msgline1 <- if (msgline1 != -1L) wd(rawToChar(charToRaw(tmp)[seq_len(msgline1 - 1L)])) else wd(tmp)
                    if (14L + strlen(dcall) + msgline1 > LONGWARN)
                        errbuf <- strncat(BUFSIZE, errbuf, tail)
                } else {
                    msgline1 <- strchr(tmp, "\n")
                    msgline1 <- if (msgline1 != -1L) msgline1 - 1L else strlen(tmp)
                    if (14L + strlen(dcall) + msgline1 > LONGWARN)
                        errbuf <- strncat(BUFSIZE, errbuf, tail)
                }
                errbuf <- strncat(BUFSIZE, errbuf, tmp)
            } else {
                errbuf <- sprintf("%s", gettext("Error: ", domain = "R", trim = FALSE))
                errbuf <- strncat(BUFSIZE, errbuf, tmp)
            }
        } else {
            errbuf <- sprintf("%s", gettext("Error: ", domain = "R", trim = FALSE))
            errbuf <- strncat(Inf, errbuf, snprintf(max(msg_len - strlen(errbuf), 0L), format, ap))
        }
        nc <- strlen(errbuf)
        if (nc > BUFSIZE - 1L - (.R_MB_CUR_MAX - 1L)) {
            end <- min(nc + 1L, (BUFSIZE + 1L) - 4L)
            if (nc > end - 1L)
                errbuf <- strncat(end - 1L, errbuf, "")
            errbuf <- strncat(BUFSIZE, errbuf, "...\n")
        } else {
            if (!endsWith(errbuf, "\n")) {
                errbuf <- strncat(BUFSIZE, errbuf, "\n")
                nc <- nc + 1L
            }
        }
        buf <- errbuf
    }
    else buf <- conditionMessage(c)
    if (is.null(outfile))
        paste(buf, collapse = "")
    else {
        cat(buf, file = outfile, sep = "")
        invisible(paste(buf, collapse = ""))
    }
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2021-2025   Iris Simmons
##


local({
    FILE.R <- tempfile(fileext = ".R")
    on.exit(unlink(FILE.R), add = TRUE)
    this.path:::.writeCode(file = FILE.R, {
        if (requireNamespace("microbenchmark", quietly = TRUE)) {
            print(this.path:::.faster_subsequent_times_test())
        } else cat("\n'package:microbenchmark' is not available :(\n")
    })
    cat("\n")
    this.path:::.Rscript(c(
        "--default-packages=NULL", "--no-save", FILE.R
    ))


    cat("\n> source(FILE.R, chdir = FALSE)\n")
    source(FILE.R, chdir = FALSE)


    cat("\n> source(FILE.R, chdir = TRUE)\n")
    source(FILE.R, chdir = TRUE)
})

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2024   Iris Simmons
##


.LINENO <- function (path, to = 1L)
{
    for (which in seq.int(to = to, by = -1L, length.out = sys.nframe() - to)) {
        call <- sys.call(which)
        srcref <- attr(call, "srcref", exact = TRUE)
        if (!is.null(srcref)) {
            ## try to get the normalized filename
            success <- tryCatch({
                value <- .External2(.C_src_path, srcref)
                TRUE
            }, error = function(e) FALSE)
            if (success) {
                ## compare the filenames
                if (.str_equal_useBytes(path, value)) {
                    # return(srcref[7L])

                    ## srcref[1L] is better, it respects #line directives
                    return(srcref[1L])
                }
            }
            else {
                ## try to get the original filename
                success <- tryCatch({
                    value <- .External2(.C_src_path, FALSE, TRUE, FALSE, FALSE, srcref)
                    TRUE
                }, error = function(e) FALSE)
                if (success) {
                    if (startsWith(value, "file://"))
                        value <- .fixslash(.file_URL_path(value))
                    else if (grepl("^(https|http|ftp|ftps)://", value, useBytes = TRUE))
                        value <- .normalizeURL(value)
                    else value <- .fixslash(value)
                    if (.str_equal_useBytes(path, value))
                        return(srcref[1L])
                }
            }
        }
    }
    NA_integer_
}


sys.LINENO <- function ()
{
    success <- tryCatch({
        n <- .External2(.C_getframenumber)
        if (is.na(n) || n < 1L)
            return(NA_integer_)
        path <- .External2(.C_sys_path)
        TRUE
    }, error = function(e) FALSE)
    if (success)
        .LINENO(path, n + 1L)
    else NA_integer_
}


env.LINENO <- function (n = 0L, envir = parent.frame(n + 1L), matchThisEnv = getOption("topLevelEnvironment"))
{
    n <- .External2(.C_asIntegerGE0, n)
    envir
    matchThisEnv ## this is unused
    value <- NA_integer_
    if (typeof(envir) == "environment") {
        parents <- sys.parents()
        for (i in seq.int(to = 1L, by = -1L, along.with = parents)) {
            ## if the parent frame matches envir,
            ## look for a srcref on the corresponding call
            if (identical(envir, sys.frame(parents[[i]]))) {
                call <- sys.call(i)
                srcref <- attr(call, "srcref", exact = TRUE)
                if (!is.null(srcref)) {
                    value <- srcref[1L]
                    break
                }
            }
        }
    }
    value
}


src.LINENO <- function (n = 0L, srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    srcfile
    tryCatch({
        .External2(.C_src_LINENO, srcfile)
    }, error = function(e) NA_integer_)
}


LINENO <- function (n = 0L, envir = parent.frame(n + 1L), matchThisEnv = getOption("topLevelEnvironment"),
    srcfile = if (n) sys.parent(n) else 0L)
{
    n <- .External2(.C_asIntegerGE0, n)
    srcfile
    value <- tryCatch({
        .External2(.C_src_LINENO, srcfile)
    }, error = function(e) NA_integer_)
    if (is.na(value)) {
        envir
        matchThisEnv
        value <- NA_integer_
        if (typeof(envir) == "environment") {
            parents <- sys.parents()
            for (i in seq.int(to = 1L, by = -1L, along.with = parents)) {
                if (identical(envir, sys.frame(parents[[i]]))) {
                    call <- sys.call(i)
                    srcref <- attr(call, "srcref", exact = TRUE)
                    if (!is.null(srcref)) {
                        value <- srcref[1L]
                        break
                    }
                }
            }
        }
        if (is.na(value)) {
            success <- tryCatch({
                n <- .External2(.C_getframenumber)
                if (is.na(n) || n < 1L)
                    return(NA_integer_)
                path <- .External2(.C_sys_path)
                TRUE
            }, error = function(e) FALSE)
            if (success)
                .LINENO(path, n + 1L)
            else NA_integer_
        }
        else value
    }
    else value
}


LINE <- function ()
{
    srcfile <- 0L
    value <- tryCatch({
        .External2(.C_src_LINENO, srcfile)
    }, error = function(e) NA_integer_)
    if (is.na(value)) {
        envir <- parent.frame()
        value <- NA_integer_
        parents <- sys.parents()
        for (i in seq.int(to = 1L, by = -1L, along.with = parents)) {
            if (identical(envir, sys.frame(parents[[i]]))) {
                call <- sys.call(i)
                srcref <- attr(call, "srcref", exact = TRUE)
                if (!is.null(srcref)) {
                    value <- srcref[1L]
                    break
                }
            }
        }
        if (is.na(value)) {
            success <- tryCatch({
                n <- .External2(.C_getframenumber)
                if (is.na(n) || n < 1L)
                    return(NA_integer_)
                path <- .External2(.C_sys_path)
                TRUE
            }, error = function(e) FALSE)
            if (success)
                .LINENO(path, n + 1L)
            else NA_integer_
        }
        else value
    }
    else value
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2026   Iris Simmons
##


## these are expressions that are used in this package,
## but which are not expected to change value in a session,
## so we can evaluate them exactly once when needed
##
## if you've never heard of an R promise before, it's a mechanism for handling
## delayed argument evaluation and substitution. a promise consists of four
## things:
##
##   * an expression
##
##   * an environment in which the expression will be evaluated
##
##   * value of evaluating said expression in said environment, initially empty
##
##   * the number of attempts to evaluate the promise, used to detect recursive
##     references and restarting interrupted promises
##
## it is called a promise because "I promise to evaluate this code in this
## environment if / / when requested"
##
## an object like an integer, a boolean, and a string will evaluate to itself.
## other objects like a symbol, a bytecode expression, or a call will evaluate
## the same as usual.
##
## if / / when the code has been evaluated, the value of the expression is
## stored inside the promise. this means the promise does not re-evaluate the
## same piece of code multiple times. for example:
##
## ```R
## delayedAssign("x", {
##     writeLines("evaluating 'x'")
##     5 + 6
## })
## x
## x
## x
## ```
##
## which produces the following output:
##
## ```
## > delayedAssign("x", {
## +     writeLines("evaluating 'x'")
## +     5 + 6
## + })
## > x
## evaluating 'x'
## [1] 11
## > x
## [1] 11
## > x
## [1] 11
## >
## ```
##
## you'll see that the expression `writeLines("evaluating 'x'")` is only
## evaluated once, as desired! you can use `ikws.debug::PRINFO()` to examine
## the state of the promise:
##
## ```R
## delayedAssign("x", {
##     writeLines("evaluating 'x'")
##     5 + 6
## })
## ikws.debug::PRINFO(x)
## x
## ikws.debug::PRINFO(x)
## ```
##
## which produces the following output:
##
## ```
## > delayedAssign("x", {
## +     writeLines("evaluating 'x'")
## +     5 + 6
## + })
## > ikws.debug::PRINFO(x)
## $PRCODE
## {
##     writeLines("evaluating 'x'")
##     5 + 6
## }
##
## $PRENV
## <environment: R_GlobalEnv>
##
## $PREXPR
## {
##     writeLines("evaluating 'x'")
##     5 + 6
## }
##
## $PRSEEN
## [1] 0
##
## > x
## evaluating 'x'
## [1] 11
## > ikws.debug::PRINFO(x)
## $PRCODE
## {
##     writeLines("evaluating 'x'")
##     5 + 6
## }
##
## $PRENV
## NULL
##
## $PREXPR
## {
##     writeLines("evaluating 'x'")
##     5 + 6
## }
##
## $PRSEEN
## [1] 0
##
## $PRVALUE
## [1] 11
##
## >
## ```
##
## "PRCODE", "PRENV", "PRVALUE", and "PRSEEN" were all mentioned earlier, but
## "PREXPR" has not been mentioned yet. if "PRCODE" is a bytecode, then
## "PREXPR" will be the uncompiled version of "PRCODE". in all other cases,
## they are identical.
##
## notice that "PRENV" is set to `NULL` once the promise has been evaluated,
## this has to do with garbage collection. it allows the garbage collector to
## collect environments which are no longer in use.
##
## "PRSEEN" was mentioned earlier, it is an integer used to determine if a
## promise leads to infinite recursion or has been interrupted and is being
## restarted. for example:
##
## ```R
## delayedAssign("x", withAutoprint({
##     5 + 6
##     ikws.debug::PRINFO(x)
##     x
## }))
## ikws.debug::PRINFO(x)
## try(x)
## ikws.debug::PRINFO(x)
## ```
##
## notice "PRSEEN" is 0 when `x` is not being evaluated, it is 1 when `x` is
## being evaluated, and it is 2 when `x` has been interrupted (by an error) and
## restarted:
##
## ```
## > ikws.debug::PRINFO(x)
## $PRCODE
## withAutoprint({
##     5 + 6
##     ikws.debug::PRINFO(x)
##     x
## })
##
## $PRENV
## <environment: R_GlobalEnv>
##
## $PREXPR
## withAutoprint({
##     5 + 6
##     ikws.debug::PRINFO(x)
##     x
## })
##
## $PRSEEN
## [1] 0
##
## > try(x)
## > 5 + 6
## [1] 11
## > ikws.debug::PRINFO(x)
## $PRCODE
## withAutoprint({
##     5 + 6
##     ikws.debug::PRINFO(x)
##     x
## })
##
## $PRENV
## <environment: R_GlobalEnv>
##
## $PREXPR
## withAutoprint({
##     5 + 6
##     ikws.debug::PRINFO(x)
##     x
## })
##
## $PRSEEN
## [1] 1
##
## > x
## Error in eval(ei, envir) :
##   promise already under evaluation: recursive default argument reference or earlier problems?
## > ikws.debug::PRINFO(x)
## $PRCODE
## withAutoprint({
##     5 + 6
##     ikws.debug::PRINFO(x)
##     x
## })
##
## $PRENV
## <environment: R_GlobalEnv>
##
## $PREXPR
## withAutoprint({
##     5 + 6
##     ikws.debug::PRINFO(x)
##     x
## })
##
## $PRSEEN
## [1] 2
##
## >
## ```
##
## these promises are created when the package is built and installed. they are
## evaluated during the regular use of the package


delayedAssign(".OS_unix"   , { .Platform$OS.type == "unix"    })
delayedAssign(".OS_windows", { .Platform$OS.type == "windows" })


## we need to determine the type of GUI in use. unfortunately, .Platform$GUI is
## not good enough. for example, 'VSCode' and 'Jupyter' do not set
## .Platform$GUI, and 'RStudio' does not set .Platform$GUI until after the
## site-wide startup profile file, a user profile, and the function .First()
## have been run (see ?Startup).
##
## as such, I've made my own ways of determining the type of GUI in use


delayedAssign(".GUI_RStudio", { commandArgs()[[1L]] == "RStudio" })


delayedAssign(".GUI_Positron", {
    interactive() &&
    .scalar_streql(Sys.getenv("POSITRON"), "1") &&
    Sys.getenv("POSITRON_VERSION") != "" &&

    (
        ( .OS_unix    && .Platform$GUI %in% c("Positron", "X11", "unknown", "none") ) ||
        ( .OS_windows && .Platform$GUI %in% c("Positron", "Rgui"                  ) )
    )
})


delayedAssign(".OS_unix_maybe_unembedded_shell"   , { .OS_unix    && .Platform$GUI %in% c("X11"  , "unknown", "none") &&                        "R" == basename2(commandArgs()[[1L]])                                             })
delayedAssign(".OS_windows_maybe_unembedded_shell", { .OS_windows && .Platform$GUI %in% c("RTerm", "unknown"        ) && grepl("(?i)^Rterm(\\.exe)?$", basename2(commandArgs()[[1L]])) && .External2(.C_CharacterMode) == "RTerm" })
delayedAssign(".maybe_unembedded_shell", { .OS_unix_maybe_unembedded_shell || .OS_windows_maybe_unembedded_shell })


## see function do_shINFO in file ./src/shfile.c
delayedAssign(".shINFO", { .External2(.C_shINFO) })


delayedAssign(".console_radian", { commandArgs()[[1L]] == "radian" })


delayedAssign(".GUI_vscode", {
    interactive() &&
    isTRUE(Sys.getenv("TERM_PROGRAM") == "vscode") &&
    (
        isFALSE(.shINFO[["has_input"]]) ||
        .console_radian
    )
})


## jupyter build a competent API challenge (impossible)
delayedAssign(".GUI_jupyter", {
    !interactive() &&

    Sys.getenv("JPY_API_TOKEN") != "" &&
    Sys.getenv("JPY_PARENT_PID") != "" &&

    .maybe_unembedded_shell &&

    isTRUE(.shINFO[["has_input"]]) &&
    is.na(.shINFO[["FILE"]]) &&
    !is.na(.shINFO[["EXPR"]]) &&

    length(commandArgs(TRUE)) == 1L &&
    file.exists(commandArgs(TRUE)) &&

    ## this is the longest check, save for last
    local({
        exprs <- tryCatch(parse(text = .shINFO[["EXPR"]], n = -1, keep.source = FALSE, srcfile = NULL),
            error = identity)
        !inherits(exprs, "error") &&
            length(exprs) &&
            .identical(
                exprs[[length(exprs)]],
                as.call(list(call("::", quote(IRkernel), quote(main))))  # quote(IRkernel::main())
            )
    })
})


delayedAssign(".GUI_emacs", {
    interactive() &&
    Sys.getenv("STATATERM") == "emacs" &&
    .maybe_unembedded_shell &&
    (.OS_unix || (.OS_windows && .shINFO[["ess"]]))
})


delayedAssign(".GUI_rkward", { commandArgs()[[1L]] == "rkward" })


delayedAssign(".GUI_powerbi", {
    !interactive() &&

    Sys.getenv("RPackagesLibrariesDirectory") != "" &&
    Sys.getenv("RScriptWrapperWorkingDirectory") != "" &&

    .maybe_unembedded_shell &&

    .scalar_streql(.shINFO[["ENC"]], "UTF-8") &&
    isTRUE(.shINFO[["has_input"]]) &&
    !is.na(.shINFO[["FILE"]]) &&
    is.na(.shINFO[["EXPR"]])
})


delayedAssign(".in_callr", {
    !interactive() &&

    .scalar_streql(Sys.getenv("CALLR_IS_RUNNING"), "true") &&

    .maybe_unembedded_shell &&

    isTRUE(.shINFO[["has_input"]]) &&
    !is.na(.shINFO[["FILE"]]) &&
    is.na(.shINFO[["EXPR"]]) &&

    "tools:callr" %in% search()
})


delayedAssign(".GUI_AQUA", { .OS_unix    && .Platform$GUI == "AQUA" && !.GUI_RStudio && !.GUI_Positron && !.GUI_vscode && !.GUI_jupyter && !.GUI_emacs && !.GUI_rkward && !.GUI_powerbi && !.in_callr                            })
delayedAssign(".GUI_Rgui", { .OS_windows && .Platform$GUI == "Rgui" && !.GUI_RStudio && !.GUI_Positron && !.GUI_vscode && !.GUI_jupyter && !.GUI_emacs && !.GUI_rkward && !.GUI_powerbi && !.in_callr && .External2(.C_RConsole) })
delayedAssign(".GUI_Tk"  , { .OS_unix    && .Platform$GUI == "Tk"   && !.GUI_RStudio && !.GUI_Positron && !.GUI_vscode && !.GUI_jupyter && !.GUI_emacs && !.GUI_rkward && !.GUI_powerbi && !.in_callr                            })


delayedAssign(".OS_unix_in_shell"   , { .OS_unix_maybe_unembedded_shell    && !.GUI_Positron && !.GUI_vscode && !.GUI_jupyter && !.GUI_emacs && !.GUI_powerbi && !.in_callr })
delayedAssign(".OS_windows_in_shell", { .OS_windows_maybe_unembedded_shell && !.GUI_Positron && !.GUI_vscode && !.GUI_jupyter && !.GUI_emacs && !.GUI_powerbi && !.in_callr })
delayedAssign(".in_shell", { .OS_unix_in_shell || .OS_windows_in_shell })


delayedAssign(".unrecognized_manner", {
    !.in_shell &&
    !.GUI_RStudio &&
    !.GUI_Positron &&
    !.GUI_vscode &&
    !.GUI_jupyter &&
    !.GUI_emacs &&
    !.GUI_rkward &&
    !.GUI_powerbi &&
    !.in_callr &&
    !.GUI_Rgui &&
    !.GUI_AQUA &&
    !.GUI_Tk
})


delayedAssign(".ucrt", { identical(R.version[["crt"]], "ucrt") })
delayedAssign(".GUI", {
    if (.in_shell) .Platform$GUI
    else if (.GUI_RStudio) "RStudio"
    else if (.GUI_Positron) "Positron"
    else if (.GUI_vscode) "vscode"
    else if (.GUI_jupyter) "jupyter"
    else if (.GUI_emacs) "emacs"
    else if (.GUI_rkward) "rkward"
    else if (.GUI_powerbi) "powerbi"
    # else if (.GUI_Rgui) "Rgui"
    # else if (.GUI_AQUA) "AQUA"
    # else if (.GUI_Tk) "Tk"
    else .Platform$GUI
})


delayedAssign("initwd", { getwd() })
getinitwd <- function ()
initwd


.unset_these_envvars <- c(
    "TERM_PROGRAM", "JPY_API_TOKEN", "JPY_PARENT_PID", "STATATERM",
    "RPackagesLibrariesDirectory", "RScriptWrapperWorkingDirectory",
    "CALLR_IS_RUNNING", "POSITRON", "POSITRON_VERSION"
)

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2026   Iris Simmons
##


main <- function ()
{
    if (getRversion() < "3.2.0") {
        dir.exists <- function(paths) {
            isdir <- file.info(paths)$isdir
            !is.na(isdir) & isdir
        }
        file.size <- function(...) file.info(...)$size
    }
    oopt <- options(encoding = "native.enc")
    on.exit(options(oopt))


    info_path <- "./tools/info.dcf"
    info <- read.dcf(info_path)
    if (nrow(info) != 1L)
        stop("contains a blank line", call. = FALSE)
    ## re-read the file, keeping white space this time
    info <- read.dcf(info_path, keep.white = colnames(info))
    if (nrow(info) != 1L)
        stop("contains a blank line", call. = FALSE)
    info <- info[1L, ]
    write_info <- function() write.dcf(t(info), info_path, keep.white = names(info))
    if (file.exists("./tools/for-r-mac-builder")) {
        devel <- TRUE
    } else {
        devel <- as.logical(getOption("R_THIS_PATH_DEVEL", NA))
        if (is.na(devel))
            devel <- as.logical(Sys.getenv("R_THIS_PATH_DEVEL"))
        if (is.na(devel))
            devel <- info[["devel"]]
        devel <- if (devel) TRUE else FALSE
    }


    regexQuote <- function(x) gsub("([.\\\\|()[{^$*+?])", "\\\\\\1", x)
    RdQuote <- function(x) gsub("([%{}\\])", "\\\\\\1", x, useBytes = TRUE)


    ## we need the contents of the DESCRIPTION file
    desc <- local({
        ## see tools:::.read_description
        desc <- read.dcf("./DESCRIPTION", keep.white = c("Description", "Authors@R", "Author", "Built", "Packaged"))
        if (nrow(desc) != 1L)
            stop("contains a blank line", call. = FALSE)
        desc <- desc[1L, ]
        if (getRversion() >= "3.4.0") {
            if (!is.na(encoding <- desc["Encoding"])) {
                if (encoding == "UTF-8") {
                    Encoding(desc) <- "UTF-8"
                    ind <- validUTF8(desc)
                    if (!all(ind)) {
                        pos <- which(!ind)
                        desc[pos] <- iconv(desc[pos], "UTF-8", "UTF-8", sub = "byte")
                    }
                } else if (encoding == "latin1") {
                    Encoding(desc) <- "latin1"
                } else desc <- iconv(desc, encoding, "", sub = "byte")
            }
        } else {
            if (!is.na(encoding <- desc["Encoding"])) {
                if (encoding %in% c("latin1", "UTF-8")) {
                    Encoding(desc) <- encoding
                } else desc <- iconv(desc, encoding, "", sub = "byte")
            }
        }
        desc
    })
    local({
        pkgname <- Sys.getenv("R_PACKAGE_NAME")
        if (pkgname != desc["Package"])
            stop(gettextf("DESCRIPTION file is for package '%s', not '%s'",
                desc["Package"], pkgname))
    })


    building <- local({
        if (is.null(wd <- getwd())) {
            FALSE
        } else {
            if (.Platform$OS.type == "windows") {
                enc <- Encoding(wd)
                wd <- gsub("\\", "/", wd, fixed = TRUE, useBytes = TRUE)
                Encoding(wd) <- enc
            }
            ## we need to know if the package is being installed by 'R CMD build'
            ## (i.e. "installing the package to process help pages")
            ##
            ## we know a package is being installed by 'R CMD build'
            ## if the working directory matches the following pattern
            pattern <- sprintf("/Rtmp[\001-\056\060-\177]{6}/Rbuild[0123456789abcdef]+/%s$", regexQuote(desc["Package"]))
            ##                       ^^^^^^^^^^^^^^^^^^^^^^^                                 6 ASCII characters excluding \0 and /
            ##                                                     ^^^^^^^^^^^^^^^^^^^       at least 1 hex digit
            ##                                                                         ^^^   ends with the package name
            grepl(pattern, wd, useBytes = TRUE)
        }
    })


    ## on the maintainer's computer, prevent the source from being destroyed
    ##
    ## file './tools/maintainers-copy' should not exist on other computers
    ## because it is excluded by the build process and does not exist on GitHub
    if (!building && file.exists("./tools/maintainers-copy"))
        stop("must 'R CMD build' before 'R CMD INSTALL' since the files are destructively modified")


    if (!building) {


        make_backports.Rd <- function() {
            make_backports <- function(...) {
                x <- list(...)
                if (!length(x) || length(x) %% 3L)
                    stop("invalid arguments")
                version <- R_system_version(as.character(x[c(TRUE , FALSE, FALSE)]))
                alias <- I(lapply(x[c(FALSE, TRUE , FALSE)], as.character))
                usage <- I(lapply(x[c(FALSE, FALSE, TRUE )], as.character))
                data.frame(version, alias, usage)
            }
            backports <- make_backports(
                "3.0.0",
                c(".mapply", "list.files", "parse"),
                c(
                    ".mapply(FUN, dots, MoreArgs)",
                    "list.files(path = \".\", pattern = NULL, all.files = FALSE,",
                    "           full.names = FALSE, recursive = FALSE,",
                    "           ignore.case = FALSE, include.dirs = FALSE, no.. = FALSE)",
                    "parse(file = \"\", n = NULL, text = NULL, prompt = \"?\",",
                    "      keep.source = getOption(\"keep.source\"),",
                    "      srcfile = NULL, encoding = \"unknown\")"
                ),
                "3.1.0",
                c("anyNA", "anyNA.data.frame", "anyNA.numeric_version", "anyNA.POSIXlt"),
                c(
                    "anyNA(x, recursive = FALSE)",
                    "\\method{anyNA}{data.frame}(x, recursive = FALSE)",
                    "\\method{anyNA}{numeric_version}(x, recursive = FALSE)",
                    "\\method{anyNA}{POSIXlt}(x, recursive = FALSE)"
                ),
                "3.2.0",
                c("isNamespaceLoaded", "dir.exists", "file.copy", "lengths", "file.mtime", "file.size", "file.info"),
                c(
                    "isNamespaceLoaded(name)",
                    "dir.exists(paths)",
                    "file.copy(from, to, overwrite = recursive, recursive = FALSE,",
                    "          copy.mode = TRUE, copy.date = FALSE)",
                    "lengths(x, use.names = TRUE)",
                    "file.mtime(...)",
                    "file.size(...)",
                    "file.info(..., extra_cols = TRUE)"
                ),
                "3.3.0",
                c("strrep", "startsWith", "endsWith"),
                c(
                    "strrep(x, times)",
                    "startsWith(x, prefix)",
                    "endsWith(x, suffix)"
                ),
                "3.4.0",
                c("print.connection", "withAutoprint"),
                c(
                    "\\method{print}{connection}(x, ...)",
                    "withAutoprint(exprs, evaluated = FALSE, local = parent.frame(),",
                    "              print. = TRUE, echo = TRUE, max.deparse.length = Inf,",
                    "              width.cutoff = max(20, getOption(\"width\")),",
                    "              deparseCtrl = c(\"keepInteger\", \"showAttributes\", \"keepNA\"),",
                    "              skip.echo = 0, spaced = FALSE, ...)"
                ),
                "3.5.0",
                c("...length", "isTRUE", "isFALSE"),
                c(
                    "...length()",
                    "isTRUE(x)",
                    "isFALSE(x)"
                ),
                "3.6.0",
                c("errorCondition", "str2expression", "str2lang", "Sys.setFileTime"),
                c(
                    "errorCondition(message, ..., class = NULL, call = NULL)",
                    "str2expression(text)",
                    "str2lang(s)",
                    "Sys.setFileTime(path, time)"
                ),
                "4.0.0",
                c("deparse1", "unlink"),
                c(
                    "deparse1(expr, collapse = \" \", width.cutoff = 500L, ...)",
                    "unlink(x, recursive = FALSE, force = FALSE, expand = TRUE)"
                ),
                "4.1.0",
                c("...elt", "bquote"),
                c(
                    "...elt(n)",
                    "bquote(expr, where = parent.frame(), splice = FALSE)"
                ),
                "4.2.0",
                c("gettext", "gettextf"),
                c(
                    "gettext(..., domain = NULL, trim = TRUE)",
                    "gettextf(fmt, ..., domain = NULL, trim = TRUE)"
                )
            )
            backports <- backports[getRversion() < backports$version, , drop = FALSE]
            if (nrow(backports))
                writeLines(c(
                    "\\name{backports}",
            sprintf("\\alias{%s}", unlist(backports$alias)),
                    "\\title{Backports For Older R Versions}",
                    "\\description{",
            sprintf("  Reimplementations of functions introduced since \\R-\\code{%s}.", backports$version[1L]),
                    "}",
                    "\\usage{",
                    paste(
                        vapply(
                            seq_len(nrow(backports)),
                            function(i) {
                                value <- c(
            sprintf("# Introduced in R %s", backports[[i, "version"]]),
                    backports[[i, "usage"]]
                                )
                                paste(value, collapse = "\n")
                            },
                            ""
                        ),
                        collapse = "\n\n"
                    ),
                    "}",
                    "\\keyword{internal}"
                ), con = "./man/backports.Rd")
        }
        make_backports.Rd()


    }


    files <- list.files(all.files = TRUE, full.names = TRUE)
    files <- setdiff(files, c("./.", "./..", "./.git", "./.Rproj.user", "./build", "./tools"))
    files <- grep("\\.Rcheck$", files, value = TRUE, invert = TRUE)
    files <- grep("(\\.tar\\.gz|\\.zip|\\.tgz)$", files, value = TRUE, invert = TRUE)
    files <- unlist(lapply(files, function(file) {
        if (dir.exists(file))
            list.files(file, all.files = TRUE, full.names = TRUE, recursive = TRUE)
        else file
    }))
    files <- grep("\\.dat$", files, value = TRUE, invert = TRUE)


    gsub_from_DESCRIPTION <- function(x) {
        # x <- c(a = "testing", b = "@R_PACKAGE_NAME@ @R_PACKAGE_VERSION@ (@R_PACKAGE_DATE@)"); stop("comment out this later")
        patterns <- c(
            "@R_PACKAGE_NAME@", "@R_PACKAGE_VERSION@", "@R_PACKAGE_DATE@",
            "@R_PACKAGE_AUTHOR@", "@R_PACKAGE_MAINTAINER@",
            "@R_PACKAGE_LIB@"
        )
        replacements <- c(
            desc[c("Package", "Version", "Date")],
            RdQuote(desc[c("Author", "Maintainer")]),
            Library = gsub(".", "_", desc[["Package"]], fixed = TRUE)
        )
        Encoding(replacements) <- "bytes"
        pattern <- paste(regexQuote(patterns), collapse = "|")
        m <- gregexpr(pattern, x)
        regmatches(x, m) <- lapply(regmatches(x, m), function(xx) {
            Encoding(xx) <- "bytes"
            i <- match(xx, patterns, 0L)
            xx[i > 0L] <- replacements[i]
            xx
        })
        x
    }


    if (info[["renamed_files"]]) {
    } else {
        new <- gsub_from_DESCRIPTION(files)
        if (any(i <- files != new)) {
            old <- files[i]
            new <- new[i]
            files[i] <- new
            if (any(i <- !file.rename(old, new)))
                stop(sprintf(ngettext(sum(i), "unable to rename file %s",
                                              "unable to rename files %s"),
                     paste(sQuote(old), collapse = ", ")))
        }
        info[["renamed_files"]] <- "TRUE"
        write_info()
    }


    ## if we have already substituted the file contents, do not bother to read
    ## all files, only a small subset
    if (info[["substituted_file_contents"]]) {
        pattern <- paste(c(
            "^\\./NEWS$",
            "^\\./inst/NEWS\\.in\\.Rd$",
            if (getRversion() < "3.2.0")
                "^\\./man/[^/]+\\.Rd$",
            "^\\./R/0\\.R$",
            "^\\./src/devel\\.h$"
        ), collapse = "|")
        files <- grep(pattern, files, value = TRUE)
    }


    text <- vapply(
        files,
        function(x) readChar(x, file.size(x), useBytes = TRUE),
        ""
    )
    Encoding(text) <- "bytes"
    o <- text
    if (info[["substituted_file_contents"]]) {
    } else {
        text <- gsub_from_DESCRIPTION(text)
        info[["substituted_file_contents"]] <- "TRUE"
    }
    if (info[["reindented_NEWS"]]) {
    } else {
        local({
            if (!devel)
                if (i <- match("./NEWS", names(text), 0L)) {
                    m <- gregexpr("(?<=^|\r\n|[\r\n])(  )?(?![ \r\n])", text[i], perl = TRUE)
                    regmatches(text[i], m) <<- lapply(regmatches(text[i], m), function(x) {
                        i <- nzchar(x)
                        x[!i] <- "                 "  ## indent sections by 17
                        x[ i] <- ""                   ## do not indent subsections
                        x[1L] <- ""                   ## do not indent title
                        x
                    })
                    info[["reindented_NEWS"]] <<- "TRUE"
                }
        })
    }


    if (!building) {
        ## we need to add the common macros to the files which do not
        ## have access. for R < 3.2.0, this is ALL of the Rd files.
        ## otherwise, only the news files will need them
        macros <- unlist(lapply(
            list.files("./man/macros", "\\.Rd$", all.files = TRUE, full.names = TRUE, ignore.case = TRUE),
            function(file) {
                conn <- file(file, "rb")
                on.exit(close(conn))
                readLines(conn)
            }
        ))
        if (length(macros)) {
            Encoding(macros) <- "bytes"
            pattern <- "^\\./inst/NEWS\\.in\\.Rd$"
            if (getRversion() < "3.2.0")
                pattern <- paste(pattern, "^\\./man/[^/]+\\.Rd$", sep = "|")
            keep <- grep(pattern, names(text))
            Rdtext <- text[keep]
            # Rdtext <- readChar("./inst/NEWS.in.Rd", file.size("./inst/NEWS.in.Rd"), useBytes = TRUE); stop("comment out this later")
            m <- gregexpr("(?<=^|\r\n|[\r\n])\\\\title\\{", Rdtext, perl = TRUE)
            if (any(i <- vapply(m, length, 0L, USE.NAMES = FALSE) > 1L))
                stop(gettextf("in %s:\n multiple lines that start with \\title{",
                     paste(sQuote(names(Rdtext)[i]), collapse = ", ")))
            if (any(i <- as.integer(m) == -1L))
                stop(gettextf("in %s:\n no lines that start with \\title{",
                     paste(sQuote(names(Rdtext)[i]), collapse = ", ")))
            regmatches(Rdtext, m) <- mapply(
                regmatches(Rdtext, m),
                ifelse(grepl("\r\n", Rdtext), "\r\n", "\n"),
                FUN = function(xx, line_end) {
                    paste0(
                        paste0(macros, line_end, collapse = ""),
                        xx
                    )
                },
                SIMPLIFY = FALSE, USE.NAMES = FALSE
            )
            text[keep] <- Rdtext
        }
        ## remove the macros folder so 'R CMD check' will not complain about it
        if (getRversion() < "3.2.0")
            unlink("./man/macros", recursive = TRUE, force = TRUE)
    }


    if (!building) {
        ## (possibly) enable development features
        if (devel) {
            text["./R/0.R"] <- paste(
                "options(R_THIS_PATH_DEVEL = TRUE)",
                if (grepl("\r\n", text[["./R/0.R"]])) "\r\n" else "\n",
                text[["./R/0.R"]],
                sep = ""
            )
            text["./src/devel.h"] <- paste(
                "#ifndef R_THIS_PATH_DEVEL",
                "#define R_THIS_PATH_DEVEL",
                "#endif",
                text[["./src/devel.h"]],
                ## *.c and *.h must use Unix (LF) under Unix-alikes
                sep = "\n"
            )
        }
    }


    if (any(i <- o != text)) {
        mapply(function(description, text) {
            ## "wb" does not convert line endings
            conn <- file(description, "wb")
            on.exit(close(conn))
            writeLines(text, conn, sep = "", useBytes = TRUE)
        }, names(text)[i], text[i], SIMPLIFY = FALSE, USE.NAMES = FALSE)
    }
    write_info()


    if (info[["replaced_CRLF_with_LF"]]) {
    } else {
        ## change the line endings to \n
        local({
            file <- "./tools/this_path_reg_ptrs.c"
            text <- readChar(file, file.size(file), useBytes = TRUE)
            Encoding(text) <- "bytes"
            text <- gsub("\r\n", "\n", text, fixed = TRUE, useBytes = TRUE)
            conn <- file(file, "wb")
            on.exit(close(conn))
            writeLines(text, conn, sep = "", useBytes = TRUE)
        })
        info[["replaced_CRLF_with_LF"]] <- "TRUE"
        write_info()
    }


    if (!building) {
        if (!devel) {
            make_this_path_reg_ptrs_lib <- function() {
                run_shlib <- function(arch) {
                    chname <- "this_path_reg_ptrs"
                    chname1 <- paste0(chname, .Platform$dynlib.ext)
                    DLLpath <- DLLpath_top <- "./dlls"
                    if (nzchar(arch))
                        DLLpath <- paste0(DLLpath, arch)
                    dir.create(DLLpath, showWarnings = FALSE, recursive = TRUE)
                    file <- paste0(DLLpath, "/", chname1)
                    if (.Platform$OS.type == "windows") {
                        args <- paste0(R.home(), "/bin", arch, "/Rcmd.exe")
                    } else {
                        args <- paste0(R.home(), "/bin/R")
                        if (nzchar(arch))
                            args <- c(args, paste0("--arch=", substr(arch, 2, 1000)))
                        args <- c(args, "CMD")
                    }
                    args <- c(args, "SHLIB", paste0("--output=", file),
                        "--clean", "--preclean", paste0("./tools/", chname, ".c"))
                    command <- paste(shQuote(args), collapse = " ")
                    # cat("$ ", command, "\n", sep = "")
                    system(command, ignore.stdout = TRUE, ignore.stderr = TRUE)
                    if (!length(dir(DLLpath))) {
                        unlink(DLLpath, recursive = TRUE, force = TRUE)
                        if (DLLpath_top != DLLpath && !length(dir(DLLpath_top)))
                            unlink(DLLpath_top, recursive = TRUE, force = TRUE)
                    }
                }
                if (.Platform$OS.type == "windows") {
                    f <- dir(paste0(R.home(), "/bin"))
                    archs <- f[f %in% c("i386", "x64")]
                    for (arch in archs) {
                        ra <- paste0("/", arch)
                        run_shlib(ra)
                    }
                } else {
                    # wd2 <- setwd(paste0(R.home(), "/bin/exec"))
                    # archs <- Sys.glob("*")
                    # setwd(wd2)
                    archs <- dir(paste0(R.home(), "/bin/exec"))
                    for (arch in archs) {
                        if (arch == "R") {
                            run_shlib("")
                        } else {
                            ra <- paste0("/", arch)
                            run_shlib(ra)
                        }
                    }
                }
                if (file.exists("./dlls")) {
                    file.copy("./dlls", Sys.getenv("R_PACKAGE_DIR"), recursive = TRUE)
                    unlink("./dlls", recursive = TRUE, force = TRUE)
                }
            }
            make_this_path_reg_ptrs_lib()
        }
    }


    if (!building) {
        old <- "./inst/NEWS.in.Rd"
        new <- "./inst/NEWS.Rd"
        if (!file.rename(old, new))
            stop(sprintf("unable to rename file '%s'", old))
    }


    invisible()
}


main()

##
## this.path : Get Executing Script's Path
## Copyright (C) 2022-2024   Iris Simmons
##


check <- function (path, dirname, basename)
{
    stopifnot(
        identical(dirname2(path), dirname),
        identical(basename2(path), basename)
    )
}


dirname2  <- this.path:::.windows_dirname2
basename2 <- this.path:::.windows_basename2


check("", "", "")


check("d:", "d:.", "")
check("//host/share", "//host/share", "")


check("d:/", "d:/", "")
check("//host/share/", "//host/share/", "")
check("/", "/", "")


check("d:/path", "d:/", "path")
check("//host/share/path", "//host/share/", "path")
check("/path", "/", "path")


check("d:file", "d:.", "file")
check("file", ".", "file")
check("d:path/to/file", "d:path/to", "file")


check("d:/path/to/file", "d:/path/to", "file")
check("//host/share/path/to/file", "//host/share/path/to", "file")
check("/path/to/file", "/path/to", "file")
check("/path/to/file/", "/path/to", "file")


check(
    this.path::path.join("/", "p1", "p2", "p3", c("file1", "file2")),
    rep("/p1/p2/p3", 2),
    c("file1", "file2")
)
check(
    this.path::path.join("/", "p1", "p2", "p3", "filename"),
    "/p1/p2/p3",
    "filename"
)


check(
    c("/usr/lib", "/usr/", "usr", "/", ".", ".."),
    c("/usr"    , "/"    , "."  , "/", ".", "."),
    c("lib"     , "usr"  , "usr", "" , ".", "..")
)


dirname2  <- this.path:::.unix_dirname2
basename2 <- this.path:::.unix_basename2


check("", "", "")


check("//host/share", "//host/share", "")


check("//host/share/", "//host/share/", "")
check("/", "/", "")


check("//host/share/path", "//host/share/", "path")
check("/path", "/", "path")


check("file", ".", "file")


check("//host/share/path/to/file", "//host/share/path/to", "file")
check("/path/to/file", "/path/to", "file")
check("/path/to/file/", "/path/to", "file")


check(
    this.path::path.join("/", "p1", "p2", "p3", c("file1", "file2")),
    rep("/p1/p2/p3", 2),
    c("file1", "file2")
)
check(
    this.path::path.join("/", "p1", "p2", "p3", "filename"),
    "/p1/p2/p3",
    "filename"
)


check(
    c("/usr/lib", "/usr/", "usr", "/", ".", ".."),
    c("/usr"    , "/"    , "."  , "/", ".", "."),
    c("lib"     , "usr"  , "usr", "" , ".", "..")
)

##
## this.path : Get Executing Script's Path
## Copyright (C) 2024   Iris Simmons
##


if (requireNamespace("box", quietly = TRUE)) {
    ## tests for with_script_path                               ----


    oscript_path <- box::script_path()
    this.path::with_script_path(box::script_path())
    ## the script_path must be restored to its original value
    stopifnot(identical(oscript_path, box::script_path()))


    ## check the namespace is never loaded if the user did not request it so
    unloadNamespace("box")
    stopifnot(!isNamespaceLoaded("box"))
    this.path::with_script_path(stopifnot(!isNamespaceLoaded("box")))
    stopifnot(!isNamespaceLoaded("box"))


    ## check the namespace hooks are not changed (except while in the function)
    hooks <- getHook(packageEvent("box"))
    this.path::with_script_path({
        stopifnot(length(hooks) + 1L == length(getHook(packageEvent("box"))))
        stopifnot(identical(hooks, getHook(packageEvent("box"))[seq_along(hooks)]))
    })
    stopifnot(identical(hooks, getHook(packageEvent("box"))))


    ## tests for "tests/package-box-supp/startsWith-NULL.R"     ----


    this.path::with_script_path(
    box::use(
        mod = ./`package-box-supp`/`startsWith-NULL`
    )
    )


    srcref <- attr(mod$fun_that_calls_src_path, "srcref")
    stopifnot(!is.null(srcref))
    srcfile <- attr(srcref, "srcfile")
    stopifnot(!is.null(srcfile))
    ## we expect the document context to not exist just yet
    stopifnot(is.null(srcfile[[".this.path::document.context"]]))


    namespace <- attr(mod, "namespace")
    stopifnot(!is.null(namespace))
    ## we expect the document context to not exist just yet
    stopifnot(is.null(attr(namespace, ".this.path::document.context")))


    mod$fun_that_calls_src_path(verbose = FALSE)
    doc_cntxt <- srcfile[[".this.path::document.context"]]
    ## now we do expect the document context to exist
    stopifnot(!is.null(doc_cntxt))
    ## we expect the source to be "path of srcfile"
    stopifnot(identical(
        as.character(list(doc_cntxt$source)),
        "path of srcfile"
    ))


    ## we expect the document context to still not exist
    stopifnot(is.null(attr(namespace, ".this.path::document.context")))


    mod$fun_that_calls_env_path(verbose = FALSE)
    doc_cntxt_2 <- attr(namespace, ".this.path::document.context")
    ## now we do expect the document context to exist
    stopifnot(!is.null(doc_cntxt_2))
    ## we expect the source to be "path of a 'package:box' namespace"
    stopifnot(identical(
        as.character(list(doc_cntxt_2$source)),
        "path of a 'package:box' namespace"
    ))


    ## we expect the contexts to not be equal
    stopifnot(!identical(doc_cntxt, doc_cntxt_2))


    box::unload(mod)


    ## tests for "tests/package-box-supp/startsWith-sys-path.R" ----


    this.path::with_script_path(
    box::use(
        mod = ./`package-box-supp`/`startsWith-sys-path`
    )
    )


    ## we expect the source to be
    ## "call to function 'load_from_source' from package 'box'"
    stopifnot(identical(
        as.character(list(mod$doc_cntxt$source)),
        "call to function 'load_from_source' from package 'box'"
    ))


    srcref <- attr(mod$fun_that_calls_src_path, "srcref")
    stopifnot(!is.null(srcref))
    srcfile <- attr(srcref, "srcfile")
    stopifnot(!is.null(srcfile))
    ## we expect the document context to be identical to that of sys.path()
    stopifnot(identical(mod$doc_cntxt, srcfile[[".this.path::document.context"]]))


    namespace <- attr(mod, "namespace")
    stopifnot(!is.null(namespace))
    ## we expect the document context to be identical to that of sys.path()
    stopifnot(identical(mod$doc_cntxt, attr(namespace, ".this.path::document.context")))


    box::unload(mod)


    ## tests for "tests/package-box-supp/startsWith-env-path.R" ----


    this.path::with_script_path(
    box::use(
        mod = ./`package-box-supp`/`startsWith-env-path`
    )
    )


    ## we expect the source to be "path of a 'package:box' namespace"
    stopifnot(identical(
        as.character(list(mod$doc_cntxt$source)),
        "path of a 'package:box' namespace"
    ))


    srcref <- attr(mod$fun_that_calls_src_path, "srcref")
    stopifnot(!is.null(srcref))
    srcfile <- attr(srcref, "srcfile")
    stopifnot(!is.null(srcfile))
    ## we expect the document context to be identical to that of env.path()
    stopifnot(identical(mod$doc_cntxt, srcfile[[".this.path::document.context"]]))


    namespace <- attr(mod, "namespace")
    stopifnot(!is.null(namespace))
    ## we expect the document context to be identical to that of env.path()
    stopifnot(identical(mod$doc_cntxt, attr(namespace, ".this.path::document.context")))


    box::unload(mod)


    ## tests for "tests/package-box-supp/startsWith-src-path.R" ----


    this.path::with_script_path(
    box::use(
        mod = ./`package-box-supp`/`startsWith-src-path`
    )
    )


    ## we expect the source to be "path of srcfile"
    stopifnot(identical(
        as.character(list(mod$doc_cntxt$source)),
        "path of srcfile"
    ))


    srcref <- attr(mod$fun_that_calls_src_path, "srcref")
    stopifnot(!is.null(srcref))
    srcfile <- attr(srcref, "srcfile")
    stopifnot(!is.null(srcfile))
    ## we expect the document context to be identical to that of src.path()
    stopifnot(identical(mod$doc_cntxt, srcfile[[".this.path::document.context"]]))


    namespace <- attr(mod, "namespace")
    stopifnot(!is.null(namespace))
    ## we expect the document context to be identical to that of src.path()
    stopifnot(identical(mod$doc_cntxt, attr(namespace, ".this.path::document.context")))


    box::unload(mod)
} else cat("\n'package:box' is not available :(\n")

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2024   Iris Simmons
##


# cat("\n> sys.calls()\n"); print(sys.calls())
# cat("\n> sys.parents()\n"); print(sys.parents())
# cat("\n> sys.frames()\n"); print(sys.frames())


.IS_SCALAR_STR <- function (x)
is.character(x) && length(unclass(x)) == 1L


.removeSource <- function (fn)
{
    recurse <- function(part) {
        if (is.name(part)) return(part)
        if (inherits(part, "srcref")) return(NULL)
        attr(part, "srcref") <- NULL
        attr(part, "wholeSrcref") <- NULL
        attr(part, "srcfile") <- NULL
        if (is.pairlist(part)) {
            for (i in seq_along(part))
                part[i] <- list(recurse(part[[i]]))
            as.pairlist(part)
        }
        else if (is.language(part) && is.recursive(part)) {
            for (i in seq_along(part))
                part[i] <- list(recurse(part[[i]]))
            part
        }
        else part
    }
    if (is.function(fn)) {
        if (!is.primitive(fn)) {
            attr(fn, "srcref") <- NULL
            at <- attributes(fn)
            formals(fn) <- recurse(formals(fn))
            attr(body(fn), "wholeSrcref") <- NULL
            attr(body(fn), "srcfile") <- NULL
            body(fn) <- recurse(body(fn))
            if (!is.null(at))
                attributes(fn) <- at
        }
        fn
    }
    else if (is.language(fn)) {
        recurse(fn)
    }
    else stop("argument is not a function or language object:", typeof(fn))
}


.R_FunctionSymbol <- as.symbol("function")


.removeSource_from_inner_functions <- function (fn)
{
    ## allow a function/language object to keep its source references
    ## while removing source references from any sub functions
    recurse <- function(part) {
        if (is.name(part))
            part
        else if (is.call(part) && identical(part[[1L]], .R_FunctionSymbol))
            .removeSource(part)
        else if (is.language(part) && is.recursive(part)) {
            for (i in seq_along(part))
                part[i] <- list(recurse(part[[i]]))
            part
        }
        else part
    }
    if (is.function(fn)) {
        if (!is.primitive(fn)) {
            at <- attributes(fn)
            formals(fn) <- lapply(formals(fn), recurse)
            body(fn) <- recurse(body(fn))
            if (!is.null(at))
                attributes(fn) <- at
        }
        fn
    }
    else if (is.language(fn)) {
        recurse(fn)
    }
    else stop("argument is not a function or language object:", typeof(fn))
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2024   Iris Simmons
##


.asArgs <- function (x)
{
    if (!length(x)) return(character())
    value <- rapply(x, function(xx) {
        if (is.null(xx))
            character()
        else if (is.object(xx)) {
            if (inherits(xx, "factor"))
                as.character.factor(xx)
            else if (inherits(xx, "POSIXct"))
                format.POSIXct(xx, format = "%Y-%m-%d %H:%M:%OS6")
            else if (inherits(xx, "POSIXlt"))
                format.POSIXlt(xx, format = "%Y-%m-%d %H:%M:%OS6")
            else as.character(xx)
        }
        else if (!length(xx))
            character()
        else if (is.numeric(xx) || is.complex(xx))
            format.default(xx, trim = TRUE, digits = 17L, decimal.mark = ".",
                drop0trailing = TRUE)
        else if (is.raw(xx))
            sprintf("0x%02x", as.integer(xx))
        else if (is.pairlist(xx))
            .asArgs(as.vector(xx, "list"))
        else as.character(xx)
    }, how = "list")
    value <- unlist(value, recursive = TRUE, use.names = FALSE)
    # if (anyNA(value))
    #     value[is.na(value)] <- "NA"
    if (any(i <- is.na(value))) value[i] <- "NA"
    return(value)
}


asArgs <- function (...)
.asArgs(list(...))


progArgs <- function (ifnotfound = character())
{
    n <- .External2(.C_getframenumber)
    if (n) {
        if (n > .toplevel.nframe()) {
            whiches <- .External2(.C_sys.whiches, n)
            ## remove the last element and reverse
            if (n <- length(whiches))
                whiches <- whiches[seq.int(to = 1L, by = -1L, length.out = n - 1L)]
            for (which in whiches) {
                if (.identical(sys.function(which), withArgs)) {
                    value <- get("args", envir = sys.frame(which), inherits = FALSE)
                    return(value)
                }
            }
            ifnotfound
        }
        ## in the site-wide startup profile file or a user profile
        else ifnotfound
    }
    else if (.in_shell)
        commandArgs(trailingOnly = TRUE)
    else ifnotfound
}


withArgs <- function (...)
{
    args <- .External2(.C_asArgs, 1L)
    ..1
}

##
## this.path : Get Executing Script's Path
## Copyright (C) 2021-2025   Iris Simmons
##


## functions to be converted to active bindings from '.onLoad' ----


.mbcslocale <- function ()
.External2(.C_mbcslocale)


.utf8locale <- function ()
l10n_info()[[2L]]


.latin1locale <- function ()
l10n_info()[[3L]]


.R_MB_CUR_MAX <- function ()
.External2(.C_R_MB_CUR_MAX)


## functions to be run from '.onLoad' ----


.fix_utils_Sweave <- function (pkgname, pkgpath)
{
    ## the arguments are unused, they only exist to match the arguments
    ## provided to a user hook (see loadNamespace, specifically runUserHook)
    nsname <- "utils"
    ns <- .getNamespace(nsname)
    sym <- "Sweave"
    fun <- ns[[sym]]
    if (typeof(fun) == "closure" &&
        is.call(b <- body(fun)) && length(b) >= 7L &&
        identical(b[[6L]], quote(text <- SweaveReadFile(file, syntax, encoding = encoding))) &&
        identical(b[[7L]], quote(attr(file, "encoding") <- encoding <- attr(text, "encoding"))))
    {
        body(fun) <- as.call(append(as.list(b), after = 6L,
            expression(file <- attr(text, "files")[1L])
        ))
        if (bindingIsLocked(sym, ns)) {
            (unlockBinding)(sym, ns)
            assign(sym, fun, envir = ns, inherits = FALSE)
            lockBinding(sym, ns)
        }
        else assign(sym, fun, envir = ns, inherits = FALSE)
    }
    invisible()
}


.fix_utils_RweaveLatexRuncode <- function (pkgname, pkgpath)
{
    ## the arguments are unused, they only exist to match the arguments
    ## provided to a user hook (see loadNamespace, specifically runUserHook)
    nsname <- "utils"
    ns <- .getNamespace(nsname)
    sym <- "RweaveLatexRuncode"
    fun <- ns[[sym]]
    if (typeof(fun) == "closure" &&
        is.call(b <- body(fun)) && length(b) >= 34L &&
        # is.call(b_33 <- b[[33L]]) && length(b_33) >= 4L &&
        ## compatibility with R < 3.1
        is.call(b_33 <- b[[indx33 <- 33L + identical(b[[29L]], quote(refline <- NA))]]) && length(b_33) >= 4L &&
        is.call(b_33_4 <- b_33[[4L]]) && length(b_33_4) >= 6L &&
        is.call(b_33_4_6 <- b_33_4[[6L]]) && length(b_33_4_6) >= 3L &&
        is.call(b_33_4_6_3 <- b_33_4_6[[3L]]) && length(b_33_4_6_3) >= 4L &&
        is.call(b_33_4_6_3_4 <- b_33_4_6_3[[4L]]) && length(b_33_4_6_3_4) >= 3L &&
        is.call(b_33_4_6_3_4_3 <- b_33_4_6_3_4[[3L]]) && length(b_33_4_6_3_4_3) >= 2L)
    {
        i <- NULL
        if (is.call(b_33_4_6_3_4_3_2 <- b_33_4_6_3_4_3[[2L]])) {
            if (length(b_33_4_6_3_4_3_2) >= 2L &&
                identical(b_33_4_6_3_4_3_2[[2L]], quote(ce)))
            {
                i <- c(indx33, 4L, 6L, 3L, 4L, 3L, 2L, 2L)
            }
        }
        ## compatibility with R < 3.0
        else if (identical(b_33_4_6_3_4_3_2, quote(ce))) {
            i <- c(indx33, 4L, 6L, 3L, 4L, 3L, 2L)
        }
        if (!is.null(i)) {
            b[[i]] <- quote(chunkexps[nce])
            body(fun) <- b
            if (bindingIsLocked(sym, ns)) {
                (unlockBinding)(sym, ns)
                assign(sym, fun, envir = ns, inherits = FALSE)
                lockBinding(sym, ns)
            }
            else assign(sym, fun, envir = ns, inherits = FALSE)
        }
    }
    invisible()
}


.fix_utils_makeRweaveLatexCodeRunner <- function (pkgname, pkgpath)
{
    ## the arguments are unused, they only exist to match the arguments
    ## provided to a user hook (see loadNamespace, specifically runUserHook)
    nsname <- "utils"
    ns <- .getNamespace(nsname)
    sym <- "makeRweaveLatexCodeRunner"
    fun <- ns[[sym]]
    if (typeof(fun) == "closure" &&
        is.call(b <- body(fun)) && length(b) == 2L &&
        is.call(b_2 <- b[[2L]]) && length(b_2) >= 3L &&
        is.call(b_2_3 <- b_2[[3L]]) && length(b_2_3) >= 34L &&
        # is.call(b_2_3_33 <- b_2_3[[33L]]) && length(b_2_3_33) >= 4L &&
        ## compatibility with R < 3.1
        is.call(b_2_3_33 <- b_2_3[[indx33 <- 33L + identical(b_2_3[[29L]], quote(refline <- NA))]]) && length(b_2_3_33) >= 4L &&
        is.call(b_2_3_33_4 <- b_2_3_33[[4L]]) && length(b_2_3_33_4) >= 6L &&
        is.call(b_2_3_33_4_6 <- b_2_3_33_4[[6L]]) && length(b_2_3_33_4_6) >= 3L &&
        is.call(b_2_3_33_4_6_3 <- b_2_3_33_4_6[[3L]]) && length(b_2_3_33_4_6_3) >= 4L &&
        is.call(b_2_3_33_4_6_3_4 <- b_2_3_33_4_6_3[[4L]]) && length(b_2_3_33_4_6_3_4) >= 3L &&
        is.call(b_2_3_33_4_6_3_4_3 <- b_2_3_33_4_6_3_4[[3L]]) && length(b_2_3_33_4_6_3_4_3) >= 2L)
    {
        i <- NULL
        if (is.call(b_2_3_33_4_6_3_4_3_2 <- b_2_3_33_4_6_3_4_3[[2L]])) {
            if (length(b_2_3_33_4_6_3_4_3_2) >= 2L &&
                identical(b_2_3_33_4_6_3_4_3_2[[2L]], quote(ce)))
            {
                i <- c(2L, 3L, indx33, 4L, 6L, 3L, 4L, 3L, 2L, 2L)
            }
        }
        ## compatibility with R < 3.0
        else if (identical(b_2_3_33_4_6_3_4_3_2, quote(ce))) {
            i <- c(2L, 3L, indx33, 4L, 6L, 3L, 4L, 3L, 2L)
        }
        if (!is.null(i)) {
            b[[i]] <- quote(chunkexps[nce])
            body(fun) <- b
            if (bindingIsLocked(sym, ns)) {
                (unlockBinding)(sym, ns)
                assign(sym, fun, envir = ns, inherits = FALSE)
                lockBinding(sym, ns)
            }
            else assign(sym, fun, envir = ns, inherits = FALSE)
        }
    }
    invisible()
}


.fix_utils <- function (pkgname, pkgpath)
{
    .fix_utils_Sweave(pkgname, pkgpath)
    .fix_utils_RweaveLatexRuncode(pkgname, pkgpath)
    .fix_utils_makeRweaveLatexCodeRunner(pkgname, pkgpath)
}


.fix_plumber_parseUTF8 <- function (pkgname, pkgpath)
{
    ## the arguments are unused, they only exist to match the arguments
    ## provided to a user hook (see loadNamespace, specifically runUserHook)
    nsname <- "plumber"
    ns <- .getNamespace(nsname)
    sym <- "parseUTF8"
    fun <- ns[[sym]]
    if (typeof(fun) == "closure" &&
        is.call(b <- body(fun)) && length(b) >= 6L &&
        identical(b[[6L]], quote(exprs <- try(parse(file, keep.source = TRUE, srcfile = src, encoding = enc)))))
    {
        b[[6L]] <- quote(exprs <- try(parse(file, keep.source = FALSE, srcfile = src, encoding = enc)))
        body(fun) <- b
        if (bindingIsLocked(sym, ns)) {
            (unlockBinding)(sym, ns)
            assign(sym, fun, envir = ns, inherits = FALSE)
            lockBinding(sym, ns)
        }
        else assign(sym, fun, envir = ns, inherits = FALSE)
    }
    invisible()
}


.maybe_setHook_packageEvent_utils_fix_utils <- function ()
{
    hookName <- packageEvent("utils")
    old <- getHook(hookName)
    new <- list(.fix_utils_Sweave, .fix_utils_RweaveLatexRuncode, .fix_utils_makeRweaveLatexCodeRunner)
    if (!length(old))
        setHook(hookName, new, "prepend")
    else if (any(i <- !(new %in% old)))
        setHook(hookName, new[i], "prepend")
}


.maybe_setHook_packageEvent_plumber_fix_plumber_parseUTF8 <- function ()
{
    hookName <- packageEvent("plumber")
    old <- getHook(hookName)
    new <- list(.fix_plumber_parseUTF8)
    if (!length(old) || !(new %in% old))
        setHook(hookName, new, "prepend")
}





if (!isTRUE(getOption("R_THIS_PATH_DEVEL"))) {
tmp <- new.env()
.get_ptrs <- evalq(envir = tmp, {
    x <- NULL
             function ()
{
    chname <- "this_path_reg_ptrs"
    path <- getNamespaceInfo("this.path", "path")
    path <- paste0(path, "/dlls")
    r_arch <- .Platform$r_arch
    chname1 <- paste0(chname, .Platform$dynlib.ext)
    DLLpath <- if (nzchar(r_arch)) paste0(path, "/", r_arch) else path
    file <- paste0(DLLpath, "/", chname1)
    if (file.exists(file)) {
        dyn.load(file, DLLpath = DLLpath)
        x <<- file
        .External2(.C_get_ptrs)
    }
}
})
.maybe_dyn_unload <- evalq(envir = tmp, {
                     function ()
if (!is.null(x)) dyn.unload(x)
})
rm(tmp)
}


## '.onLoad' and '.onUnload' ----


.onLoad <- function (libname, pkgname)
.External2(.C_onLoad, libname, pkgname)


.onUnload <- function (libpath)
.External2(.C_onUnload, libpath)

##
## this.path : Get Executing Script's Path
## Copyright (C) 2024   Iris Simmons
##


.shFILE <- evalq(envir = new.env(), {
    delayedAssign("ofile", { if (.in_shell) .shINFO[["FILE"]] else NA_character_ })
    delayedAssign("file" , { .normalizePath(ofile) })
           function (original = TRUE, for.msg = FALSE)
.External2(.C_shFILE, original, for.msg)
})


delayedAssign(".has_shFILE", { !is.na(.shFILE()) })


shFILE <- function (original = FALSE, for.msg = FALSE, default, else.)
{
    if (missing(default)) {
        if (missing(else.))
            .shFILE(original, for.msg)
        else stop("'shFILE' with 'else.' but not 'default' makes no sense")
    }
    else {
        if (missing(else.)) {
            if (.has_shFILE)
                .shFILE(original, for.msg)
            else if (for.msg)
                NA_character_
            else default
        }
        else {
            if (.has_shFILE) {
                value <- .shFILE(original, for.msg)
                (else.)(value)
            }
            else if (for.msg) {
                value <- NA_character_
                (else.)(value)
            }
            else default
        }
    }
}


try.shFILE <- function ()
tryCatch(.shFILE(FALSE), error = function(e) .shFILE())

##
## this.path : Get Executing Script's Path
## Copyright (C) 2021-2024   Iris Simmons
##


# URL_pattern <- "^((?:ftp|ftps|http|https)://[^/]+)(?:/+(.*))?$"
# ##              ^                                            ^ start and end of the string
# ##                   ^^^^^^^^^^^^^^^^^^^^^^^                   ftp:// or ftps:// or http:// or https://
# ##                                          ^^^^^              series of non-slash characters
# ##               ^                               ^             the root of the URL
# ##                                                   ^^^^^^    more paths
# ##                                                          ^  more paths are optional
# ##                 ^^                              ^^          non-capturing groups


# UNC_pattern <- "^(?:(//[^/]+)/+([^/]+))(?:/+(.*))?$"
# ##              ^                                 ^ start and end of the string
# ##                   ^^                             two slashes
# ##                     ^^^^^    ^^^^^               series of non-slash characters, the host and share name
# ##                           ^^                     one or more slashes
# ##                                        ^^^^^^    more paths
# ##                                               ^  more paths are optional
# ##                ^^                    ^^          non-capturing groups


# c(
#     "https://raw.githubusercontent.com/ArcadeAntics/this.path/main/tests/this.path_w_URLs.R",
#     "//host/share/path/to/file"
# )


check.path <- function (...)
{
    expected <- path.join(...)
    if (!.IS_SCALAR_STR(expected))
        stop(gettextf("'%s' must be a character string", "expected", domain = "R"))
    if (!nzchar(expected))
        stop(gettextf("'%s' must not be \"\"", "expected"))
    expected <- path.split.1(expected)
    if (check_wd <- expected[[1L]] == ".")
        expected <- expected[-1L]
    thispath <- .External2(.C_this_path)
    thispath <- path.split.1(thispath)
    if (length(thispath) < length(expected) || {
        i <- seq_along(thispath) > length(thispath) - length(expected)
        any(thispath[i] != expected)
    })
    {
        expected <- path.unsplit(expected)
        thispath <- path.unsplit(thispath[i])
        stop("'this.path()' and expected path do not match\n",
            paste0(c("* this.path(): ", "* expected   : "), {
                encodeString(c(thispath, expected), quote = "\"")
            }, collapse = "\n"))
    }
    if (check_wd) {
        expected <- path.unsplit(thispath[!i])
        if (relpath(expected) != ".") {
            stop("'getwd()' and expected path do not match\n",
                paste0(c("* getwd() : ", "* expected: "), {
                    encodeString(c(getwd(), expected), quote = "\"")
                }, collapse = "\n"))
        }
    }
    invisible(TRUE)
}


check.dir <- function (...)
{
    expected <- path.join(...)
    if (!.IS_SCALAR_STR(expected))
        stop(gettextf("'%s' must be a character string", "expected", domain = "R"))
    if (!nzchar(expected))
        stop(gettextf("'%s' must not be \"\"", "expected"))
    expected <- path.split.1(expected)
    if (check_wd <- expected[[1L]] == ".")
        expected <- expected[-1L]
    thisdir <- .External2(.C_this_path)
    thisdir <- .dir(thisdir)
    thisdir <- path.split.1(thisdir)
    if (length(thisdir) < length(expected) || {
        i <- seq_along(thisdir) > length(thisdir) - length(expected)
        any(thisdir[i] != expected)
    })
    {
        expected <- path.unsplit(expected)
        thisdir <- path.unsplit(thisdir[i])
        stop("'this.dir()' and expected path do not match\n",
            paste0(c("* this.dir(): ", "* expected  : "), {
                encodeString(c(thisdir, expected), quote = "\"")
            }, collapse = "\n"))
    }
    if (check_wd) {
        expected <- path.unsplit(thisdir[!i])
        if (relpath(expected) != ".") {
            stop("'getwd()' and expected path do not match\n",
                paste0(c("* getwd() : ", "* expected: "), {
                  encodeString(c(getwd(), expected), quote = "\"")
                }, collapse = "\n"))
        }
    }
    invisible(TRUE)
}


check.proj <- function (...)
{
    expected <- path.join(...)
    if (!.IS_SCALAR_STR(expected))
        stop(gettextf("'%s' must be a character string", "expected", domain = "R"))
    if (!nzchar(expected))
        stop(gettextf("'%s' must not be \"\"", "expected"))
    expected <- path.split.1(expected)
    if (check_wd <- expected[[1L]] == ".")
        expected <- expected[-1L]
    thispath <- .External2(.C_this_path)
    thisproj <- .proj(.dir(thispath))
    thispath <- path.split.1(thispath)
    thisproj <- path.split.1(thisproj)
    i <- seq_along(thispath) > length(thisproj)
    if (sum(i) != length(expected) || any(thispath[i] != expected))  {
        expected <- path.unsplit(expected)
        thispath <- path.unsplit(thispath[i])
        stop("path within project and expected path do not match\n",
             paste0(c("* project path: ", "* expected    : "), {
                 encodeString(c(thispath, expected), quote = "\"")
             }, collapse = "\n"))
    }
    if (check_wd) {
        expected <- path.unsplit(thispath[!i])
        if (relpath(expected) != ".") {
            stop("'getwd()' and expected path do not match\n",
                 paste0(c("* getwd() : ", "* expected: "), {
                     encodeString(c(getwd(), expected), quote = "\"")
                 }, collapse = "\n"))
        }
    }
    invisible(TRUE)
}


# path.split(c(
#     this.path(),
#     "~/this.path",
#     "testing//this/out",
#     "/Users/testing/this/out",
#     "//host-name/share-name/path/to/file",
#     "ftp://host/path/to/file",
#     "ftps://host/path/to/file",
#     "http://host/path/to/file",
#     "https://host/path/to/file"
# ))
#
#
# check.path("this.path/R/checkpath.R")
# check.path("this.path/R/checkpath.")
# check.dir("this.path/R")
# check.dir("this.path/r")

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2024   Iris Simmons
##


local({


    FILE.R <- tempfile(fileext = ".R")
    on.exit(unlink(FILE.R))
    writeLines("
test.this.path <- function ()
{
    try(print(this.path::sys.path(verbose = TRUE)))
    try(print(this.path::env.path(verbose = TRUE)))
    try(print(this.path::src.path(verbose = TRUE)))
    try(print(this.path::this.path(verbose = TRUE)))
}


test.this.path()", FILE.R)


    sourcelike <- function(file) {
        ofile <- file
        file <- this.path::set.sys.path(file, Function = "sourcelike")
        lines <- readLines(file, warn = FALSE)
        filename <- this.path::sys.path(local = TRUE, for.msg = TRUE)
        isFile <- !is.na(filename)
        if (isFile) {
            timestamp <- file.info(filename)$mtime[1]
            if (is.na(timestamp))
                timestamp <- Sys.time()
        }
        else {
            filename <- if (is.character(ofile)) ofile else "<connection>"
            timestamp <- Sys.time()
        }
        srcfile <- srcfilecopy(filename, lines, timestamp, isFile)
        this.path::set.src.path(srcfile)
        envir <- new.env(hash = TRUE, parent = .BaseNamespaceEnv)
        envir$.packageName <- filename
        oopt <- options(topLevelEnvironment = envir)
        on.exit(options(oopt))
        this.path::set.env.path(envir)
        exprs <- parse(text = lines, srcfile = srcfile)
        this.path::source.exprs(exprs, evaluated = TRUE, envir)
        `.this.path::document.context`
    }


    this.path::source.exprs({


x <- sourcelike(this.path::relpath(FILE.R))
print(x, quote = FALSE)


y <- local({
    FILE.zip <- tempfile(fileext = ".zip")
    on.exit(unlink(FILE.zip))
    local({
        owd <- getwd()
        on.exit(setwd(owd))
        setwd(this.path::dirname2(FILE.R))
        status <- utils::zip(FILE.zip, this.path::basename2(FILE.R))
        on.exit()
        setwd(owd)
        if (status == 0L) {
            local({
                conn <- unz(this.path::relpath(FILE.zip), this.path::basename2(FILE.R))
                on.exit(close(conn))
                sourcelike(conn)
            })
        }
    })
})
print(y, quote = FALSE)


    })


})

##
## this.path : Get Executing Script's Path
## Copyright (C) 2023-2026   Iris Simmons
##


path.functions <- .removeSource_from_inner_functions(
                  function (file, delayed = FALSE, local = FALSE, n = 0L, envir = parent.frame(n + 1L),
    matchThisEnv = getOption("topLevelEnvironment"),
    srcfile = if (n) sys.parent(n) else 0L)
{
    if (delayed) {
        ofile <- file
        set.sys.path(ofile, path.only = TRUE, allow.url = TRUE,
            allow.file.uri = TRUE, delayed = TRUE, assign_to = "file")
        ## remove the variable(s) created by set.sys.path()
        unset.sys.path()
        delayedAssign("directory", .dir(file))
        `attr<-`(environment(), "name", paste0("make_fix_funs:", ofile))
        this.path <- function(original = FALSE, contents = FALSE) {
            ## force the promise
            original
            if (contents) {
                if (.isfalse(original))
                    readLines(file)
                else stop(gettextf("'%s' must be FALSE when '%s' is TRUE", "original", "contents"))
            }
            else {
                if (.istrue(original)) {
                    ofile
                }
                else if (.External2(.C_promise_is_unevaluated, "file")) {
                    if (.isfalse(original))
                        .External2(.C_forcePromise_no_warn, "file")
                    else ofile
                }
                else file
            }
        }
    }
    else {
        if (missing(file)) {
            n <- .External2(.C_asIntegerGE0, n)
            ofile <- .External2(.C_this_path, FALSE, TRUE, FALSE, FALSE, local, envir, matchThisEnv, srcfile)
            file <- .External2(.C_this_path, local, envir, matchThisEnv, srcfile)
        }
        else {
            ofile <- file
            file <- set.sys.path(file, path.only = TRUE, allow.url = TRUE,
                allow.file.uri = TRUE)
            ## remove the variable(s) created by set.sys.path()
            unset.sys.path()
        }
        directory <- .dir(file)
        `attr<-`(environment(), "name", paste0("make_fix_funs:", file))
        this.path <- function(original = FALSE, contents = FALSE) {
            ## force the promise
            original
            if (contents) {
                if (.isfalse(original))
                    readLines(file)
                else stop(gettextf("'%s' must be FALSE when '%s' is TRUE", "original", "contents"))
            }
            else {
                if (.istrue(original))
                    ofile
                else file
            }
        }
    }
    rm(srcfile, matchThisEnv, envir, n, local, delayed)
    this.dir <- function() directory
    here <- function(..., .. = 0L) {
        base <- .here(file, ..)
        path.join(base, ...)
    }
    delayedAssign("ici", { here })
    this.proj <- function(...) {
        base <- .proj(directory)
        path.join(base, ...)
    }
    rel2here <- function(path) .relpath(path, directory, normalize = FALSE)
    rel2proj <- function(path) .relpath(path, .proj(directory), normalize = FALSE)
    LINENO <- function() .LINENO(file)
    lockEnvironment(environment(), bindings = TRUE)
    environment()
}
)


delayedAssign("make_fix_funs", { path.functions })

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2024   Iris Simmons
%%


\name{OS.type}
\alias{OS.type}
\title{Detect the Operating System Type}
\description{
  \code{OS.type} is a list of \code{TRUE} / / \code{FALSE} values dependent on
  the platform under which this package was built.
}
\usage{
OS.type
}
\value{

  %\item{BSD}{Built under BSD (DragonFly BSD, FreeBSD, OpenBSD, NetBSD).}

  %\item{POSIX}{Built under a POSIX compliant OS.}

  A list with at least the following components:

  \item{AIX}{

    Built under IBM AIX.}

  \item{HPUX}{

    Built under Hewlett-Packard HP-UX.}

  \item{linux}{

    Built under some distribution of Linux.}

  \item{darwin}{

    Built under Apple OSX and iOS (Darwin).}

  \item{iOS.simulator}{

    Built under iOS in Xcode simulator.}

  \item{iOS}{

    Built under iOS on iPhone, iPad, etc.}

  \item{macOS}{

    Built under OSX.}

  \item{solaris}{

    Built under Solaris (SunOS).}

  \item{cygwin}{

    Built under Cygwin POSIX under Microsoft Windows.}

  \item{windows}{

    Built under Microsoft Windows.}

  \item{win64}{

    Built under Microsoft Windows (64-bit).}

  \item{win32}{

    Built under Microsoft Windows (32-bit).}

  \item{UNIX}{

    Built under a UNIX-style OS.}
}
\source{
  \href{http://web.archive.org/web/20191012035921/http://nadeausoftware.com/articles/2012/01/c_c_tip_how_use_compiler_predefined_macros_detect_operating_system}{http://web.archive.org/web/20191012035921/http://nadeausoftware.com/articles/2012/01/c_c_tip_how_use_compiler_predefined_macros_detect_operating_system}
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2024   Iris Simmons
%%


%% used in ./inst/NEWS.Rd
\newcommand{\codeorsQuote}{\ifelse{text}{\sQuote{#1}}{\code{#1}}}
\newcommand{\commandorsQuote}{\ifelse{text}{\sQuote{#1}}{\command{#1}}}
\newcommand{\optionorsQuote}{\ifelse{text}{\sQuote{#1}}{\option{#1}}}
\newcommand{\pkgorsQuote}{\ifelse{text}{\sQuote{#1}}{\pkg{#1}}}


%% does not work well with characters that have
%% special meaning in HTML or LaTeX or Rd
\newcommand{\string}{\ifelse{text}{"#1"}{\code{"#1"}}}
\newcommand{\blankstring}{\ifelse{text}{""}{\code{""}}}
\newcommand{\codelink2}{\ifelse{html}{\code{\link[#1]{#2}}}{\codeorsQuote{#2}}}
\newcommand{\codelink3}{\ifelse{html}{\code{\link[#1]{#2}#3}}{\codeorsQuote{#2#3}}}
\newcommand{\codelink4}{\ifelse{html}{\code{#1\link[#2]{#3}#4}}{\codeorsQuote{#1#3#4}}}


\newcommand{\CRANpackage}{\href{https://CRAN.R-project.org/package=#1}{\pkgorsQuote{package:#1}}}


\newcommand{\Emacs}{\href{https://www.gnu.org/software/emacs/}{Emacs}}
\newcommand{\ESS}{\href{https://ess.r-project.org/}{ESS}}
\newcommand{\Jupyter}{\href{https://jupyter.org/}{Jupyter}}
\newcommand{\Positron}{\href{https://github.com/posit-dev/positron}{Positron}}
\newcommand{\Python}{\href{https://www.python.org/}{Python}}
\newcommand{\radian}{\href{https://github.com/randy3k/radian}{radian}}
\newcommand{\REditorSupport}{\href{https://code.visualstudio.com/docs/languages/r}{REditorSupport}}
\newcommand{\RStudio}{\href{https://posit.co/products/open-source/rstudio/}{RStudio}}
\newcommand{\VSCode}{\href{https://code.visualstudio.com/}{VSCode}}


\newcommand{\sourcelink}{\codelink3{base}{source}{()}}
\newcommand{\syssourcelink}{\codelink3{base}{sys.source}{()}}
\newcommand{\debugSourceinRStudiolink}{\ifelse{text}{\sQuote{debugSource()} in \sQuote{RStudio}}{\href{https://support.posit.co/hc/en-us/articles/205612627-Debugging-with-the-RStudio-IDE}{\code{debugSource}}\code{()} in \sQuote{\RStudio}}}
\newcommand{\compilerloadcmplink}{\codelink3{compiler:compile}{compiler::loadcmp}{()}}
\newcommand{\utilsSweavelink}{\codelink3{utils:Sweave}{utils::Sweave}{()}}
\newcommand{\boxuselink}{\codelink3{box:use}{box::use}{()}}
\newcommand{\knitrknitlink}{\codelink3{knitr:knit}{knitr::knit}{()}}
\newcommand{\plumberplumblink}{\codelink3{plumber:plumb}{plumber::plumb}{()}}
\newcommand{\shinyrunApplink}{\codelink3{shiny:runApp}{shiny::runApp}{()}}
\newcommand{\targetslink}{\CRANpackage{targets}}
\newcommand{\targetstarcallrinnertrylink}{\codelink3{targets:tar_callr_inner_try}{targets::tar_callr_inner_try}{()}}
\newcommand{\targetstarloadglobalslink}{\codelink3{targets:tar_load_globals}{targets::tar_load_globals}{()}}
\newcommand{\targetstarsourcelink}{\codelink3{targets:tar_source}{targets::tar_source}{()}}
\newcommand{\targetstarworkspacelink}{\codelink3{targets:tar_workspace}{targets::tar_workspace}{()}}
\newcommand{\testthatsourcefilelink}{\codelink3{testthat:source_file}{testthat::source_file}{()}}
\newcommand{\Rscriptlink}{\codelink2{utils}{Rscript}}

\newcommand{\sourcenolink}{\codeorsQuote{source()}}
\newcommand{\syssourcenolink}{\codeorsQuote{sys.source()}}
\newcommand{\debugSourceinRStudionolink}{\codeorsQuote{debugSource()} in \sQuote{RStudio}}
\newcommand{\compilerloadcmpnolink}{\codeorsQuote{compiler::loadcmp()}}
\newcommand{\utilsSweavenolink}{\codeorsQuote{utils::Sweave()}}
\newcommand{\boxusenolink}{\codeorsQuote{box::use()}}
\newcommand{\knitrknitnolink}{\codeorsQuote{knitr::knit()}}
\newcommand{\plumberplumbnolink}{\codeorsQuote{plumber::plumb()}}
\newcommand{\shinyrunAppnolink}{\codeorsQuote{shiny::runApp()}}
\newcommand{\targetsnolink}{\pkgorsQuote{package:targets}}
\newcommand{\targetstarcallrinnertrynolink}{\codeorsQuote{targets::tar_callr_inner_try()}}
\newcommand{\targetstarloadglobalsnolink}{\codeorsQuote{targets::tar_load_globals()}}
\newcommand{\targetstarsourcenolink}{\codeorsQuote{targets::tar_source()}}
\newcommand{\targetstarworkspacenolink}{\codeorsQuote{targets::tar_workspace()}}
\newcommand{\testthatsourcefilenolink}{\codeorsQuote{testthat::source_file()}}
\newcommand{\Rscriptnolink}{\codeorsQuote{Rscript}}


\newcommand{\syspathcompatible}{\itemize{\item{\source#1} \item{\syssource#1} \item{\debugSourceinRStudio#1} \item{\compilerloadcmp#1} \item{\utilsSweave#1} \item{\boxuse#1} \item{\knitrknit#1} \item{\plumberplumb#1} \item{\shinyrunApp#1} \item{\targets#1} \item{\testthatsourcefile#1}}}
\newcommand{\syspathcompatiblelinks}{\syspathcompatible{link}}
\newcommand{\syspathcompatiblenolinks}{\syspathcompatible{nolink}}


\newcommand{\sources}{\itemize{\item{\source#1} \item{\syssource#1} \item{\debugSourceinRStudio#1} \item{\compilerloadcmp#1} \item{\utilsSweave#1} \item{\boxuse#1} \item{\knitrknit#1} \item{\plumberplumb#1} \item{\shinyrunApp#1} \item{\targetstarcallrinnertry#1 \cr \targetstarloadglobals#1 \cr \targetstarsource#1 \cr \targetstarworkspace#1} \item{\testthatsourcefile#1}}}
\newcommand{\sourceslinks}{\sources{link}}
\newcommand{\sourcesnolinks}{\sources{nolink}}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2021-2025   Iris Simmons
%%


\name{this.path}
\alias{this.path}
\alias{this.dir}
\title{Determine Script's Filename}
\description{
  \code{this.path()} returns the
  \ifelse{html}{\codelink2{base:normalizePath}{normalized}}{normalized} path of
  the script in which it was written.

  \code{this.dir()} returns the directory of \code{this.path()}.
}
\usage{
this.path(verbose = getOption("verbose"), original = FALSE,
          for.msg = FALSE, contents = FALSE, local = FALSE,
          n = 0, envir = parent.frame(n + 1),
          matchThisEnv = getOption("topLevelEnvironment"),
          srcfile = if (n) sys.parent(n) else 0,
          default, else.)

this.dir(verbose = getOption("verbose"), local = FALSE,
         n = 0, envir = parent.frame(n + 1),
         matchThisEnv = getOption("topLevelEnvironment"),
         srcfile = if (n) sys.parent(n) else 0,
         default, else.)
}
\arguments{
  \item{verbose}{\code{TRUE} or \code{FALSE}; should the method in which the
    path was determined be printed?}

  \item{original}{\code{TRUE}, \code{FALSE}, or \code{NA}; should the original
    or the normalized path be returned? \code{NA} means the normalized path
    will be returned if it has already been forced, and the original path
    otherwise.}

  \item{for.msg}{\code{TRUE} or \code{FALSE}; do you want the path for the
    purpose of printing a diagnostic message / / warning / / error? This will
    return \codelink2{base:NA}{NA_character_} in most cases where an error
    would have been thrown.

    \code{for.msg = TRUE} will ignore \code{original = FALSE}, and will use
    \code{original = NA} instead.}

  \item{contents}{\code{TRUE} or \code{FALSE}; should the contents of the
    script be returned instead?

    In \sQuote{Jupyter}, a list of character vectors will be returned, the
    contents separated into cells. Otherwise, a character vector will be
    returned. If the executing script cannot be determined and \code{for.msg}
    is \code{TRUE}, \code{NULL} will be returned.

    You could use
    \ifelse{html}{\code{\link[base:character]{as.character}(\link[base]{unlist}(this.path(contents = TRUE)))}}{\code{as.character(unlist(this.path(contents = TRUE)))}}
    if you require a character vector.

    This is intended for logging purposes. This is useful in \sQuote{Rgui},
    \sQuote{RStudio}, \sQuote{VSCode}, and \sQuote{Emacs} when the source
    document has contents but no path.}

  \item{local}{\code{TRUE} or \code{FALSE}; should the search for the executing
    script be confined to the local environment in which
    \code{\link{set.sys.path}()} was called?}

  \item{n}{the number of additional generations to go back. By default,
    \code{this.path()} will look for a path based on the
    \ifelse{html}{\codelink2{base:srcfile}{srcref}}{\code{srcref}} of the call
    to \code{this.path()} and the environment in which \code{this.path()} was
    called. This can be changed to be based on the \code{srcref} of the call
    and the calling environment \code{n} generations up the stack. See section
    \strong{Argument 'n'} for more details.}

  \item{envir, matchThisEnv}{arguments passed to
    \codelink3{base:ns-topenv}{topenv}{()} to determine the top level
    environment in which to search for an associated path.}

  \item{srcfile}{source file in which to search for a pathname, or an object
    containing a source file. This includes a source reference, a call, an
    expression object, or a closure.}

  \item{default}{this value is returned if there is no executing script.}

  \item{else.}{function to apply if there is an executing script. See
    \code{\link{tryCatch2}()} for inspiration.}
}
\details{
  \code{this.path()} starts by examining argument \code{srcfile}. It looks at
  the bindings \code{filename} and \code{wd} to determine the associated file
  path. A source file of class \string{srcfilecopy} in which binding
  \code{isFile} is \code{FALSE} will be ignored. A source file of class
  \string{srcfilealias} will use the aliased \code{filename} in determining the
  associated path. Filenames such as \blankstring, \string{clipboard}, and
  \string{stdin} will be ignored since they do not refer to files.

  If it does not find a path associated with \code{srcfile}, it will next
  examine arguments \code{envir} and \code{matchThisEnv}. Specifically, it
  calculates \codelink3{base:ns-topenv}{topenv}{(envir, matchThisEnv)} then
  looks for an associated path. It will find a path associated with the top
  level environment in two ways:

  \itemize{
    \item{from a \CRANpackage{box} module's namespace}

    \item{from an attribute \string{path}}
  }

  If it does not find an associated path with \code{envir} and
  \code{matchThisEnv}, it will next examine the call stack looking for a source
  call; a call to one of these functions: \sourceslinks If a source call is
  found, the file argument is returned from the function's evaluation
  environment. If you have your own \sourcenolink-like function that you would
  like to be recognized by \code{this.path()}, please use
  \code{\link{set.sys.path}()} or contact the package maintainer so that it can
  be implemented.

  If no source call is found up the calling stack, it will next examine the GUI
  in use. If \R is running from:

  \describe{
    \item{a shell, such as the Windows command-line / / Unix terminal}{

      then the shell arguments are searched for \option{-f} \option{FILE} or
      \option{--file=FILE} (the two methods of taking input from \option{FILE})
      (\option{-f} \option{-} and \option{--file=-} are ignored). The last
      \option{FILE} is extracted and returned. If no arguments of either type
      are supplied, an error is thrown.

      If \R is running from a shell under a Unix-alike with \option{-g}
      \option{Tk} or \option{--gui=Tk}, an error is thrown. \sQuote{Tk} does
      not make use of its \option{-f} \option{FILE}, \option{--file=FILE}
      arguments.}

    \item{\sQuote{Rgui}}{

      then the source document's filename (the document most recently
      interacted with) is returned (at the time of evaluation). Please note
      that minimized documents \emph{WILL} be included when looking for the
      most recently used document. It is important to not leave the current
      document (either by closing the document or interacting with another
      document) while any calls to \code{this.path()} have yet to be evaluated
      in the run selection. If no documents are open or the source document
      does not exist (not saved anywhere), an error is thrown.}

    \item{\sQuote{\RStudio}}{

      then the active document's filename (the document in which the cursor is
      active) is returned (at the time of evaluation). If the active document
      is the \R console, the source document's filename (the document open in
      the current tab) is returned (at the time of evaluation). Please note
      that the source document will \emph{NEVER} be a document open in another
      window (with the \strong{Show in new window} button). Please also note
      that an active document open in another window can sometimes lose focus
      and become inactive, thus returning the incorrect path. It is best to not
      run \R code from a document open in another window. It is important to
      not leave the current tab (either by closing or switching tabs) while any
      calls to \code{this.path()} have yet to be evaluated in the run
      selection. If no documents are open or the source document does not exist
      (not saved anywhere), an error is thrown.}

    \item{\sQuote{\Positron}}{

      then the source document's filename is returned (at the time of
      evaluation). It is important to not leave the current tab (either by
      closing or switching tabs) while any calls to \code{this.path()} have yet
      to be evaluated in the run selection. If no documents are open or the
      source document does not exist (not saved anywhere), an error is thrown.}

    \item{\sQuote{\VSCode} + \sQuote{\REditorSupport}}{

      then the source document's filename is returned (at the time of
      evaluation). It is important to not leave the current tab (either by
      closing or switching tabs) while any calls to \code{this.path()} have yet
      to be evaluated in the run selection. If no documents are open or the
      source document does not exist (not saved anywhere), an error is thrown.}

    \item{\sQuote{\Jupyter}}{

      then the source document's filename is guessed by looking for \R
      notebooks in the initial working directory, then searching the contents
      of those files for an expression matching the top-level expression.
      Please be sure to save your notebook before using \code{this.path()}, or
      explicitly use \code{\link{set.jupyter.path}()}.}

    \item{\sQuote{\Emacs} + \sQuote{\ESS}}{

      then the source document's filename is returned (at the time of
      evaluation). \sQuote{Emacs} must be running as a server, either by
      running \code{(server-start)} (consider adding to your \file{~/.emacs}
      file) or typing \code{M-x server-start}. It is important to not leave the
      current window (either by closing or switching buffers) while any calls
      to \code{this.path()} have yet to be evaluated in the run selection. If
      multiple frames are active, \code{this.path()} will pick the first frame
      containing the corresponding \R session.

      If multiple \sQuote{Emacs} sessions are active, \code{this.path()} will
      only work in the primary session due to limitations in
#ifdef windows
      \sQuote{emacsclient.exe}.
#endif
#ifdef unix
      \sQuote{emacsclient}.
#endif
      If you want to run multiple \R sessions, it is better to run one
      \sQuote{Emacs} session with multiple frames, one \R session per frame.
      Use \code{M-x make-frame} to make a new frame, or \code{C-x 5 f} to visit
      a file in a new frame.

      Additionally, never use \code{C-c C-b} to send the current buffer to the
      \R process. This copies the buffer contents to a new file which is then
      \sourcenolink-ed. The source references now point to the wrong file.
      Instead, use \code{C-x h} to select the entire buffer then \code{C-c C-r}
      to evaluate the selection.}

    \item{\sQuote{AQUA}}{

      then the executing script's path cannot be determined. Until such a time
      that there is a method for requesting the path of an open document,
      consider using \sQuote{RStudio}, \sQuote{Positron}, \sQuote{VSCode}, or
      \sQuote{Emacs}.}
  }

  If \R is running in another manner, an error is thrown.

  If your GUI of choice is not implemented with \code{this.path()}, please
  contact the package maintainer so that it can be implemented.
}
\value{
  \code{default} if there is no executing script.

  If \code{contents} is \code{TRUE}, there are a variety of return values. If a
  custom GUI is implemented with \code{\link{set.gui.path}()}, any \R object.
  If the executing script cannot be determined and \code{for.msg} is
  \code{TRUE}, then \code{NULL}. In \Jupyter, a list of character vectors, the
  contents separated into cells. Otherwise, a character vector.

  Otherwise, a character string.
}
\section{Argument 'n'}{
  By default, \code{this.path()} will look for a path based on the
  \code{srcref} of the call to \code{this.path()} and the environment in which
  \code{this.path()} was called. For example: \preformatted{{
#line 1 "file1.R"
fun <- function() this.path::this.path(original = TRUE)
fun()
}


{
#line 1 "file2.R"
fun()
}}

  Both of these will return \string{file1.R} because that is where the call to
  \code{this.path()} is written.

  But suppose we do not care to know where \code{this.path()} is called, but
  instead want to know where \code{fun()} is called. Pass argument \code{n = 1};
  \code{this.path()} will inspect the call and the calling environment one
  generation up the stack: \preformatted{{
#line 1 "file1.R"
fun <- function() this.path::this.path(original = TRUE, n = 1)
fun()
}


{
#line 1 "file2.R"
fun()
}}

  These will return \string{file1.R} and \string{file2.R}, respectively,
  because those are where the calls to \code{fun()} are written.

  But now suppose we wish to make a second function that uses \code{fun()}. We
  do not care to know where \code{fun()} is called, but instead want to know
  where \code{fun2()} is called. Add a formal argument \code{n = 0} to each
  function and pass \code{n = n + 1} to each sub-function: \preformatted{{
#line 1 "file1.R"
fun <- function(n = 0) {
    this.path::this.path(original = TRUE, n = n + 1)
}
fun()
}


{
#line 1 "file2.R"
fun2 <- function(n = 0) fun(n = n + 1)
list(fun = fun(), fun2 = fun2())
}


{
#line 1 "file3.R"
fun3 <- function(n = 0) fun2(n = n + 1)
list(fun = fun(), fun2 = fun2(), fun3 = fun3())
}}

  Within each file, all these functions will return the path in which they are
  called, regardless of how deep \code{this.path()} is called.
}
\note{
  If you need to use \code{this.path()} inside a user profile, please use
  \code{\link{with_init.file}()}. i.e. instead of writing:
  \preformatted{<expr 1>
<expr 2>
<...>}
  write this: \preformatted{this.path::with_init.file({
<expr 1>
<expr 2>
<...>
})}
}
\seealso{
  \code{\link{shFILE}()}

  \code{\link{set.sys.path}()}
}
\examples{
FILE1.R <- tempfile(fileext = ".R")
writeLines("writeLines(sQuote(this.path::this.path()))", FILE1.R)

## 'this.path()' works with 'source()'
source(FILE1.R)

## 'this.path()' works with 'sys.source()'
sys.source(FILE1.R, envir = environment())

## 'this.path()' works with 'debugSource()' in 'RStudio'
if (.Platform$GUI == "RStudio")
    get("debugSource", "tools:rstudio", inherits = FALSE)(FILE1.R)

## 'this.path()' works with 'testthat::source_file()'
if (requireNamespace("testthat"))
    testthat::source_file(FILE1.R, chdir = FALSE, wrap = FALSE)

## 'this.path()' works with 'compiler::loadcmp()'
if (requireNamespace("compiler")) {
    FILE2.Rc <- tempfile(fileext = ".Rc")
    compiler::cmpfile(FILE1.R, FILE2.Rc)
    compiler::loadcmp(FILE2.Rc)
    unlink(FILE2.Rc)
}

## 'this.path()' works with 'Rscript'
this.path:::.Rscript(c(
    "--default-packages=NULL", "--no-save", FILE1.R
))

## 'this.path()' also works when 'source()'-ing a URL
## (included tryCatch in case an internet connection is not available)
tryCatch({
    source(paste0("https://raw.githubusercontent.com/ArcadeAntics/",
                  "this.path/main/tests/sys-path-with-urls.R"))
}, condition = this.path:::.cat_condition)

unlink(FILE1.R)
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2025   Iris Simmons
%%


\name{shFILE}
\alias{shFILE}
\title{Get 'FILE' Provided to R by a Shell}
\description{
  Look through the command line arguments, extracting \option{FILE} from either
  of the following: \option{-f} \option{FILE} or \option{--file=FILE}
}
\usage{
shFILE(original = FALSE, for.msg = FALSE, default, else.)
}
\arguments{
  \item{original}{\code{TRUE}, \code{FALSE}, or \code{NA}; should the original
    or the normalized path be returned? \code{NA} means the normalized path
    will be returned if it has already been forced, and the original path
    otherwise.}

  \item{for.msg}{\code{TRUE} or \code{FALSE}; do you want the path for the
    purpose of printing a diagnostic message / / warning / / error?
    \code{for.msg = TRUE} will ignore \code{original = FALSE}, and will use
    \code{original = NA} instead.}

  \item{default}{if \option{FILE} is not found, this value is returned.}

  \item{else.}{missing or a function to apply if \option{FILE} is found. See
    \ifelse{html}{\code{\link{tryCatch2}()}}{\code{tryCatch2()}} for
    inspiration.}
}
\value{
  character string, or \code{default} if \option{FILE} was not found.
}
\note{
  The original and the normalized path are saved; this makes them faster when
  called subsequent times.

  On Windows, the normalized path will use \ifelse{text}{/}{\code{/}} as the
  file separator.
}
\seealso{
  \code{\link{this.path}()}, \code{\link{here}()}
}
\examples{
FILE.R <- tempfile(fileext = ".R")
this.path:::.writeCode({
    this.path:::.withAutoprint({
        this.path::shFILE(original = TRUE)
        this.path::shFILE()
        this.path::shFILE(default = {
            stop("since 'FILE.R' will be found,\\n",
                 "this error will not be thrown")
        })
    }, spaced = TRUE, verbose = FALSE, width.cutoff = 60L)
}, FILE.R)
this.path:::.Rscript(
    c("--default-packages=NULL", "--no-save", FILE.R)
)
unlink(FILE.R)


for (expr in c("shFILE(original = TRUE)",
               "shFILE(original = TRUE, default = NULL)",
               "shFILE()",
               "shFILE(default = NULL)"))
{
    cat("\\n\\n")
    this.path:::.Rscript(
        c("--default-packages=NULL", "--no-save", "-e", expr)
    )
}
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2024   Iris Simmons
%%


\name{LINENO}
\alias{LINENO}
\title{Line Number of Executing Expression}
\description{
  Get the line number of the executing expression.
}
\usage{
LINENO(n = 0, envir = parent.frame(n + 1),
       matchThisEnv = getOption("topLevelEnvironment"),
       srcfile = if (n) sys.parent(n) else 0)
}
\arguments{
  \item{n, envir, matchThisEnv, srcfile}{See \code{?\link{this.path}()}.}
}
\details{
  \code{LINENO()} only works if the expressions have a
  \codelink2{base:srcfile}{srcref}.

  Scripts run with \Rscriptlink do not store their \code{srcref}, even when
  \codelink3{base:options}{getOption}{("keep.source")} is \code{TRUE}.

  For \sourcelink and \syssourcelink, make sure to supply argument
  \code{keep.source = TRUE} directly, or set options \string{keep.source} and
  \string{keep.source.pkgs} to \code{TRUE}.

  For \debugSourceinRStudiolink, it has no argument \code{keep.source}, so set
  option \string{keep.source} to \code{TRUE} before calling.

  For \compilerloadcmplink, the \code{srcref} is never stored for the compiled
  code, there is nothing that can be done.

  For \utilsSweavelink, the \code{srcref} is never stored, there is nothing
  that can be done.

  For \knitrknitlink, the \code{srcref} is never stored, there is nothing that
  can be done. I am looking into a fix.

  For \targetslink, set option \string{keep.source} to \code{TRUE} before
  calling associated functions.

  For \boxuselink, \plumberplumblink, \shinyrunApplink, and
  \testthatsourcefilelink, the \code{srcref} is always stored.
}
\value{
  integer; \code{NA_integer_} if the line number is not found.
}
\note{
  You can get a more accurate line number by wrapping \code{LINENO()} in
  braces: \preformatted{{ LINENO() }}
}
\examples{
FILE.R <- tempfile(fileext = ".R")
writeLines(c("
LINENO()
LINENO()
## LINENO() respects #line directives
#line 15
LINENO()
#line 1218
cat(sprintf('invalid value \%d at \%s, line \%d\\\\n',
            -5, try.this.path(), LINENO()))
"), FILE.R)


if (getRversion() >= "4.3.0") {
    source(FILE.R, echo = TRUE, verbose = FALSE,
        max.deparse.length = Inf, keep.source = TRUE)
} else {
    this.path:::.source(FILE.R, echo = TRUE, verbose = FALSE,
        max.deparse.length = Inf, keep.source = TRUE)
}


unlink(FILE.R)
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2024   Iris Simmons
%%


\name{R.from.shell}
\alias{R.from.shell}
\title{Using R From a Shell}
\description{
  How to use \R from a shell (including the Windows command-line / / Unix
  terminal).
}
\details{
  For the purpose of running \R scripts, there are four ways to do it. Suppose
  our \R script has filename \option{script1.R}, we could write any of:

  \itemize{
    \item{\verb{R -f script1.R}}

    \item{\verb{R --file=script1.R}}

    \item{\verb{R CMD BATCH script1.R}}

    \item{\verb{Rscript script1.R}}
  }

  The first two are different ways of writing equivalent statements. The third
  statement is the first statement plus options \option{--restore}
  \option{--save} (plus option \option{--no-readline} under Unix-alikes), and
  it also saves the \codelink2{base:showConnections}{stdout} and
  \codelink2{base:showConnections}{stderr} in a file of your choosing. The
  fourth statement is the second statement plus options \option{--no-echo}
  \option{--no-restore}. You can try:

  \itemize{
    \item{\verb{R --help}}

    \item{\verb{R CMD BATCH --help}}

    \item{\verb{Rscript --help}}
  }

  for a help message that describes what these options mean. In general,
  \command{Rscript} is the one you want to use. It should be noted that
  \command{Rscript} has some exclusive
  \codelink2{base:EnvVar}{environment variables} (not used by the other
  executables) that will make its behaviour different from \command{R}.

  For the purpose of making packages, \command{R CMD} is what you will need.
  Most commonly, you will use:

  \itemize{
    \item{\verb{R CMD build}}

    \item{\verb{R CMD INSTALL}}

    \item{\verb{R CMD check}}
  }

  \command{R CMD build} will turn an \R package (specified by a directory) into
  tarball. This allows for easy sharing of \R packages with other people,
  including
  \href{https://cran.r-project.org/submit.html}{submitting a package to CRAN}.
  \command{R CMD INSTALL} will install an \R package (specified by a directory
  or tarball), and is used by
  \codelink3{utils:install.packages}{install.packages}{()}.
  \command{R CMD check} will check an \R package (specified by a tarball) for
  possible errors in code, documentation, tests, and much more.

#ifdef windows
  If, when you execute one of the previous commands, you see the following
  error message: \dQuote{\sQuote{R} is not recognized as an internal or
  external command, operable program or batch file.}, see section
  \strong{Ease of Use on Windows}.
#endif
}
#ifdef windows
\section{Ease of Use on Windows}{
  Under Unix-alikes, it is easy to invoke an \R session from a shell by typing
  the name of the \R executable you wish to run. On Windows, you should see
  that typing the name of the \R executable you wish to run does not run that
  application, but instead signals an error. Instead, you will have to type the
  full path of the directory where your \R executables are located (see section
  \strong{Where are my R executable files located?}), followed by the name of
  the \R executable you wish to run.

  This is not very convenient to type everytime something needs to be run from
  a shell, plus it has another issue of being computer dependent. The solution
  is to add the path of the directory where your \R executables are located to
  the \env{Path} environment variable. The \env{Path} environment variable is a
  list of directories where executable programs are located. When you type the
  name of an executable program you wish to run, Windows looks for that program
  through each directory in the \env{Path} environment variable. When you add
  the full path of the directory where your \R executables are located to your
  \env{Path} environment variable, you should be able to run any of those
  executable programs by their basenames (\sQuote{R}, \sQuote{Rcmd},
  \sQuote{Rscript}, and \sQuote{Rterm}) instead of their full paths.

  To add a new path to your \env{Path} environment variable:

  \enumerate{
    \item{Open the Windows search bar (press \strong{Windows} key)}

    \item{Type \strong{Edit environment variables for your account}}

    \item{Click the variable \env{Path}}

    \item{Click the button \strong{Edit...}}

    \item{Click the button \strong{New}}

    \item{Type (or paste) the full path of the directory where your \R
      executables are located and click the \strong{OK} button}
  }

  This will modify your environment variable \env{Path}, not the systems. If
  another user wishes to run \R from a shell, they will have to add the
  directory to their \env{Path} environment variable as well.

  If you wish to modify the system environment variable \env{Path} (you will
  need admin permissions):

  \enumerate{
    \item{Open the Windows search bar (press \strong{Windows} key)}

    \item{Type \strong{Edit the system environment variables}}

    \item{Click the button \strong{Environment Variables...}}

    \item{Click the variable \env{Path} in section \strong{System variables}}

    \item{Click the button \strong{Edit...}}

    \item{Click the button \strong{New}}

    \item{Type (or paste) the full path of the directory where your \R
      executables are located and click the \strong{OK} button}
  }

  To check that this worked correctly, open a shell and execute the following
  commands:

  \itemize{
    \item{\verb{R --help}}

    \item{\verb{R --version}}
  }

  You should see that the first prints the usage message for the \R executable
  while the second prints information about the version of \R currently being
  run. If you have multiple versions of \R installed, make sure this is the
  version of \R you wish to run.
}
#endif
\section{Where are my R executable files located?}{
  In an \R session, you can find the location of your \R executable files with
  the following command:

  \code{R.home("bin")}

  For me, this is:

#ifdef unix
  \file{\verb{/usr/lib/R/bin}}
#endif
#ifdef windows
  \file{\verb{C:/PROGRA~1/R/R-}\Sexpr[stage=render]{getRversion()}\verb{/bin/x64}}
#endif
}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2026   Iris Simmons
%%


\name{path.split}
\alias{path.split}
\alias{path.split.1}
\alias{path.unsplit}
\title{Split File Path Into Individual Components}
\description{
  Split the path to a file into components in a platform-\strong{DEPENDENT}
  way.
}
\usage{
path.split(path)
path.split.1(path)
path.unsplit(...)
}
\arguments{
  \item{path}{character vector.}

  \item{\dots}{character vectors, or one list of character vectors.}
}
\value{
  for \code{path.split()}, a list of character vectors.

  for \code{path.split.1()} and \code{path.unsplit()}, a character vector.
}
\note{
  \code{path.unsplit()} is \strong{NOT} the same as \code{\link{path.join}()}.
}
\examples{
#ifdef windows
path <- c(
    NA,
    "",
    paste0("https://raw.githubusercontent.com/ArcadeAntics/PACKAGES/",
           "src/contrib/Archive/this.path/this.path_1.0.0.tar.gz"),
    "\\\\\\\\host\\\\share\\\\path\\\\to\\\\file",
    "\\\\\\\\host\\\\share\\\\",
    "\\\\\\\\host\\\\share",
    "C:\\\\path\\\\to\\\\file",
    "C:path\\\\to\\\\file",
    "path\\\\to\\\\file",
    "\\\\path\\\\to\\\\file",
    "~\\\\path\\\\to\\\\file",
    ## paths with character encodings
    `Encoding<-`("path/to/fil\\xe9", "latin1"),
    "C:/Users/iris/Documents/\\u03b4.R"
)
print(x <- path.split(path))
print(path.unsplit(x))
#endif
#ifdef unix
path <- c(
    NA,
    "",
    paste0("https://raw.githubusercontent.com/ArcadeAntics/PACKAGES/",
           "src/contrib/Archive/this.path/this.path_1.0.0.tar.gz"),
    "//host/share/path/to/file",
    "//host/share/",
    "//host//share",
    "/path/to/file",
    "~/path/to/file",
    "path/to/file",
    ## paths with character encodings
    `Encoding<-`("path/to/fil\\xe9", "latin1"),
    "/home/iris/Documents/\\u03b4.R"
)
print(x <- path.split(path))
print(path.unsplit(x))
#endif
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2024   Iris Simmons
%%


\name{try.this.path}
\alias{try.this.path}
\alias{try.shFILE}
\title{Attempt to Determine Script's Filename}
\description{
  \code{try.this.path()} attempts to return \code{\link{this.path}()},
  returning \code{this.path(original = TRUE)} if that fails, returning
  \codelink2{base:NA}{NA_character_} if that fails as well.
}
\usage{
try.this.path(contents = FALSE, local = FALSE, n = 0,
              envir = parent.frame(n + 1),
              matchThisEnv = getOption("topLevelEnvironment"),
              srcfile = if (n) sys.parent(n) else 0)

try.shFILE()
}
\arguments{
  \item{contents, local, n, envir, matchThisEnv, srcfile}{See
    \code{?\link{this.path}()}.}
}
\details{
  This should \strong{NOT} be used to construct file paths against the script's
  directory. This should exclusively be used for diagnostic messages / /
  warnings / / errors / / logging. The returned path may not exist, may be
  relative instead of absolute, or may be undefined.
}
\value{
  character string.
}
\examples{
try.shFILE()
try.this.path()
try.this.path(contents = TRUE)
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2024   Iris Simmons
%%


\name{Sys.putenv}
\alias{Sys.putenv}
\title{Set Environment Variables}
\description{
  \code{Sys.putenv()} sets environment variables (for other processes called
  from within \R or future calls to \codelink3{base}{Sys.getenv}{()} from this
  \R process).
}
\usage{
Sys.putenv(x)
}
\arguments{
  \item{x}{a character vector, or an object coercible to character. Strings
    must be of the form
    \ifelse{text}{"name=value"}{\ifelse{latex}{\code{"name=value"}}{\code{"\var{name}=\var{value}"}}}.}
}
\value{
  A logical vector, with elements being true if setting the corresponding
  variable succeeded.
}
\seealso{
  \codelink3{base}{Sys.setenv}{()}
}
\examples{
Sys.putenv(c("R_TEST=testit", "A+C=123"))
Sys.getenv("R_TEST")
Sys.unsetenv("R_TEST")  ## under Unix-alikes may warn and not succeed
Sys.getenv("R_TEST", unset = NA)
}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2024   Iris Simmons
%%


\name{tryCatch2}
\alias{tryCatch2}
\alias{tryCatch3}
\alias{last.condition}
\title{Condition Handling and Recovery}
\description{
  Variants of \codelink3{base:conditions}{tryCatch}{()} that accept an
  \code{else.} argument, similar to \code{try except} in \sQuote{\Python}.

  \code{last.condition} will be the last thrown and caught condition in
  \code{tryCatch3()}.
}
\usage{
tryCatch2(expr, ..., else., finally)
tryCatch3(expr, ..., else., finally)

last.condition
}
\arguments{
  \item{expr}{expression to be evaluated.}

  \item{\dots}{for \code{tryCatch2()}, condition handlers. for
    \code{tryCatch3()}, expressions to be conditionally evaluated.

    Arguments which are missing will use the next non-missing argument. If
    there is no next non-missing argument, \code{NULL} will be returned
    invisibly.}

  \item{else.}{expression to be evaluated if evaluating \code{expr} does not
    throw an error nor a condition is caught.}

  \item{finally}{expression to be evaluated before returning or exiting.}
}
\details{
  The use of the \code{else.} argument is better than adding additional code to
  \code{expr} because it avoids accidentally catching a condition that was not
  being protected by the \code{tryCatch()} call.
}
\examples{
FILES <- tempfile(c("existent-file_", "non-existent-file_"))
writeLines("line1\\nline2", FILES[[1L]])
for (FILE in FILES) {
    conn <- file(FILE)
    tryCatch2({
        open(conn, "r")
    }, condition = function(cond) {
        cat("cannot open", FILE, "\\n")
    }, else. = {
        cat(FILE, "has", length(readLines(conn)), "lines\\n")
    }, finally = {
        close(conn)
    })
    # ## or more naturely with tryCatch3:
    # tryCatch3({
    #     open(conn, "r")
    # }, condition = {
    #     cat("cannot open", FILE, "\\n")
    # }, else. = {
    #     cat(FILE, "has", length(readLines(conn)), "lines\\n")
    # }, finally = {
    #     close(conn)
    # })
}
unlink(FILES)
}
\keyword{error}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2021-2025   Iris Simmons
%%


\name{progArgs}
\alias{asArgs}
\alias{progArgs}
\alias{withArgs}
\title{Providing Arguments to a Script}
\description{
  \code{withArgs()} allows you to \sourcelink an \R script while providing
  arguments. As opposed to running with \Rscriptlink, the code will be
  evaluated in the same session in an environment of your choosing.

  \code{progArgs()} is a generalized version of
  \codelink3{base}{commandArgs}{(trailingOnly = TRUE)}, allowing you to access
  the program's arguments whether it was \sourcenolink-ed or run from a shell.

  \code{asArgs()} coerces \R objects into a character vector, for use with
  command line applications and \code{withArgs()}.
}
\usage{
asArgs(...)
progArgs(ifnotfound = character())
withArgs(...)
}
\arguments{
  \item{\dots}{\R objects to turn into script arguments; typically
    \codelink2{base}{logical}, \codelink2{base}{numeric},
    \codelink2{base}{character}, \codelink2{base:Dates}{Date}, and
    \codelink2{base:DateTimeClasses}{POSIXt} vectors.

    for \code{withArgs()}, the first argument should be an (unevaluated) call
    to a \sourcenolink-like function.}

  \item{ifnotfound}{any \R object, e.g., \code{character(0)}, \code{NULL}, a
    call to \codelink3{base}{stop}{()}}
}
\details{
  \code{progArgs()} will return the arguments associated with the executing
  script, or \code{commandArgs(trailingOnly = TRUE)} when there is no executing
  script.

  \code{asArgs()} coerces objects into command-line arguments. \code{\dots} is
  first put into a list, and then each non-list element is converted to
  character. They are converted as follows:

  \describe{
    \item{Factors (class \string{factor})}{

      using \codelink3{base:character}{as.character.factor}{()}}

    \item{Date-Times (class \string{POSIXct} and \string{POSIXlt})}{

      using format
      \ifelse{text}{"\%Y-\%m-\%d \%H:\%M:\%OS6"}{\code{"\%Y-\%m-\%d \%H:\%M:\%OS6"}}
      (retains as much precision as possible)}

    \item{Numbers (class \string{numeric} and \string{complex})}{

      with 17 significant digits (retains as much precision as possible) and
      \string{.} as the decimal point character.}

    \item{Raw Bytes (class \string{raw})}{

      using
      \ifelse{html}{\code{\link[base]{sprintf}("0x\%02x", )}}{\code{sprintf("0x\%02x", )}}
      (for easy conversion back to raw with \codelink3{base:raw}{as.raw}{()} or
      \codelink3{base:vector}{as.vector}{(, "raw")})}
  }

  All others will be converted to character using
  \codelink3{base:character}{as.character}{()} and its methods.

  The arguments will then be unlisted, and all attributes will be removed.
  Arguments that are \code{NA_character_} after conversion will be converted to
  \string{NA} (since the command-line arguments also never have missing
  strings).
}
\value{
  for \code{asArgs()} and \code{progArgs()}, a character vector.

  for \code{withArgs()}, the result of evaluating the first argument.
}
\examples{
this.path::asArgs(NULL, c(TRUE, FALSE, NA), 1:5, pi, exp(6i),
    letters[1:5], as.raw(0:4), Sys.Date(), Sys.time(),
    list(list(list("lists are recursed"))))


FILE.R <- tempfile(fileext = ".R")
this.path:::.writeCode({
    this.path:::.withAutoprint({
        this.path::sys.path()
        this.path::progArgs()
    }, spaced = TRUE, verbose = FALSE, width.cutoff = 60L)
}, FILE.R)


## wrap your source call with a call to withArgs()
this.path::withArgs(
    source(FILE.R, local = TRUE, verbose = FALSE),
    letters[6:10], pi, exp(1)
)
this.path:::.Rscript(c(
    "--default-packages=NULL", "--no-save", FILE.R,
    this.path::asArgs(letters[16:20], pi, Sys.time())
))
this.path:::.Rscript(c(
    "--default-packages=NULL", "--no-save",
    rbind("-e", readLines(FILE.R)[-2L]),
    this.path::asArgs(letters[16:20], pi, Sys.time())
))


# ## with R >= 4.1.0, use the forward pipe operator '|>' to
# ## make calls to withArgs() more intuitive:
# source(FILE.R, local = TRUE, verbose = FALSE) |> this.path::withArgs(
#     letters[6:10], pi, exp(1))


## withArgs() also works with set.sys.path()
sourcelike <- function (file, envir = parent.frame())
{
    file <- set.sys.path(file)
    envir <- as.environment(envir)
    exprs <- parse(n = -1, file = file)
    for (i in seq_along(exprs)) eval(exprs[i], envir)
}
this.path::withArgs(sourcelike(FILE.R), letters[21:26])


unlink(FILE.R)
}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2024   Iris Simmons
%%


\name{sys.path}
\alias{sys.path}

\alias{src.path}
\alias{src.dir}
\alias{src.here}
\alias{src.proj}
\alias{rel2src.dir}
\alias{rel2src.proj}
\alias{src.LINENO}
\alias{try.src.path}

\alias{env.path}
\alias{env.dir}
\alias{env.here}
\alias{env.proj}
\alias{rel2env.dir}
\alias{rel2env.proj}
\alias{env.LINENO}
\alias{try.env.path}

\alias{sys.path}
\alias{sys.dir}
\alias{sys.here}
\alias{sys.proj}
\alias{rel2sys.dir}
\alias{rel2sys.proj}
\alias{sys.LINENO}
\alias{try.sys.path}

\title{Variants of Core Functions in Package 'this.path'}
\description{
  \code{\link{this.path}()} is composed of three sections:
  \enumerate{
    \item{examining argument \code{srcfile}.}

    \item{examining arguments \code{envir} and \code{matchThisEnv}.}

    \item{examining the call stack and the GUI in use.}
  }
  \code{src.path()}, \code{env.path()}, and \code{sys.path()} can be used to
  perform any of those sections individually.

  The other functions listed here are similar such variants of the core
  functions \code{\link{this.dir}()}, \code{\link{here}()},
  \code{\link{this.proj}()}, \code{\link{rel2here}()},
  \code{\link{rel2proj}()}, \code{\link{LINENO}()}, and
  \code{\link{try.this.path}()}.
}
\usage{
src.path(verbose = getOption("verbose"), original = FALSE,
         for.msg = FALSE, contents = FALSE, n = 0,
         srcfile = if (n) sys.parent(n) else 0,
         default, else.)
src.dir(verbose = getOption("verbose"), n = 0,
        srcfile = if (n) sys.parent(n) else 0,
        default, else.)
src.here(..., n = 0,
         srcfile = if (n) sys.parent(n) else 0, .. = 0)
src.proj(..., n = 0,
         srcfile = if (n) sys.parent(n) else 0)
rel2src.dir(path, n = 0,
            srcfile = if (n) sys.parent(n) else 0)
rel2src.proj(path, n = 0,
             srcfile = if (n) sys.parent(n) else 0)
src.LINENO(n = 0, srcfile = if (n) sys.parent(n) else 0)
try.src.path(contents = FALSE, n = 0,
             srcfile = if (n) sys.parent(n) else 0)

env.path(verbose = getOption("verbose"), original = FALSE,
         for.msg = FALSE, contents = FALSE, n = 0,
         envir = parent.frame(n + 1),
         matchThisEnv = getOption("topLevelEnvironment"),
         default, else.)
env.dir(verbose = getOption("verbose"), n = 0,
        envir = parent.frame(n + 1),
        matchThisEnv = getOption("topLevelEnvironment"),
        default, else.)
env.here(..., n = 0, envir = parent.frame(n + 1),
         matchThisEnv = getOption("topLevelEnvironment"), .. = 0)
env.proj(..., n = 0, envir = parent.frame(n + 1),
         matchThisEnv = getOption("topLevelEnvironment"))
rel2env.dir(path, n = 0, envir = parent.frame(n + 1),
            matchThisEnv = getOption("topLevelEnvironment"))
rel2env.proj(path, n = 0, envir = parent.frame(n + 1),
             matchThisEnv = getOption("topLevelEnvironment"))
env.LINENO(n = 0, envir = parent.frame(n + 1),
           matchThisEnv = getOption("topLevelEnvironment"))
try.env.path(contents = FALSE, n = 0,
             envir = parent.frame(n + 1),
             matchThisEnv = getOption("topLevelEnvironment"))

sys.path(verbose = getOption("verbose"), original = FALSE,
         for.msg = FALSE, contents = FALSE, local = FALSE,
         default, else.)
sys.dir(verbose = getOption("verbose"), local = FALSE,
        default, else.)
sys.here(..., local = FALSE, .. = 0)
sys.proj(..., local = FALSE)
rel2sys.dir(path, local = FALSE)
rel2sys.proj(path, local = FALSE)
sys.LINENO()
try.sys.path(contents = FALSE, local = FALSE)
}
\details{
  Before \pkg{this.path 2.0.0}, \code{this.path()} did not have arguments
  \code{srcfile}, \code{envir}, nor \code{matchThisEnv}. It only examined the
  call stack and the GUI in use. In the major update, the original
  implementation of \code{this.path()} was renamed to \code{sys.path()}.
  \code{sys.path()} and related are available for backwards compatibility.
  However, they are unlikely to be useful compared to the advancements made in
  \code{this.path()}.

  \code{env.path()} and \code{src.path()} and related are unlikely to be useful
  when called with no arguments.
}
\examples{
## the important difference between 'this.path()' and 'sys.path()'
FILE1.R <- tempfile("FILE1-", fileext = ".R")
this.path:::.writeCode({
    fun <- function() {
        cat("\\n> this.path()\\n")
        print(this.path::this.path(verbose = TRUE))
        cat("\\n> sys.path()\\n")
        print(this.path::sys.path(verbose = TRUE))
    }
    ## 'this.path()' and 'sys.path()' should be identical because the
    ## executing script is the same as the script of the source file
    fun()
}, FILE1.R)
source(FILE1.R, keep.source = TRUE)

FILE2.R <- tempfile("FILE2-", fileext = ".R")
this.path:::.writeCode({
    ## 'this.path()' and 'sys.path()' should no longer be identical
    ## since FILE2.R is now the executing script, and FILE1.R is not
    fun()
}, FILE2.R)
source(FILE2.R)

unlink(c(FILE1.R, FILE2.R))
}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2025   Iris Simmons
%%


\name{is.main}
\alias{is.main}
\alias{from.shell}
\title{Top-Level Code Environment}
\description{
  Determine if a program is the main program, or if a program was run from a
  shell.
}
\usage{
is.main()
from.shell()
}
\details{
  If no scripts are being \sourcelink-ed, then \code{is.main()} will be
  \code{TRUE}, and \code{FALSE} otherwise.

  If \code{is.main()} is \code{TRUE} and \R was invoked with the command-line
  arguments \option{-f} \option{FILE}, \option{--file=FILE}, or \option{-e}
  \option{EXPR}, then \code{from.shell()} will be \code{TRUE}, and \code{FALSE}
  otherwise.
}
\value{
  \code{TRUE} or \code{FALSE}.
}
\examples{
FILE.R <- tempfile(fileext = ".R")
this.path:::.writeCode({
    this.path:::.withAutoprint({
        this.path::from.shell()
        this.path::is.main()
    }, spaced = TRUE, verbose = FALSE,
       prompt.echo = "FILE.R> ", continue.echo = "FILE.R+ ")
}, FILE.R)

this.path:::.Rscript(
    c("--default-packages=NULL", "--no-save", FILE.R),
    show.command = FALSE,
    show.output.on.console = TRUE
)

source(FILE.R, verbose = FALSE)

this.path:::.Rscript(
    c(
        "--default-packages=NULL", "--no-save",
        "-e", "cat(\\"\\n> this.path::from.shell()\\\\n\\")",
        "-e", "this.path::from.shell()",
        "-e", "cat(\\"\\n> this.path::is.main()\\\\n\\")",
        "-e", "this.path::is.main()",
        "-e", "cat(\\"\\n> source(commandArgs(trailingOnly = TRUE))\\\\n\\")",
        "-e", "source(commandArgs(trailingOnly = TRUE))", FILE.R
    ),
    show.command = FALSE,
    show.output.on.console = TRUE
)

unlink(FILE.R)
}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2025   Iris Simmons
%%


\name{make_fix_funs}
\alias{make_fix_funs}
\alias{path.functions}
\title{Constructs Path Functions Similar to 'this.path()'}
\description{
  \code{make_fix_funs()} accepts a pathname and constructs a set of
  path-related functions, similar to \code{this.path()} and associated.
}
\usage{
make_fix_funs(file, delayed = FALSE, local = FALSE, n = 0,
              envir = parent.frame(n + 1),
              matchThisEnv = getOption("topLevelEnvironment"),
              srcfile = if (n) sys.parent(n) else 0)

## alias for 'make_fix_funs'
path.functions(file, delayed = FALSE, local = FALSE, n = 0,
               envir = parent.frame(n + 1),
               matchThisEnv = getOption("topLevelEnvironment"),
               srcfile = if (n) sys.parent(n) else 0)
}
\arguments{
  \item{file}{a character string giving the pathname of the file or URL.}

  \item{delayed}{\code{TRUE} or \code{FALSE}; should the
    \ifelse{html}{\codelink2{base:normalizePath}{normalizing}}{normalizing} of
    \code{file} be delayed?}

  \item{local, n, envir, matchThisEnv, srcfile}{See \code{?\link{this.path}()}.}
}
\value{
  An environment with at least the following bindings:

  \item{this.path}{

    Function with formals \code{(original = FALSE, contents = FALSE)} which
    returns the
    \ifelse{html}{\codelink2{base:normalizePath}{normalized}}{normalized} file
    path, the original file path, or the contents of the file.}

  \item{this.dir}{

    Function with no formals which returns the directory of the
    \ifelse{html}{\codelink2{base:normalizePath}{normalized}}{normalized} file
    path.}

  \item{here, ici}{

    Function with formals \code{(..., .. = 0)} which constructs file paths,
    starting with the file's directory.}

  \item{this.proj}{

    Function with formals \code{(..., .. = 0)} which constructs file paths,
    starting with the project root.}

  \item{rel2here, rel2proj}{

    Functions with formals \code{(path)} which turn absolute paths into
    relative paths, against the file's directory / / project root.}

  \item{LINENO}{

    Function with no formals which returns the line number of the executing
    expression in \code{file}.}
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2024   Iris Simmons
%%


\name{getinitwd}
\alias{getinitwd}
\alias{initwd}
\title{Get Initial Working Directory}
\description{
  \code{getinitwd()} returns an absolute filepath representing the working
  directory at the time of loading this package.
}
\usage{
getinitwd()
initwd
}
\value{
  character string, or \code{NULL} if the initial working directory is not
  available.
}
\examples{
cat("\\ninitial working directory:\\n"); getinitwd()
cat("\\ncurrent working directory:\\n"); getwd()
}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2024   Iris Simmons
%%


\name{path.join}
\alias{path.join}
\title{Construct Path to File}
\description{
  Construct the path to a file from components / / paths in a
  platform-\strong{DEPENDENT} way.
}
\usage{
path.join(...)
}
\arguments{
  \item{\dots}{character vectors.}
}
\details{
  When constructing a path to a file, the last absolute path is selected and
  all trailing components are appended. This is different from
  \codelink3{base}{file.path}{()} where all trailing paths are treated as
  components.
}
\value{
  A character vector of the arguments concatenated term-by-term and separated
  by \string{/}.
}
\examples{
#ifdef unix
path.join("/", "test1")

path.join("/path/to/file1", "/path/to/file2")

path.join("//host-name/share-name/path/to/file1", "/path/to/file2")

path.join("testing", "/testing", "~", "~/testing", "//host",
    "//host/share", "//host/share/path/to/file", "not-an-abs-path")

path.join("/test1", "test2", "test3")

path.join("test1", "/test2", "test3", "//host/share/test4", "test5",
    "/test6", "test7", "test8", "test9")
#endif
#ifdef windows
path.join("C:", "test1")

path.join("C:/", "test1")

path.join("C:/path/to/file1", "/path/to/file2")

path.join("//host-name/share-name/path/to/file1", "/path/to/file2")

path.join("C:testing", "C:/testing", "~", "~/testing", "//host",
    "//host/share", "//host/share/path/to/file", "not-an-abs-path")

path.join("c:/test1", "c:test2", "C:test3")

path.join("test1", "c:/test2", "test3", "//host/share/test4", "test5",
    "c:/test6", "test7", "c:test8", "test9")
#endif
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2025   Iris Simmons
%%


\name{set.gui.path}
\alias{set.gui.path}
\alias{thisPathNotExistsError}
\alias{thisPathNotFoundError}
\title{Declare GUI's Active Document}
\description{
  \code{set.gui.path()} can be used to implement \code{\link{this.path}()} for
  arbitrary GUIs.
}
\usage{
set.gui.path(...)

thisPathNotExistsError(..., call. = TRUE, domain = NULL,
                       call = .getCurrentCall())

thisPathNotFoundError(..., call. = TRUE, domain = NULL,
                      call = .getCurrentCall())
}
\arguments{
  \item{\dots, call., domain, call}{See details.}
}
\details{
  \code{thisPathNotExistsError()} and \code{thisPathNotFoundError()} are
  provided for use inside \code{set.gui.path()}, and should not be used
  elsewhere.

  If no arguments are passed to \code{set.gui.path()}, the default behaviour
  will be restored.

  If one argument is passed to \code{set.gui.path()}, it must be a function
  that returns the path of the active document in your GUI. It must accept the
  following arguments: \code{(verbose, original, for.msg, contents)} (default
  values are unnecessary). This makes sense for a GUI which can edit and run \R
  code from several different documents such as RGui, \RStudio, \Positron,
  \VSCode + \REditorSupport, and \Emacs + \ESS.

  If two or three arguments are passed to \code{set.gui.path()}, they must be
  the name of the GUI, the path of the active document, and optionally a
  function to get the contents of the document. If provided, the function must
  accept at least one argument which will be the normalized path of the
  document. This makes sense for a GUI which can edit and run \R code from only
  one document such as \Jupyter and shell.

  It is best to call this function as a user hook.
  \preformatted{setHook(packageEvent("this.path"),
function(pkgname, pkgpath)
{
    this.path::set.gui.path(<...>)
}, action = "prepend")}

  An example for a GUI which can run code from multiple documents:
  \preformatted{evalq(envir = new.env(parent = .BaseNamespaceEnv), {
    .guiname <- "myGui"
    .custom_gui_path <- function(verbose, original, for.msg, contents) {
        if (verbose)
            cat("Source: document in", .guiname, "\\n")

        ## your GUI needs to know which document is active
        ## and some way to retrieve that document from R
        doc <- <.myGui_activeDocument()>

        ## if no documents are open, 'doc' should be NULL
        ## or some other object to represent no documents open
        if (is.null(doc)) {
            if (for.msg)
                NA_character_
            else stop(this.path::thisPathNotExistsError(
                "R is running from ", .guiname, " with no documents open\\n",
                " (or document has no path)"))
        }
        else if (contents) {
            ## somehow, get and return the contents of the document
            <doc$contents>
        }
        else {
            ## somehow, get the path of the document
            path <- <doc$path>
            if (nzchar(path)) {
                ## if the path is not normalized, this will normalize it
                if (isFALSE(original))
                    normalizePath(path, "/", TRUE)
                else path
                # ## otherwise, you could just do:
                # path
            }
            else if (for.msg)
                ## return "Untitled" possibly translated
                gettext("Untitled", domain = "RGui", trim = FALSE)
            else
                stop(this.path::thisPathNotFoundError(
                    "document in ", .guiname, " has no associated path (has yet to be saved)"))
        }
    }
    ## recommended to prevent tampering
    lockEnvironment(environment(), bindings = TRUE)
    setHook(packageEvent("this.path"),
    function(pkgname, pkgpath) {
        this.path::set.gui.path(.custom_gui_path)
    }, action = "prepend")
})}

  An example for a GUI which can run code from only one document:
  \preformatted{evalq(envir = new.env(parent = .BaseNamespaceEnv), {
    .guiname <- "myGui"
    .path <- "~/example.R"
    .custom_get_contents <- function(path) {
        ## get the contents of the document
        readLines(path, warn = FALSE)
    }
    ## recommended to prevent tampering
    lockEnvironment(environment(), bindings = TRUE)
    setHook(packageEvent("this.path"), function(pkgname, pkgpath) {
        this.path::set.gui.path(.guiname, .path, .custom_get_contents)
    }, action = "prepend")
    # ## if your GUI does not have/need a .custom_get_contents
    # ## function, then this works just as well:
    # setHook(packageEvent("this.path"), function(pkgname, pkgpath) {
    #     this.path::set.gui.path(.guiname, .path)
    # }, action = "prepend")
})}
}
\value{
  a list of the previous settings for \code{set.gui.path()}, similar to
  \codelink3{base:options}{options}{()}.
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2024   Iris Simmons
%%


\name{Enhances}
\alias{with_script_path}
\alias{make_fix_file}
\title{Functions That Enhance the Use of Other Packages}
\description{
  These functions improve the user experience of other packages.
}
\usage{
## enchances 'package:box'
with_script_path(expr, file, local = FALSE, n = 0, envir = parent.frame(n + 1),
    matchThisEnv = getOption("topLevelEnvironment"),
    srcfile = if (n) sys.parent(n) else 0)

## enchances 'package:rprojroot'
make_fix_file(criterion, local = FALSE, n = 0, envir = parent.frame(n + 1),
    matchThisEnv = getOption("topLevelEnvironment"),
    srcfile = if (n) sys.parent(n) else 0)
}
\arguments{
  \item{expr}{an expression to evaluate after setting the current script in
    \pkg{package:box}; most commonly a call to
    \codelink3{box:use}{box::use}{()}.}

  \item{file}{a character string giving the pathname of the file.}

  \item{criterion}{argument passed to
    \codelink3{rprojroot:find_root}{rprojroot::find_root}{()}.}

  \item{local, n, envir, matchThisEnv, srcfile}{See
    \code{?\link{this.path}()}.}
}
\details{
  \code{with_script_path()} improves the experience of \CRANpackage{box}; it
  sets the current script in \pkg{package:box} to \code{file} or
  \code{\link{this.path}()} using
  \codelink3{box:script_path}{box::set_script_path}{()}, then evaluates its
  argument, most commonly a \pkg{package:box} import statement.

  \code{make_fix_file()} improves the experience of \CRANpackage{rprojroot}; it
  looks for a project root starting with \code{\link{this.dir}()}, then makes a
  function that constructs file paths against said project root.
}
\value{
  for \code{with_script_path()}, the result of evaluating \code{expr}.

  for \code{make_fix_file()}, a function with formals \code{(..., .. = 0)} that
  returns a character vector.
}
\examples{
# this.path::with_script_path(
# box::use(
#     <import 1>,
#     <import 2>,
#     <...>
# )
# )


# ## replace 'rprojroot::is_r_package' with desired criterion
#
# fix_file <- this.path::make_fix_file(rprojroot::is_r_package)
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2021-2024   Iris Simmons
%%


\name{here}
\alias{here}
\alias{ici}
\alias{this.proj}
\alias{reset.proj}
\title{Construct Path to File, Starting With Script's Directory}
\description{
  \code{here()} constructs file paths starting with \code{\link{this.dir}()}.

  \code{this.proj()} constructs file paths starting with the project root of
  \code{this.dir()}.

  \code{reset.proj()} resets the path cache of \code{this.proj()}. This can be
  useful if you create a new project that you would like to be detected without
  restarting your \R session.
}
\usage{
here(..., local = FALSE, n = 0,
     envir = parent.frame(n + 1),
     matchThisEnv = getOption("topLevelEnvironment"),
     srcfile = if (n) sys.parent(n) else 0, .. = 0)

this.proj(..., local = FALSE, n = 0,
          envir = parent.frame(n + 1),
          matchThisEnv = getOption("topLevelEnvironment"),
          srcfile = if (n) sys.parent(n) else 0)

reset.proj()

## alias for 'here'
ici(..., local = FALSE, n = 0,
    envir = parent.frame(n + 1),
    matchThisEnv = getOption("topLevelEnvironment"),
    srcfile = if (n) sys.parent(n) else 0, .. = 0)
}
\arguments{
  \item{\dots}{further arguments passed to \code{\link{path.join}()}.}

  \item{local, n, envir, matchThisEnv, srcfile}{See
    \code{?\link{this.path}()}.}

  \item{..}{the number of directories to go back.}
}
\details{
  For \code{this.proj()}, the project root has the same criterion as
  \code{here::here()}, but unlike \code{here::here()}, \code{this.proj()}
  supports sub-projects and multiple projects in use at once. Additionally,
  \code{this.proj()} is independent of working directory, whereas
  \code{here::here()} relies on the working directory being set somewhere
  within the project when \pkg{package:here} is loaded. Arguably, this makes it
  better than \code{here::here()}.
}
\value{
  A character vector of the arguments concatenated term-by-term.
}
\examples{
tmpdir <- tempfile(pattern = "dir")
dir.create(tmpdir)

writeLines("this file signifies that its directory is the project root",
    this.path::path.join(tmpdir, ".here"))

FILE.R <- this.path::path.join(tmpdir, "src", "R", "script1.R")
dir.create(this.path::dirname2(FILE.R), recursive = TRUE)
this.path:::.writeCode({
    this.path::this.path()
    this.path::this.proj()
    ## use 'here' to run another script located nearby
    this.path::here("script2.R")
    ## or maybe to read input from a file
    this.path::here(.. = 2, "input", "data1.csv")
    ## but sometimes it is easier to use the project root
    ## this allows you to move the R script up or down
    ## a directory without changing the .. number
    this.path::this.proj("input", "data1.csv")
}, FILE.R)

source(FILE.R, echo = TRUE)

unlink(tmpdir, recursive = TRUE)
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2026   Iris Simmons
%%


\name{set.sys.path}
\alias{set.sys.path}
\alias{set.env.path}
\alias{set.src.path}
\alias{unset.sys.path}
\alias{set.sys.path.function}
\alias{with_sys.path}
\alias{wrap.source}
\title{Implement 'this.path()' For Arbitrary 'source()'-Like Functions}
\description{
  \code{\link{sys.path}()} is implemented to work with these functions and
  packages: \syspathcompatiblelinks

  \code{set.sys.path()} can be used to implement \code{sys.path()} for any
  other \sourcenolink-like functions. Please note that it can only be used
  within a function.

  \code{set.env.path()} and \code{set.src.path()} can be used alongside
  \code{set.sys.path()} to implement \code{\link{env.path}()} and
  \code{\link{src.path}()}, thereby fully implementing
  \code{\link{this.path}()}. Note that \code{set.env.path()} only makes sense
  if the code is being modularized, see \strong{Examples}.

  \code{unset.sys.path()} will undo a call to \code{set.sys.path()}. You will
  need to use this if you wish to call \code{set.sys.path()} multiple times
  within a function.

  \code{set.sys.path.function()} is a special variant of \code{set.sys.path()}
  to be called within \code{callr::r()} on a function with an appropriate
  \ifelse{html}{\codelink2{base:srcfile}{srcref}}{\code{srcref}}.

  \code{with_sys.path()} is a convenient way to evaluate code within the
  context of a file.

  See \code{?\link{sys.path}(local = TRUE)} which returns the path of the
  executing script, confining the search to the local environment in which
  \code{set.sys.path()} was called.

  \code{wrap.source()} should not be used, save for one specific use-case. See
  details.
}
\usage{
set.sys.path(file,
    path.only = FALSE,
    character.only = path.only,
    file.only = path.only,
    conv2utf8 = FALSE,
    allow.blank.string = FALSE,
    allow.clipboard = !file.only,
    allow.stdin = !file.only,
    allow.url = !file.only,
    allow.file.uri = !path.only,
    allow.unz = !path.only,
    allow.pipe = !file.only,
    allow.terminal = !file.only,
    allow.textConnection = !file.only,
    allow.rawConnection = !file.only,
    allow.sockconn = !file.only,
    allow.servsockconn = !file.only,
    allow.customConnection = !file.only,
    ignore.all = FALSE,
    ignore.blank.string = ignore.all,
    ignore.clipboard = ignore.all,
    ignore.stdin = ignore.all,
    ignore.url = ignore.all,
    ignore.file.uri = ignore.all,
    Function = NULL, ofile, delayed = FALSE, assign_to = NULL)

set.env.path(envir, matchThisEnv = getOption("topLevelEnvironment"))

set.src.path(srcfile)

unset.sys.path()

set.sys.path.function(fun)

with_sys.path(file, expr, ...)

wrap.source(expr,
    path.only = FALSE,
    character.only = path.only,
    file.only = path.only,
    conv2utf8 = FALSE,
    allow.blank.string = FALSE,
    allow.clipboard = !file.only,
    allow.stdin = !file.only,
    allow.url = !file.only,
    allow.file.uri = !path.only,
    allow.unz = !path.only,
    allow.pipe = !file.only,
    allow.terminal = !file.only,
    allow.textConnection = !file.only,
    allow.rawConnection = !file.only,
    allow.sockconn = !file.only,
    allow.servsockconn = !file.only,
    allow.customConnection = !file.only,
    ignore.all = FALSE,
    ignore.blank.string = ignore.all,
    ignore.clipboard = ignore.all,
    ignore.stdin = ignore.all,
    ignore.url = ignore.all,
    ignore.file.uri = ignore.all)
}
\arguments{
  \item{expr}{for \code{with_sys.path()}, an expression to evaluate within the
    context of a file.

    for \code{wrap.source()}, an (unevaluated) call to a \sourcenolink-like
    function.}

  \item{file}{a
    \ifelse{html}{\codelink2{base:connections}{connection}}{connection} or a
    character string giving the pathname of the file or URL to read from.}

  \item{path.only}{must \code{file} be an existing path? This implies
    \code{character.only} and \code{file.only} are \code{TRUE} and implies
    \code{allow.file.uri} and \code{allow.unz} are \code{FALSE}, though these
    can be manually changed.}

  \item{character.only}{must \code{file} be a character string?}

  \item{file.only}{must \code{file} refer to an existing file?}

  \item{conv2utf8}{if \code{file} is a character string, should it be converted
    to UTF-8?}

  \item{allow.blank.string}{may \code{file} be a blank string, i.e.
    \blankstring?}

  \item{allow.clipboard}{may \code{file} be \string{clipboard} or a clipboard
    connection?}

  \item{allow.stdin}{may \code{file} be \string{stdin}? Note that
    \string{stdin} refers to the C-level \sQuote{standard input} of the
    process, differing from \codelink3{base:showConnections}{stdin}{()} which
    refers to the R-level \sQuote{standard input}.}

  \item{allow.url}{may \code{file} be a URL pathname or a connection of class
    \string{url-libcurl} / / \string{url-wininet}?}

  \item{allow.file.uri}{may \code{file} be a \samp{file://} URL?}

  \item{allow.unz, allow.pipe, allow.terminal, allow.textConnection,
        allow.rawConnection, allow.sockconn, allow.servsockconn}{may
    \code{file} be a connection of class
    \ifelse{html}{\code{"\link[base:connections]{unz}"}}{\string{unz}}
    / / \string{pipe} / / \string{terminal} / / \string{textConnection} / /
    \string{rawConnection} / / \string{sockconn} / / \string{servsockconn}?}

  \item{allow.customConnection}{may \code{file} be a custom connection?}

  \item{ignore.all, ignore.blank.string, ignore.clipboard, ignore.stdin,
        ignore.url, ignore.file.uri}{ignore the special meaning of these types
    of strings, treating it as a path instead?}

  \item{Function}{character vector of length \ifelse{text}{1}{\code{1}} or
    \ifelse{text}{2}{\code{2}}; the name of the function and package in which
    \code{set.sys.path()} is called.}

  \item{ofile}{a
    \ifelse{html}{\codelink2{base:connections}{connection}}{connection} or a
    character string specifying the original file argument. This overwrites the
    value returned by \code{\link{sys.path}(original = TRUE)}.}

  \item{delayed}{\code{TRUE} or \code{FALSE}; should the
    \ifelse{html}{\codelink2{base:normalizePath}{normalizing}}{normalizing} of
    the path be delayed? Mostly for use with \code{\link{make_fix_funs}()} and
    similar.}

  \item{assign_to}{\code{NULL} or a character string or a symbol; the variable
    name for which to assign the return value in the currently active
    environment.}

  \item{envir, matchThisEnv}{arguments passed to
    \codelink3{base:ns-topenv}{topenv}{()} to determine the top level
    environment in which to assign an associated path.}

  \item{srcfile}{source file in which to assign a pathname.}

  \item{fun}{function with a \code{srcref}.}

  \item{\dots}{further arguments passed to \code{set.sys.path()}.}
}
\details{
  \code{set.sys.path()} should be added to the body of your \sourcenolink-like
  function before reading / / evaluating the expressions.

  \code{wrap.source()}, unlike \code{set.sys.path()}, does not accept an
  argument \code{file}. Instead, an attempt is made to extract the file from
  \code{expr}, after which \code{expr} is evaluated. It is assumed that the
  file is the first argument of the function, as is the case with most
  \sourcenolink-like functions. The function of the call is evaluated, its
  \ifelse{html}{\codelink3{base}{formals}{()}}{\code{formals()}} are retrieved,
  and then the arguments of \code{expr} are searched for a name matching the
  name of the first formal argument. If a match cannot be found by name, the
  first unnamed argument is taken instead. If no such argument exists, the file
  is assumed missing.

  \code{wrap.source()} does non-standard evaluation and does some guess work to
  determine the file. As such, it is less desirable than \code{set.sys.path()}
  when the option is available. I can think of exactly one scenario in which
  \code{wrap.source()} might be preferable: suppose there is a
  \sourcenolink-like function \code{sourcelike()} in a foreign package (a
  package for which you do not have write permission). Suppose that you write
  your own function in which the formals are \code{(\dots)} to wrap
  \code{sourcelike()}: \preformatted{wrapper <- function (...)
{
    ## possibly more args to wrap.source()
    wrap.source(sourcelike(...))
}}

  This is the only scenario in which \code{wrap.source()} is preferable, since
  extracting the file from the \code{\dots} list would be a pain. Then again,
  you could simply change the formals of \code{wrapper()} from \code{(\dots)}
  to \code{(file, \dots)}. If this does not describe your exact scenario, use
  \code{set.sys.path()} instead.
}
\value{
  for \code{set.sys.path()}, if \code{file} is a path, then the normalized
  path with the same attributes, otherwise \code{file} itself. Please note that
  if \code{delayed = TRUE}, the return value will be an absolute path, but not
  the normalized path.

  for \code{set.env.path()}, \code{envir} invisibly.

  for \code{set.src.path()}, \code{srcfile} invisibly.

  for \code{unset.sys.path()} and \code{set.sys.path.function()}, \code{NULL}
  invisibly.

  for \code{with_sys.path()} and \code{wrap.source()}, the result of evaluating
  \code{expr}.
}
\section{Using 'ofile'}{
  \code{ofile} can be used when the \code{file} argument supplied to
  \code{set.sys.path()} is not the same as the \code{file} argument supplied to
  the \sourcenolink-like function:

\preformatted{sourcelike <- function (file)
{
    ofile <- file
    if (!is.character(ofile) || length(ofile) != 1)
        stop(gettextf("'\%s' must be a character string", "file"))
    ## if the file exists, do nothing
    if (file.exists(file)) {
    }
    ## look for the file in the home directory
    ## if it exists, do nothing
    else if (file.exists(file <- this.path::path.join("~", ofile))) {
    }
    ## you could add other directories to look in,
    ## but this is good enough for an example
    else stop(gettextf("'\%s' is not an existing file", ofile))
    file <- this.path::set.sys.path(file, ofile = ofile)
    exprs <- parse(n = -1, file = file)
    for (i in seq_along(exprs)) eval(exprs[i], envir)
    invisible()
}}
}
\examples{
FILE.R <- tempfile(fileext = ".R")
this.path:::.writeCode({
    this.path::sys.path(verbose = TRUE)
    try(this.path::env.path(verbose = TRUE))
    this.path::src.path(verbose = TRUE)
    this.path::this.path(verbose = TRUE)
}, FILE.R)


## here we have a source-like function, suppose this
## function is in a package for which you have write permission
sourcelike <- function (file, envir = parent.frame())
{
    ofile <- file
    file <- set.sys.path(file, Function = "sourcelike")
    lines <- readLines(file, warn = FALSE)
    filename <- sys.path(local = TRUE, for.msg = TRUE)
    isFile <- !is.na(filename)
    if (isFile) {
        timestamp <- file.mtime(filename)[1]
        ## in case 'ofile' is a URL pathname / / 'unz' connection
        if (is.na(timestamp))
            timestamp <- Sys.time()
    }
    else {
        filename <- if (is.character(ofile)) ofile else "<connection>"
        timestamp <- Sys.time()
    }
    srcfile <- srcfilecopy(filename, lines, timestamp, isFile)
    set.src.path(srcfile)
    exprs <- parse(text = lines, srcfile = srcfile, keep.source = FALSE)
    invisible(source.exprs(exprs, evaluated = TRUE, envir = envir))
}


sourcelike(FILE.R)
sourcelike(conn <- file(FILE.R)); close(conn)


## here we have another source-like function, suppose this function
## is in a foreign package for which you do not have write permission
sourcelike2 <- function (pathname, envir = globalenv())
{
    if (!(is.character(pathname) && file.exists(pathname)))
        stop(gettextf("'\%s' is not an existing file",
             pathname, domain = "R-base"))
    envir <- as.environment(envir)
    lines <- readLines(pathname, warn = FALSE)
    srcfile <- srcfilecopy(pathname, lines, isFile = TRUE)
    exprs <- parse(text = lines, srcfile = srcfile, keep.source = FALSE)
    invisible(source.exprs(exprs, evaluated = TRUE, envir = envir))
}


## the above function is similar to sys.source(), and it
## expects a character string referring to an existing file
##
## with the following, you should be able
## to use 'sys.path()' within 'FILE.R':
wrap.source(sourcelike2(FILE.R), path.only = TRUE)


# ## with R >= 4.1.0, use the forward pipe operator '|>' to
# ## make calls to 'wrap.source' more intuitive:
# sourcelike2(FILE.R) |> wrap.source(path.only = TRUE)


## 'wrap.source' can recognize arguments by name, so they
## do not need to appear in the same order as the formals
wrap.source(sourcelike2(envir = new.env(), pathname = FILE.R),
    path.only = TRUE)


## it it much easier to define a new function to do this
sourcelike3 <- function (...)
wrap.source(sourcelike2(...), path.only = TRUE)


## the same as before
sourcelike3(FILE.R)


## however, this is preferable:
sourcelike4 <- function (pathname, ...)
{
    ## pathname is now normalized
    pathname <- set.sys.path(pathname, path.only = TRUE)
    sourcelike2(pathname = pathname, ...)
}
sourcelike4(FILE.R)


## perhaps you wish to run several scripts in the same function
fun <- function (paths, ...)
{
    for (pathname in paths) {
        pathname <- set.sys.path(pathname, path.only = TRUE)
        sourcelike2(pathname = pathname, ...)
        unset.sys.path(pathname)
    }
}


## here we have a source-like function which modularizes its code
sourcelike5 <- function (file)
{
    ofile <- file
    file <- set.sys.path(file, Function = "sourcelike5")
    lines <- readLines(file, warn = FALSE)
    filename <- sys.path(local = TRUE, for.msg = TRUE)
    isFile <- !is.na(filename)
    if (isFile) {
        timestamp <- file.mtime(filename)[1]
        ## in case 'ofile' is a URL pathname / / 'unz' connection
        if (is.na(timestamp))
            timestamp <- Sys.time()
    }
    else {
        filename <- if (is.character(ofile)) ofile else "<connection>"
        timestamp <- Sys.time()
    }
    srcfile <- srcfilecopy(filename, lines, timestamp, isFile)
    set.src.path(srcfile)
    envir <- new.env(hash = TRUE, parent = .BaseNamespaceEnv)
    envir$.packageName <- filename
    oopt <- options(topLevelEnvironment = envir)
    on.exit(options(oopt))
    set.env.path(envir)
    exprs <- parse(text = lines, srcfile = srcfile, keep.source = FALSE)
    source.exprs(exprs, evaluated = TRUE, envir = envir)
    envir
}


sourcelike5(FILE.R)


## the code can be made much simpler in some cases
sourcelike6 <- function (file)
{
    ## we expect a character string refering to a file
    ofile <- file
    filename <- set.sys.path(file, path.only = TRUE, ignore.all = TRUE,
        Function = "sourcelike6")
    lines <- readLines(filename, warn = FALSE)
    timestamp <- file.mtime(filename)[1]
    srcfile <- srcfilecopy(filename, lines, timestamp, isFile = TRUE)
    set.src.path(srcfile)
    envir <- new.env(hash = TRUE, parent = .BaseNamespaceEnv)
    envir$.packageName <- filename
    oopt <- options(topLevelEnvironment = envir)
    on.exit(options(oopt))
    set.env.path(envir)
    exprs <- parse(text = lines, srcfile = srcfile, keep.source = FALSE)
    source.exprs(exprs, evaluated = TRUE, envir = envir)
    envir
}


sourcelike6(FILE.R)


unlink(FILE.R)
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2024   Iris Simmons
%%


\name{sys}
\alias{sys.srcref}
\alias{sys.whiches}
\title{Call Stack Inspection}
\description{
  \code{sys.srcref()} is the R-level version of the function that implements
  \code{\link{this.path}()}. It retrieves the
  \code{\link[base:srcfile]{srcref}} of the call \code{n} generations back from
  itself.

  \code{sys.whiches()} is the R-level version of the function that implements
  \code{sys.srcref()}. It returns the frame numbers of all calls with a common
  parent frame \code{n} generations back from itself.
}
\usage{
sys.srcref(n = 1, which = if (n) sys.parent(n) else 0)
sys.whiches(n = 1, which = if (n) sys.parent(n) else 0)
}
\arguments{
  \item{n}{See \code{?\link{this.path}()}. However, note that \code{n} defaults
    to \code{0} in \code{this.path()} but defaults to \code{1} here.}

  \item{which}{frame number to inspect. An alternative to specifying \code{n}.}
}
\value{
  for \code{sys.srcref()}, a \code{srcref} object or \code{NULL}.

  for \code{sys.whiches()}, an integer vector.
}
\examples{
## this example will not work with 'Run examples'
## which uses 'package:knitr' since knitted
## documents do not store source references
fun <- function ()
{
    list(
        `sys.srcref()` = this.path::sys.srcref(),
        ## while this might seem like a simpler alternative,
        ## you will see it does not work in a couple cases below
        `attr(sys.call(sys.parent()), "srcref")` =
            attr(sys.call(sys.parent()), "srcref")
    )
}


## the braces are unnecessary when using example("sys.srcref"),
## but are needed when copied into the R Console
{ fun() }
{ print(fun()) }
{ try(print(fun())) }





fun2 <- function ()
{
    list(
        ## if copied directly into the R Console,
        ## should be 1 6 7
        ## 1 is the frame number associated with 'try'
        ## 6                                     'force'
        ## 7                                     'fun2'
        this.path::sys.whiches(),


        ## if copied directly into the R Console,
        ## should be  8 12 13
        ##  8 is the frame number associated with 'tryCatch'
        ## 12                                     'identity'
        ## 13                                     'sys.whiches'
        tryCatch(identity(this.path::sys.whiches(0)),
            error = function(e) NULL)
    )
}


try(force(fun2()))


}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2021-2024   Iris Simmons
%%


\name{this.path-package}
\alias{this.path-package}
\docType{package}
\title{Get Script's Path}
\description{
  Determine the path of the executing script.

  Compatible with several popular GUIs:

  \itemize{
    \item{\sQuote{Rgui}}

    \item{\sQuote{\RStudio} (including
      \href{https://docs.posit.co/ide/user/ide/guide/tools/jobs.html}{background jobs})}

    \item{\sQuote{\Positron}}

    \item{\sQuote{\VSCode} + \sQuote{\REditorSupport}}

    \item{\sQuote{\Jupyter}}

    \item{\sQuote{\Emacs} + \sQuote{\ESS}}

    \item{\sQuote{\ifelse{html}{\Rscriptlink}{\ifelse{text}{Rscript}{\code{Rscript}}}}
      (shell)}
  }

  Compatible with several functions and packages: \syspathcompatiblelinks
}
\details{
  The most important functions from \pkg{package:this.path} are
  \code{\link{this.path}()}, \code{\link{this.dir}()}, \code{\link{here}()},
  and \code{\link{this.proj}()}:

  \itemize{
    \item{\code{\link{this.path}()} returns the normalized path of the script
      in which it is written.}

    \item{\code{\link{this.dir}()} returns the directory of
      \code{this.path()}.}

    \item{\code{\link{here}()} constructs file paths against
      \code{this.dir()}.}

    \item{\code{\link{this.proj}()} constructs file paths against the project
      root of \code{this.dir()}.}
  }

  New additions include:

  \itemize{
    \item{\code{\link{LINENO}()} returns the line number of the executing
      expression.}

    \item{\code{\link{shFILE}()} looks through the command line arguments,
      extracting \option{FILE} from either of the following: \option{-f}
      \option{FILE} or \option{--file=FILE}}

    \item{\code{\link{set.sys.path}()} implements \code{\link{this.path}()} for
      any \sourcelink-like functions outside of the builtins.}

    \item{\code{\link{with_init.file}()} allows \code{this.path()} and related
      to be used in a user profile.}
  }

  \pkg{package:this.path} also provides functions for constructing and
  manipulating file paths:

  \itemize{
    \item{\code{\link{path.join}()}, \code{\link{basename2}()}, and
      \code{\link{dirname2}()} are drop in replacements for
      \codelink3{base}{file.path}{()}, \codelink3{base}{basename}{()}, and
      \codelink3{base:basename}{dirname}{()} which better handle drives and
      network shares.}

    \item{\code{\link{splitext}()}, \code{\link{removeext}()},
      \code{\link{ext}()}, and \code{\link{ext<-}()} split a path into root and
      extension, remove a file extension, get an extension, or set an extension
      for a file path.}

    \item{\code{\link{path.split}()}, \code{\link{path.split.1}()}, and
      \code{\link{path.unsplit}()} split the path to a file into components.}

    \item{\code{\link{relpath}()}, \code{\link{rel2here}()}, and
      \code{\link{rel2proj}()} turn absolute paths into relative paths.}
  }
}
\note{
  This package started from a stack overflow posting:

  \Sexpr[results=rd,stage=install]{if (getRversion() >= "3.0.0") "\\\\href{https://stackoverflow.com/questions/1815606/determine-path-of-the-executing-script/64129649#64129649}{https://stackoverflow.com/questions/1815606/determine-path-of-the-executing-script/64129649#64129649}" else "\\\\ifelse{latex}{\\\\href{https://stackoverflow.com/questions/1815606/determine-path-of-the-executing-script/64129649}{https://stackoverflow.com/questions/1815606/determine-path-of-the-executing-script/64129649}}{\\\\href{https://stackoverflow.com/questions/1815606/determine-path-of-the-executing-script/64129649#64129649}{https://stackoverflow.com/questions/1815606/determine-path-of-the-executing-script/64129649#64129649}}"}

  If you like this package, please consider upvoting my answer so that more
  people will see it! If you have an issue with this package, please use
  \codelink3{utils:bug.report}{bug.report}{(package = "this.path")} to
  report your issue.
}
\author{
  Iris Simmons [aut, cre]

  Maintainer: Iris Simmons <ikwsimmo@gmail.com>
}
\keyword{package}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2024   Iris Simmons
%%


\name{print.ThisPathDocumentContext}
\alias{print.ThisPathDocumentContext}
\alias{format.ThisPathDocumentContext}
\alias{as.character.ThisPathDocumentContext}
\title{Printing "ThisPathDocumentContext" Objects}
\description{
  Print a \string{ThisPathDocumentContext} object.
}
\usage{
\method{print}{ThisPathDocumentContext}(x, ..., quote = TRUE)

\method{format}{ThisPathDocumentContext}(x, ...)

\method{as.character}{ThisPathDocumentContext}(x, ...)
}
\arguments{
  \item{x}{object of class \string{ThisPathDocumentContext}.}

  \item{\dots}{unused.}

  \item{quote}{logical, indicating whether or not strings should be printed
    with surrounding quotes.}
}
\details{
  An object of class \string{ThisPathDocumentContext} is generated upon calling
  \code{\link{set.sys.path}()}, \code{\link{wrap.source}()},
  \code{\link{sys.path}()}, \code{\link{env.path}()}, or
  \code{\link{src.path}()}, and by extension \code{\link{this.path}()}. It
  contains information about the path of the executing / / current script.

  These objects are not usually user-visible.
}
\value{
  for \code{print.ThisPathDocumentContext()}, \code{x} invisibly.

  for \code{format.ThisPathDocumentContext()}, a character vector of lines.

  for \code{as.character.ThisPathDocumentContext()}, a character string of
  concatenated lines.
}
\examples{
fun <- function (file)
{
    set.sys.path(file, Function = "fun")
    `.this.path::document.context`
}


fun()


fun("clipboard")


fun(paste0("https://raw.githubusercontent.com/ArcadeAntics/",
           "this.path/main/tests/sys-path-with-urls.R"))


FILE.R <- tempfile(fileext = ".R"); file.create(FILE.R)

x <- fun(FILE.R)
print(x)
print(x, quote = FALSE)
format(x)
as.character(x)

unlink(FILE.R)
}
\keyword{internal}
\keyword{print}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2025   Iris Simmons
%%


\name{ext}
\alias{splitext}
\alias{removeext}
\alias{ext}
\alias{ext<-}
\title{File Extensions}
\description{
  \code{splitext()} splits an extension from a path.

  \code{removeext()} removes an extension from a path.

  \code{ext()} gets the extension of a path.

  \code{ext<-()} sets the extension of a path.
}
\usage{
splitext(path, compression = FALSE, expand = TRUE)
removeext(path, compression = FALSE, expand = TRUE)
ext(path, compression = FALSE, expand = TRUE)
ext(path, compression = FALSE, expand = TRUE) <- value
}
\arguments{
  \item{path}{character vector, containing path names.}

  \item{compression}{should compression extensions \string{.gz}, \string{.bz2},
    and \string{.xz} be taken into account when removing / / getting an
    extension?}

  \item{expand}{logical. Should tilde (see \codelink2{base}{path.expand}) be
    expanded?}

  \item{value}{a character vector, typically of length
    \ifelse{text}{1}{\code{1}} or \code{length(path)}, or \code{NULL}.}
}
\details{
  Trailing path separators are removed before dissecting the path.

  Except for \code{path <- NA_character_}, it will always be true that
  \code{path == paste0(removeext(path), ext(path))}.
}
\value{
  for \code{splitext()}, a matrix with \ifelse{text}{2}{\code{2}} rows and
  \code{length(path)} columns. The first row will be the roots of the paths,
  the second row will be the extensions of the paths.

  for \code{removeext()} and \code{ext()}, a character vector the same length
  as \code{path}.

  for \code{ext<-()}, the updated object.
}
\examples{
splitext(character(0))
splitext("")

splitext("file.ext")

path <- c("file.tar.gz", "file.tar.bz2", "file.tar.xz")
splitext(path, compression = FALSE)
splitext(path, compression = TRUE)

path <- "this.path_2.8.0.tar.gz"
ext(path) <- ".png"
path

path <- "this.path_2.8.0.tar.gz"
ext(path, compression = TRUE) <- ".png"
path
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2024-2026   Iris Simmons
%%


\name{startup_files}
\alias{site.file}
\alias{init.file}
\alias{with_site.file}
\alias{with_init.file}
\title{Files Related to Initialization of the R Session}
\description{
  \code{site.file()} and \code{init.file()} return the
  \ifelse{html}{\codelink2{base:normalizePath}{normalized}}{normalized} paths
  of the site-wide startup profile file and the user profile that were run at
  \codelink2{base:Startup}{startup}.

  \code{with_init.file()} declares that the current script is the user profile
  then evaluates and auto-prints the sub-expressions of its argument.
}
\usage{
site.file(original = FALSE, for.msg = FALSE, default, else.)
init.file(original = FALSE, for.msg = FALSE, default, else.)

with_site.file(expr)
with_init.file(expr)
}
\arguments{
  \item{original, for.msg, default, else.}{Same as \code{\link{shFILE}()}.}

  \item{expr}{a braced expression, the sub-expressions of which to evaluate and
    auto-print.}
}
\value{
  for \code{site.file()} and \code{init.file()}, a character string, or
  \code{default} if it was not found.

  for \code{with_site.file()} and \code{with_init.file()}, \code{expr}
  invisibly.
}
\note{
  \code{with_site.file()} is unneeded now that the site-wide startup profile
  file can be automatically detected.
}
\examples{
## if you wish to use this.path() in a user profile,
## instead of writing:
##
## <expr 1>
## <expr 2>
## <...>
##
## write this:
##
## this.path::with_init.file({
## <expr 1>
## <expr 2>
## <...>
## })
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2021-2024   Iris Simmons
%%


\name{check.path}
\alias{check.path}
\alias{check.dir}
\alias{check.proj}
\title{Check 'this.path()' is Functioning Correctly}
\description{
  Add \code{check.path("path/to/file")} to the start of your script to
  initialize \code{\link{this.path}()} and check that it is returning the
  expected path.
}
\usage{
check.path(...)
check.dir(...)

check.proj(...)
}
\arguments{
  \item{\dots}{further arguments passed to \code{\link{path.join}()} which must
    return a character string; the path you expect \code{this.path()} or
    \code{this.dir()} to return. The specified path can be as deep as necessary
    (just the basename, the last directory and the basename, the last two
    directories and the basename, \ldots), but do not use an absolute path.
    \code{this.path()} makes \R scripts portable, but using an absolute path in
    \code{check.path()} or \code{check.dir()} makes an \R script non-portable,
    defeating a major purpose of this package.}
}
\details{
  \code{check.proj()} is a specialized version of \code{check.path()} that
  checks the path up to the project root.
}
\value{
  if the expected path / / directory matches \code{this.path()} / /
  \code{this.dir()}, then \code{TRUE} invisibly, otherwise an error is thrown.
}
\examples{
# ## I have a project called 'd_cead'
# ##
# ## Within this project, I have a folder called 'code'
# ## where I place all of my scripts.
# ##
# ## One of these scripts is called 'provrun.R'
# ##
# ## So, at the top of that R script, I could write:
#
#
# this.path::check.path("d_cead", "code", "provrun.R")
#
# ## or:
#
# this.path::check.path("d_cead/code/provrun.R")
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2024   Iris Simmons
%%


\name{set.jupyter.path}
\alias{set.jupyter.path}
\title{Declare Executing 'Jupyter' Notebook's Filename}
\description{
  \code{\link{this.path}()} does some guess work to determine the path of the
  executing notebook in \sQuote{\Jupyter}. This involves listing all the files
  in the initial working directory, filtering those which are \R notebooks,
  then filtering those with contents matching the top-level expression.

  This could possibly select the wrong file if the same top-level expression is
  found in another file. As such, you can use \code{set.jupyter.path()} to
  declare the executing \sQuote{Jupyter} notebook's filename.
}
\usage{
set.jupyter.path(...)
}
\arguments{
  \item{\dots}{further arguments passed to \code{\link{path.join}()}. If no
    arguments are provided or exactly one argument is provided that is
    \codelink2{base}{NA} or \codelink2{base}{NULL}, the \sQuote{Jupyter} path
    is unset.}
}
\details{
  This function may only be called from a top-level context in
  \sQuote{Jupyter}. It is recommended that you do \strong{NOT} provide an
  absolute path. Instead, provide just the basename and the directory will be
  determined by the initial working directory.
}
\value{
  character string, invisibly; the declared path for \sQuote{Jupyter}.
}
\examples{
# ## if you opened the file "~/file50b816a24ec1.ipynb", the initial
# ## working directory should be "~". You can write:
#
# set.jupyter.path("file50b816a24ec1.ipynb")
#
# ## and then this.path() will return "~/file50b816a24ec1.ipynb"
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2024   Iris Simmons
%%


\name{FILE}
\alias{FILE}
\alias{LINE}
\title{Macros in Package 'this.path'}
\description{
  \code{FILE()} and \code{LINE()} are intended to be used in a similar manner
  to the macros \code{__FILE__} and \code{__LINE__} in C. They are useful for
  generating a diagnostic message / / warning / / error to about the status of
  the program.
}
\usage{
FILE()
LINE()
}
\examples{
FILE.R <- tempfile(fileext = ".R")
writeLines("fun <- function ()
{
    message(sprintf('invalid value \%d at \%s, line \%d',
        -1, FILE(), LINE()))
}
", FILE.R)
source(FILE.R, verbose = FALSE, keep.source = TRUE)
fun()
unlink(FILE.R)
}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2023-2024   Iris Simmons
%%


\name{source.exprs}
\alias{source.exprs}
\title{Evaluate and Print Expressions}
\description{
  \code{source.exprs()} evaluates and auto-prints expressions as if in a
  toplevel context.
}
\usage{
source.exprs(exprs, evaluated = FALSE, envir = parent.frame(),
             echo = TRUE, print.eval = TRUE)
}
\arguments{
  \item{exprs, evaluated, envir, echo, print.eval}{See
    \codelink3{base:source}{withAutoprint}{()}.}
}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2025   Iris Simmons
%%


\name{basename2}
\alias{basename2}
\alias{dirname2}
\title{Manipulate File Paths}
\description{
  \code{basename2()} removes all of the path up to and including the last path
  separator (if any).

  \code{dirname2()} returns the part of the \code{path} up to but excluding the
  last path separator, or \string{.} if there is no path separator.
}
\usage{
basename2(path, expand = TRUE)
dirname2(path, expand = TRUE)
}
\arguments{
  \item{path}{character vector, containing path names.}

  \item{expand}{logical. Should tilde (see \codelink2{base}{path.expand}) be
    expanded?}
}
\details{
  Trailing path separators are removed before dissecting the path, and for
  \code{dirname2()} any trailing file separators are removed from the result.
}
\value{
  A character vector of the same length as \code{path}.
}
\section{Behaviour on Windows}{
  If \code{path} is an empty string, then both \code{dirname2()} and
  \code{basename2()} return an emty string.

  \ifelse{text}{\\}{\code{\\}} and \ifelse{text}{/}{\code{/}} are accepted as
  path separators, and \code{dirname2()} does \strong{NOT} translate the path
  separators.

  Recall that a network share looks like \string{//host/share} and a drive
  looks like \string{d:}.

  For a path which starts with a network share or drive, the path specification
  is the portion of the string immediately afterward, e.g.
  \string{/path/to/file} is the path specification of
  \string{//host/share/path/to/file} and \string{d:/path/to/file}. For a path
  which does not start with a network share or drive, the path specification is
  the entire string.

  The path specification of a network share will always be empty or absolute,
  but the path specification of a drive does not have to be, e.g.
  \string{d:file} is a valid path despite the fact that the path specification
  does not start with \string{/}.

  If the path specification of \code{path} is empty or is \string{/}, then
  \code{dirname2()} will return \code{path} and \code{basename2()} will return
  an empty string.
}
\section{Behaviour under Unix-alikes}{
  If \code{path} is an empty string, then both \code{dirname2()} and
  \code{basename2()} return an emty string.

  Recall that a network share looks like \string{//host/share}.

  For a path which starts with a network share, the path specification is the
  portion of the string immediately afterward, e.g. \string{/path/to/file} is
  the path specification of \string{//host/share/path/to/file}. For a path
  which does not start with a network share, the path specification is the
  entire string.

  If the path specification of \code{path} is empty or is \string{/}, then
  \code{dirname2()} will return \code{path} and \code{basename2()} will return
  an empty string.
}
\examples{
path <- c("/usr/lib", "/usr/", "usr", "/", ".", "..")
x <- cbind(path, dirname = dirname2(path), basename = basename2(path))
print(x, quote = FALSE, print.gap = 3)
}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2021-2024   Iris Simmons
%%


\name{this.path-defunct}
\alias{this.path-defunct}
\alias{this.path2}
\alias{this.dir2}
\alias{this.dir3}
\alias{normalized.shFILE}
\alias{as.rel.path}
\alias{as.relative.path}
\alias{local.path}
\alias{reset.this.proj}
\alias{set.this.path.jupyter}
\alias{inside.source}
\alias{set.this.path}
\alias{unset.this.path}
\alias{set.sys.path.jupyter}
\alias{fileArgs}
\title{Defunct Functions in Package 'this.path'}
\description{
  The functions or variables listed here are no longer part of
  \pkg{this.path} as they are no longer needed.
}
\usage{
# Defunct in this.path 1.1.0 (2022-12-01)
this.path2(...)
this.dir2(...)
this.dir3(...)
normalized.shFILE(default, else.)

# Defunct in this.path 1.3.0 (2023-04-09)
as.rel.path(path)
as.relative.path(path)

# Defunct in this.path 1.4.0 (2023-04-18)
local.path(verbose = getOption("verbose"), original = FALSE,
    for.msg = FALSE, contents = FALSE, default, else.)

# Defunct in this.path 2.0.0 (2023-08-09)
reset.this.proj()
set.this.path.jupyter(...)
inside.source(file, path.only = FALSE, character.only = path.only,
    file.only = path.only, conv2utf8 = FALSE, allow.blank.string = FALSE,
    allow.clipboard = !file.only, allow.stdin = !file.only, allow.url = !file.only,
    allow.file.uri = !path.only, allow.unz = !path.only, allow.pipe = !file.only,
    allow.terminal = !file.only, allow.textConnection = !file.only,
    allow.rawConnection = !file.only, allow.sockconn = !file.only,
    allow.servsockconn = !file.only, allow.customConnection = !file.only,
    ignore.all = FALSE, ignore.blank.string = ignore.all, ignore.clipboard = ignore.all,
    ignore.stdin = ignore.all, ignore.url = ignore.all, ignore.file.uri = ignore.all,
    Function = NULL, ofile)
set.this.path(file, path.only = FALSE, character.only = path.only,
    file.only = path.only, conv2utf8 = FALSE, allow.blank.string = FALSE,
    allow.clipboard = !file.only, allow.stdin = !file.only, allow.url = !file.only,
    allow.file.uri = !path.only, allow.unz = !path.only, allow.pipe = !file.only,
    allow.terminal = !file.only, allow.textConnection = !file.only,
    allow.rawConnection = !file.only, allow.sockconn = !file.only,
    allow.servsockconn = !file.only, allow.customConnection = !file.only,
    ignore.all = FALSE, ignore.blank.string = ignore.all, ignore.clipboard = ignore.all,
    ignore.stdin = ignore.all, ignore.url = ignore.all, ignore.file.uri = ignore.all,
    Function = NULL, ofile)
unset.this.path()

# Defunct in this.path 2.4.0 (2024-02-16)
set.sys.path.jupyter(...)
fileArgs()
}
\details{
  \code{this.path2()}, \code{this.dir2()}, and \code{this.dir3()} were used
  before \code{\link{this.path}()} and \code{\link{this.dir}()} were given an
  argument \code{default}. Should be replaced by
  \code{\link{this.path}(default = NULL)},
  \code{\link{this.dir}(default = NULL)}, and
  \ifelse{html}{\code{\link{this.dir}(default = \link[base]{getwd}())}}{\code{this.dir(default = getwd())}},
  respectively.

  \code{normalized.shFILE()} was used before \code{\link{shFILE}()} was given
  an argument \code{original}. Should be replaced by \code{\link{shFILE}()}.

  \code{as.rel.path()} and \code{as.relative.path()} were poorly chosen names
  given their functionality. Should be replaced by \code{\link{rel2here}()}.

  \code{local.path()} was used when it called a different C function than
  \code{\link{this.path}()}. Should be replaced by
  \code{\link{this.path}(local = TRUE)}.

  \code{reset.this.proj()} was used when \code{\link{this.proj}()} was the only
  project related function. Now that \code{\link{src.proj}()},
  \code{\link{env.proj}()}, and \code{\link{sys.proj}()} exist, the function
  was renamed to \code{\link{reset.proj}()}.

  In \pkg{this.path 2.0.0}, the original implementation of
  \code{\link{this.path}()} was renamed to \code{\link{sys.path}()}, so
  \code{set.this.path.jupyter()}, \code{set.this.path()}, and
  \code{unset.this.path()} were accordingly renamed to
  \code{\link{set.jupyter.path}()}, \code{\link{set.sys.path}()}, and
  \code{\link{unset.sys.path}()}. Since \code{set.this.path()} was already
  being made defunct, \code{inside.source()} was also made defunct.

  \code{set.sys.path.jupyter()} was renamed to
  \code{\link{set.jupyter.path}()}.

  \code{fileArgs()} should be replaced by \code{\link{progArgs}()}.
}
\seealso{
  \code{\link{Defunct}}
}
\keyword{internal}

%%
%% this.path : Get Executing Script's Path
%% Copyright (C) 2022-2026   Iris Simmons
%%


\name{relpath}
\alias{relpath}
\alias{rel2here}
\alias{rel2proj}
\title{Make a Path Relative to Another}
\description{
  When working with \pkg{this.path}, you will be dealing with a lot of absolute
  paths. These paths are not portable for saving within files nor tables, so
  convert them to relative paths with \code{relpath()}.
}
\usage{
#ifdef unix
relpath(path, relative.to = getwd())
#endif
#ifdef windows
relpath(path, relative.to = normalizePath(getwd(), "/", TRUE))
#endif

rel2here(path, local = FALSE, n = 0, envir = parent.frame(n + 1),
         matchThisEnv = getOption("topLevelEnvironment"),
         srcfile = if (n) sys.parent(n) else 0)

rel2proj(path, local = FALSE, n = 0,
         envir = parent.frame(n + 1),
         matchThisEnv = getOption("topLevelEnvironment"),
         srcfile = if (n) sys.parent(n) else 0)
}
\arguments{
  \item{path}{character vector of file / / URL pathnames.}

  \item{relative.to}{character string; the file / / URL pathname to make
    \code{path} relative to.}

  \item{local, n, envir, matchThisEnv, srcfile}{See
    \code{?\link{this.path}()}.}
}
\details{
  Tilde-expansion (see \codelink4{?}{base}{path.expand}{()}) is first done on
  \code{path} and \code{relative.to}.

  If \code{path} and \code{relative.to} are equivalent, \string{.} will be
  returned. If \code{path} and \code{relative.to} have no base in common, the
  \ifelse{html}{\code{\link[base:normalizePath]{normalized}}}{normalized}
  \code{path} will be returned.
}
\value{
  character vector of the same length as \code{path}.
}
\examples{
#ifdef unix
\dontrun{
relpath(
    c(
        ## paths which are equivalent will return "."
        "/home/effective_user/Documents/this.path/man",


        ## paths which have no base in common return as themselves
        paste0("https://raw.githubusercontent.com/ArcadeAntics/",
               "this.path/main/tests/sys-path-with-urls.R"),
        "//host-name/share-name/path/to/file",


        "/home/effective_user/Documents/testing",
        "/home/effective_user",
        "/home/effective_user/Documents/R/thispath.R"
    ),
    relative.to = "/home/effective_user/Documents/this.path/man"
)
}
#endif
#ifdef windows
\dontrun{
relpath(
    c(
        ## paths which are equivalent will return "."
        "C:/Users/effective_user/Documents/this.path/man",


        ## paths which have no base in common return as themselves
        paste0("https://raw.githubusercontent.com/ArcadeAntics/",
               "this.path/main/tests/sys-path-with-urls.R"),
        "D:/",
        "//host-name/share-name/path/to/file",


        "C:/Users/effective_user/Documents/testing",
        "C:\\\\Users\\\\effective_user",
        "C:/Users/effective_user/Documents/R/thispath.R"
    ),
    relative.to = "C:/Users/effective_user/Documents/this.path/man"
)
}
#endif
}
